/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/******************************************************************************
 * this file provides the functionality of the options 'loadappl
 * node' and 'hide appl node' in menu 'cl'
 *****************************************************************************/

#include "listener.h"
#include <exErrors.h>

static char SccsID[] = "@(#)loadhide.c	1.2\t11/25/92";            /* sccs keyword */

static MenuDescr menus[MDS_MAXAPPL][MDS_MAXMENU];     /* menus created */

extern Listener *listener;

MHmenu *loadApplMenu();
MHmenu *hideApplMenu();
int createItemList();
Widget getPullDownDescr();
int redisplayMenuBar();
int destroyPopup();

char *getenv();

/* loadApplNode **************************************************************
 * application function to (re)load user application nodes
 *****************************************************************************/

loadApplNode(parms)
     RequestParm *parms;
{
  applId *userAppl[MDS_MAXAPPL];
  MHmenu *mptrAppl;
  ApplicationNode *aptr;
  exMenu *m;                                          
  char *result, *file, *menupath;
  char fullName[256];
  char msg[100];

  int acnt = MDS_MAXMENU, i;
  
  bzero((char *) userAppl, MDS_MAXAPPL * sizeof(applId *));

  if ((! (mptrAppl = loadApplMenu(listener, userAppl, MDS_MAXAPPL))) || 
      (! (m = exMakeMenu((ApplicationNode *)NULL, (exMenubar *)NULL, 
			 parms->parent, mptrAppl)))) {
    for (i = 0; i < MDS_MAXAPPL; i++) {
      if (userAppl[i] != (applId *) NULL) {
	free(userAppl[i]->name);
	free(userAppl[i]->file);
      }
      free(userAppl[i]);
    }

    return(-1);
  }

  exCallMenu(m);

  if (! (result = exGetMenuReturnValue(m))) {
    freeMenu(m);
    for (i = 0; i < MDS_MAXAPPL; i++) {
      if (userAppl[i] != (applId *) NULL) {
	free(userAppl[i]->name);
	free(userAppl[i]->file);
      }
      free(userAppl[i]);
    }
    return(-1);
  }

  /* check if application node is already loaded */
  file = userAppl[atoi(result)]->file;
  if (*file == '/') {				/* absolute file name */
    strcpy(fullName, file);
  } else {					/* create file name */
    if (! (menupath = getenv("MENUPATH"))) {
      if (! (menupath = getenv("HOME"))) {
	freeMenu(m);
	for (i = 0; i < MDS_MAXAPPL; i++) {
	  if (userAppl[i] != (applId *) NULL) {
	    free(userAppl[i]->name);
	    free(userAppl[i]->file);
	  }
	  free(userAppl[i]);
	}
	return(-1);
      } 
    }
    
    if (makeFullName(menupath, file, fullName, 256) < 0) {
      setError(ERR_MAX);
      freeMenu(m);  
      for (i = 0; i < MDS_MAXAPPL; i++) {
	if (userAppl[i] != (applId *) NULL) {
	  free(userAppl[i]->name);
	  free(userAppl[i]->file);
	}
	free(userAppl[i]);
      }
      return(-1);
    }
  }
  
  if (aptr = getApplByName(GetNameFromFile(fullName))) {
    sprintf(msg, "Application node --%s-- already loaded; replace it ?",
	    aptr->applicationName);
    if (exNotify2User(listener->listener_shell, -1, -1, msg) != ANSWERYES) {
      freeMenu(m);
      for (i = 0; i < MDS_MAXAPPL; i++) {
	if (userAppl[i] != (applId *) NULL) {
	  free(userAppl[i]->name);
	  free(userAppl[i]->file);
	}
	free(userAppl[i]);
      }
      return(1);
    }
    
    redisplayMenuBar(listener->menubar, aptr);
    destroyPopup(menus[getMDSindex(aptr)]);
    removeApplNode(aptr);  
  }
  
  if (restoreFromFile(parms->parent, userAppl[atoi(result)]->file, 0))
    loadCurrentApplication(listener->menubar, listener->listener_shell, 
			 		menus[getCurAppl()], &acnt); 

  freeMenu(m);
  for (i = 0; i < MDS_MAXAPPL; i++) {
    if (userAppl[i] != (applId *) NULL) {
      free(userAppl[i]->name);
      free(userAppl[i]->file);
    }
    free(userAppl[i]);
  }

  return(1);
}

/* hideApplNode ***************************************************************
 * application function to hide user application nodes
 *****************************************************************************/

/*ARGSUSED*/
hideApplNode(parms)
     RequestParm *parms;
{
  char *applMenuItems[MDS_MAXAPPL];
  MHmenu *mptrAppl;
  ApplicationNode *hideAptr;
  exMenu *m;                                          
  char *result;
  int i;
  
  for (i = 0; i < MDS_MAXAPPL; i++) 
    applMenuItems[i] = (char *) NULL;

  if (! (mptrAppl = 
	 hideApplMenu(applMenuItems, MDS_MAXAPPL, listener->menubar)) ||
      ! (m = exMakeMenu((ApplicationNode *)NULL, (exMenubar *)NULL,
			listener->parent, mptrAppl))) 
    return(-1);

  exCallMenu(m);

  if ((result = exGetMenuReturnValue(m)) == NULL) {
    freeMenu(m);
    return(-1);
  }

  hideAptr = getApplByName(applMenuItems[atoi(result)]);
  redisplayMenuBar(listener->menubar, hideAptr);
  destroyPopup(menus[getMDSindex(hideAptr)]);
  removeApplNode(hideAptr);  

  freeMenu(m);
  for (i = 0; i < MDS_MAXAPPL; i++) {
    if (applMenuItems[i])
      free(applMenuItems[i]);
    else
      break;
  }
  return(0);
}

/* loadApplMenu **************************************************************
 * creates menu which lists user application names
 ****************************************************************************/

MHmenu *loadApplMenu(l, userApplList, size)
     Listener *l;
     applId *userApplList[];
     int size;
{
  char *getenv();
  MHmenu *applMenu;
  char *applMenuItems[MDS_MAXAPPL], *menupath;
  int cnt, i, j, indx;

  if ((menupath = getenv("MENUPATH")) == (char *) 0) {
    setError(ERR_NULL);
    exAlert(listener->listener_shell, -1, -1, 
	    		"No environment-variable set: MENUPATH");
    return((MHmenu *) NULL);
  }

  if ((cnt = getApplIdentifier(menupath, userApplList, size)) == -1) 
    return((MHmenu *) NULL);
  
  setWorkApplication();

  for (i = 0, j = 0; i < cnt; i++)	/* check for system menus */
    if (userApplList[i] && 
	strcmp(userApplList[i]->name, l->menubar->buttons[0].menu->applName)) 
      applMenuItems[j++] = strCopy(userApplList[i]->name);
    else {                      /* don't list the system application node */
      indx = i;
      bcopy(&userApplList[indx + 1], &userApplList[indx], 
	    				(cnt - indx - 1) * sizeof(applId *));
    }

  if (! (applMenu = buildMenuFromStrings("applications nodes", 
       				MDS_L_POPUP, applMenuItems, NULL, j))) {
    setError(ERR_NOEXIST);
    return((MHmenu *) NULL);
  }

  setMenuPos(applMenu, 0, 0, MDS_POS_CREL);  
  insertMenu(applMenu); 

  return(applMenu);
}

/* hideApplMenu ***************************************************************
 * creates a menu containing the names of all applications loaded from the 
 * user
 *****************************************************************************/

MHmenu *hideApplMenu(itemList, size, mbar)
     char *itemList[];
     int size;
     exMenubar *mbar;
{
  MHmenu *applMenu;
  int cnt;

  setWorkApplication();

  cnt = createItemList(itemList, size, mbar);

  if (! (applMenu = buildMenuFromStrings("loaded applications", 
	 		MDS_L_POPUP, itemList, NULL, cnt))) {
    setError(ERR_CREATE);
    return((MHmenu *) NULL);
  }

  insertMenu(applMenu);
  setMenuPos(applMenu, 0, 0, MDS_POS_CREL);

  return(applMenu);
}

/* createItemList *********************************************************
 * creates item list with names of applications which are contained in the
 * MDS-structure but are not temporary or system applications; don't create
 * an entry for the current application's menus.
 ***************************************************************************/

createItemList(iList, num, mb)
     char *iList[];
     int num;
     exMenubar *mb;
{
  register i, j, count = 0;

  for (i = 0; (i < MDS_MAXAPPL) && count < num; i++)
    if (MDS[i] && (! (MDS[i]->a_access & (MDS_TEMP | MDS_SYS))) &&
	strcmp(MDS[i]->applicationName, 
	       listener->menubar->buttons[0].menu->applName)) {
      if (menus[i][0].menu)    /* user has loaded MDS[i] with popup menus */
	iList[count++] = strCopy(MDS[i]->applicationName); 
      else 
	for (j = 0; j < MDS[i]->menuCnt; j++)
	  if (MDS[i]->m_table[j] && 
	      ((MDS[i]->m_table[j]->m_type == MDS_M_PULLD) | 
	       (MDS[i]->m_table[j]->m_type == MDS_L_PULLD))) {
	    if (getPullDownDescr(MDS[i], mb, 
				 MDS[i]->m_table[j]->m_repr->name)) {
	               /* user has loaded MDS[i] containing pulldown menus */
	      iList[count++] = strCopy(MDS[i]->applicationName);
	      break;
	    }
	  }
    }

  return(count);
}

/* redisplayMenuBar ***********************************************************
 * destroys the button for a pulldown menu; a callback is activated when a 
 * button is deleted that automatically frees the menu (see destroyMenu)
 *****************************************************************************/

redisplayMenuBar(mb, aptr)
     exMenubar *mb;
     ApplicationNode *aptr;
{
  Widget button;
  register i;

  for (i = 0; i < MDS_MAXMENU; i++)
    if (aptr->m_table[i] && 
	((aptr->m_table[i]->m_type == MDS_M_PULLD) ||
	 		(aptr->m_table[i]->m_type == MDS_L_PULLD))) {
      if (button = getPullDownDescr(aptr, mb, aptr->m_table[i]->m_repr->name))
	XtDestroyWidget(button);
    }
}

/* getPullDownDescr ***********************************************************
 * searches for pulldown menu with input name in menu-descriptions 
 *****************************************************************************/
 
Widget getPullDownDescr(aptr, mbar, name)
     ApplicationNode *aptr;
     exMenubar *mbar;
     char *name;
{
  register i;

  if (mbar->num_buttons < 1)			/* no pulldown menus */
    return((Widget) NULL);

  for (i = 0; i < mbar->num_buttons; i++) {
    if (! mbar->buttons[i].menu)		/* just to be shure */
      continue;
    if (strcmp(aptr->applicationName, mbar->buttons[i].menu->applName) == 0)
      if (strcmp(mbar->buttons[i].menu->name, name) == 0)
	return(mbar->buttons[i].button);	/* found it */
  }
  return((Widget) NULL);				/* not found */
}

/* destroyPopup ***************************************************************
 * destroys popup menu windows
 *****************************************************************************/

destroyPopup(popupwin)
     register MenuDescr popupwin[];
{
  register i;

  for (i = 0; i < MDS_MAXMENU; i++)
    if (popupwin[i].menu) {
      if (popupwin[i].type == MDS_M_POPUP || popupwin[i].type == MDS_L_POPUP)
	XtDestroyWidget(popupwin[i].menu->w);
      popupwin[i].menu = (exMenu *) NULL;
      free(popupwin[i].name);
    }
}

/* getPopup: ******************************************************************
 * looks if a popup menu with the given name is conatained in the popups-array,
 * that holds descriptions of all popuMenus created sofar. It returns the
 * menu-id if a menu with the given is found.
 *****************************************************************************/

exMenu *getPopup(name)
     register char *name;
{
  register applCnt, menuCnt;

  for (applCnt = 0; applCnt < MDS_MAXAPPL; applCnt++) {
    for (menuCnt = 0; menuCnt < MDS_MAXMENU; menuCnt++) {
      if (menus[applCnt][menuCnt].type == MDS_M_POPUP 
	                         && menus[applCnt][menuCnt].name)
	if (strcmp(menus[applCnt][menuCnt].name, name) == 0)
	  return(menus[applCnt][menuCnt].menu);
    }
  }

  return((exMenu *) 0);
}

/* createPopupList: ***********************************************************
 * create a list of names of the currenly loaded popupmenus 
 *****************************************************************************/

createPopupList(list, size)
     char *list[];
     int size;
{
  register applCnt, menuCnt, cnt = 0;

  for (applCnt = 0; applCnt < MDS_MAXAPPL; applCnt++)
    for (menuCnt = 0; menuCnt < MDS_MAXMENU; menuCnt++)
      if (menus[applCnt][menuCnt].type == MDS_M_POPUP 
	  		&& menus[applCnt][menuCnt].name) {
	list[cnt++] = strCopy(menus[applCnt][menuCnt].name);
	if (cnt > size)
	  return(size);
      }
  return(cnt);
}
