/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
  *                                                                      *
 ************************************************************************/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <signal.h>
#include <strings.h>

#include "fm.h"

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/ToggleB.h>

static char SccsID[] = "@(#)aedinfo.c	1.3\t11/25/92";            /* sccs keyword */

struct passwd *pw, *getpwuid();
struct group *gr, *getgrgid(), *getgrnam();

static void toggleAccessBit();
static void toggle_sAccessBit();
static void saveAed();
static void revertAed();
static void quitAed();
static void closeAed();
static void nextAed();
static void prevAed();
static void setGroup();
static entryHelp();
static bell();

Boolean needsUpdate();

extern char *getenv();

static XmString label_string = NULL;

static char msg[128];

static struct {
  group_elem *groups; 
  int group_cnt;
} group_list;

static editCmd cmdTable [] = {
  {ctrl('f'), right},		/* cursor right */
  {ctrl('b'), left},		/* cursor left  */
  {ctrl('a'), beglin},		/* beginning of line */
  {ctrl('e'), endlin},		/* end of line */
  {ctrl('d'), delnxt},	       	/* delete next character */
  {127,       delprev},         /* delete previous char */
  {ctrl('k'), krest},		/* delete to end of line */
  {ctrl('l'), newscr},		/* redraw screen */
  {ctrl('m'), bell},		/* ring the bell */
  {ctrl('j'), bell},		/* ring the bell */
  {ctrl('c'), bell},      	/* ring the bell */
  {EOS,       (int (*)()) 0},	/* end of list */
};
	
static editCmd ctrlxTable [] = {		       
  {'k', killBuffer},		/* clear the current buffer */
  {'?', entryHelp},		/* display help information */
  {ctrl('c'), bell},        	/* ring the bell */
  {EOS, (int (*)()) 0},		
};

static editCmd escTable [] = {
  {'=', memory},		/* display statistics */
  {'b', wordl},			/* move to left word */
  {'f', wordr},			/* move to right word */
  {'w', fdelword},		/* delete next word */
  {127, bdelword},		/* delete previous word */
  {ctrl('c'), bell},        	/* ring the bell1 */
  {EOS, (int (*)()) 0},		
};

static cmdDescr entryCmds[] = {
  {entryHelp,	      " provide some help information"}
};

/* exCreateAedForm: **********************************************************
 * create the interactive attribute editor 
 ****************************************************************************/

aedForm *exCreateAedForm(parent, x, y, exit_on_close)
     Widget parent;
     int x, y;
     Boolean exit_on_close;
{
  Arg args[20];
  int i, j, cnt;
  aedForm *aed;
  
  aed = (aedForm *) Malloc(sizeof(aedForm));
  aed->display   = XtDisplay(parent);
  aed->ownerBuf = (char *) malloc(20); 
  *aed->ownerBuf = '\0';
  
  /* create toplevel shell widget */
  cnt = 0;
  XtSetArg(args[cnt], XmNx, x); cnt++;
  XtSetArg(args[cnt], XmNy, y); cnt++;
  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;

  aed->aed_shell = XtCreatePopupShell("aed", 
			 applicationShellWidgetClass, parent, args, cnt);

  /* create toplevel form-widget */
  cnt = 0;
  XtSetArg(args[cnt], XmNmarginHeight, 5); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, 5); cnt++;
  aed->aed_form = XtCreateManagedWidget("form", xmFormWidgetClass,
					  	aed->aed_shell, args, cnt);

  /* create the various labels */
  for (i = 0; i < INODE; i++) {
    cnt = 0;
    XtSetArg(args[cnt], XmNalignment, XmALIGNMENT_BEGINNING); cnt++;
    if (i == NAME) {
      XtSetArg(args[cnt], XmNtopOffset, 10); cnt++;
      XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
    } else {
      XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
      XtSetArg(args[cnt], XmNtopWidget, aed->labels[i - 1]); cnt++;
    }
    XtSetArg(args[cnt], XmNleftOffset, 10); cnt++;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNrecomputeSize, True); cnt++;
    XtSetArg(args[cnt], XmNlabelString, 
	     label_string ? label_string : XmStringCreate("                                                                             ", 
				XmSTRING_DEFAULT_CHARSET)); cnt++;
    aed->labels[i] = XtCreateManagedWidget("label", xmLabelWidgetClass, 
					   aed->aed_form, args, cnt);
  }

  /* create more labels */
  for (i = INODE; i < ORIG_ACCESS; i++) {
    cnt = 0;
    XtSetArg(args[cnt], XmNalignment, XmALIGNMENT_BEGINNING); cnt++;
    if (i == INODE) {
      XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
      XtSetArg(args[cnt], XmNtopWidget, aed->labels[NAME]); cnt++;
    } else {
      XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
      XtSetArg(args[cnt], XmNtopWidget, aed->labels[i - 1]); cnt++;
    } 

    XtSetArg(args[cnt], XmNleftOffset, 300); cnt++;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNlabelString, label_string); cnt++;
    
    aed->labels[i] = XtCreateManagedWidget("label", xmLabelWidgetClass, 
					   aed->aed_form, args, cnt);
  }

  /* create entry for owner */
  cnt = 0;
  XtSetArg(args[cnt], XmNleftOffset, 10); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, aed->labels[LASTCHANGE]); cnt++;
  XtSetArg(args[cnt], XmNadjustMargin, False); cnt++;

  aed->ownerForm = XtCreateManagedWidget("owner", 
		   xmDrawingAreaWidgetClass, aed->aed_form, args, cnt);

  aed->owner = exMakeEntry(aed->ownerForm, &aed->ownerBuf, 1,
       20, "owner:", cmdTable, escTable, ctrlxTable, entryCmds, True, False);

  cnt = 0;
  XtSetArg(args[cnt], XmNleftOffset, 20); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, aed->ownerForm); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreate("group: ", XmSTRING_DEFAULT_CHARSET)); cnt++; 

  aed->labels[GROUP_LABEL] = XtCreateManagedWidget("label", 
			     xmLabelWidgetClass, aed->aed_form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, aed->labels[GROUP_LABEL]); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, aed->ownerForm); cnt++;
  XtSetArg(args[cnt], XmNalignment, XmALIGNMENT_BEGINNING); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 	
      XmStringCreate("                  ", XmSTRING_DEFAULT_CHARSET)); cnt++;
  
  aed->group = XtCreateManagedWidget("group", 
			xmPushButtonWidgetClass, aed->aed_form, args, cnt);
  XtAddCallback(aed->group, XmNactivateCallback, setGroup, aed);


  /* create access label */
  cnt = 0;
  XtSetArg(args[cnt], XmNalignment, XmALIGNMENT_BEGINNING); cnt++;
  XtSetArg(args[cnt], XmNtopOffset, 20); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, aed->labels[GROUP_LABEL]); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	XmStringCreate("access rights: ", XmSTRING_DEFAULT_CHARSET)); cnt++; 
  
  aed->labels[ACCESS_LABEL] = XtCreateManagedWidget("label", 
			xmLabelWidgetClass, aed->aed_form, args, cnt);


  for (i = ORIG_ACCESS; i < GROUP_LABEL; i++) {
    cnt = 0;
    XtSetArg(args[cnt], XmNalignment, XmALIGNMENT_BEGINNING); cnt++;
    XtSetArg(args[cnt], XmNtopOffset, 20); cnt++;
    XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNtopWidget, aed->labels[GROUP_LABEL]); cnt++;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNleftWidget, aed->labels[i - 1]); cnt++;
    if (i == ORIG_ACCESS) { 
    XtSetArg(args[cnt], XmNlabelString, 
	XmStringCreate("           ", XmSTRING_DEFAULT_CHARSET)); cnt++; 
    } else {
      XtSetArg(args[cnt], XmNlabelString, 
	XmStringCreate("            ", XmSTRING_DEFAULT_CHARSET)); cnt++; 
    }
    
    aed->labels[i] = XtCreateManagedWidget("label", 
				xmLabelWidgetClass, aed->aed_form, args, cnt);
  }
    
  /* create access bits */
  for (i = 0; i < 4; i++) {
    cnt = 0;
    XtSetArg(args[cnt], XmNleftOffset, 40); cnt++;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNalignment, XmALIGNMENT_BEGINNING); cnt++;
    if (i == 0) {
      XtSetArg(args[cnt], XmNtopOffset, 5); cnt++;
      XtSetArg(args[cnt], XmNtopWidget, aed->labels[ACCESS_LABEL]); cnt++;
    } else {
      XtSetArg(args[cnt], XmNtopWidget, aed->bits[i - 1].form); cnt++;
    }
    XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
    aed->bits[i].form = XtCreateManagedWidget("access", 
				 xmFormWidgetClass, aed->aed_form, args, cnt);
    
    cnt = 0;
    XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNalignment, XmALIGNMENT_BEGINNING); cnt++;
    XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreate(acc_labels[i], XmSTRING_DEFAULT_CHARSET)); cnt++; 
    
    aed->bits[i].label = XtCreateManagedWidget("label", 
			   xmLabelWidgetClass, aed->bits[i].form, args, cnt);


    for (j = 0; j < 3; j++) {
      cnt = 0;
      XtSetArg(args[cnt], XmNfillOnSelect, False); cnt++;
      XtSetArg(args[cnt], XmNindicatorOn, False); cnt++;
      XtSetArg(args[cnt], XmNalignment, XmALIGNMENT_BEGINNING); cnt++;
      XtSetArg(args[cnt], XmNlabelString, access_bits[0]); cnt++;
      XtSetArg(args[cnt], XmNuserData, &aed->bits[i]); cnt++;

      aed->bits[i].toggles[j] = XtCreateManagedWidget("button", 
		    xmToggleButtonWidgetClass, aed->bits[i].form, args, cnt);
      XtAddCallback(aed->bits[i].toggles[j], 
		    XmNvalueChangedCallback, toggleAccessBit, &j);
      
      aed->bits[i].bit[j] = 0;
    }
    
    aed->bits[i].indx = i;
    aed->bits[i].aed = aed;

    cnt = 0;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNleftWidget, aed->bits[i].label); cnt++;
    XtSetValues(aed->bits[i].toggles[0], args, cnt);

    cnt = 0;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNleftWidget, aed->bits[i].toggles[0]); cnt++;
    XtSetValues(aed->bits[i].toggles[1], args, cnt);

    cnt = 0;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNleftWidget, aed->bits[i].toggles[1]); cnt++;
    XtSetValues(aed->bits[i].toggles[2], args, cnt);
  }

  /* create next, prev save, revert and quit buttons */
  cnt = 0;
  XtSetArg(args[cnt], XmNleftOffset, 300); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNtopOffset, 5); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, aed->labels[ACCESS_LABEL]); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreate("save", XmSTRING_DEFAULT_CHARSET)); cnt++; 

  aed->saveButton = XtCreateManagedWidget("save", 
			xmPushButtonWidgetClass, aed->aed_form, args, cnt);
  XtAddCallback(aed->saveButton, XmNactivateCallback, saveAed, aed);

  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, aed->saveButton); cnt++;
  XtSetArg(args[cnt], XmNtopOffset, 5); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, aed->labels[ACCESS_LABEL]); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreate("revert", XmSTRING_DEFAULT_CHARSET)); cnt++; 

  aed->revertButton = XtCreateManagedWidget("revert", 
			xmPushButtonWidgetClass, aed->aed_form, args, cnt);
  XtAddCallback(aed->revertButton, XmNactivateCallback, revertAed, aed);

  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, aed->revertButton); cnt++;
  XtSetArg(args[cnt], XmNtopOffset, 5); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, aed->labels[ACCESS_LABEL]); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreate("quit", XmSTRING_DEFAULT_CHARSET)); cnt++; 

  aed->quitButton = XtCreateManagedWidget("quit", 
			xmPushButtonWidgetClass, aed->aed_form, args, cnt);

  XtAddCallback(aed->quitButton, XmNactivateCallback, 
		(exit_on_close) ? quitAed : closeAed, aed);

  cnt = 0;
  XtSetArg(args[cnt], XmNleftOffset, 300); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNtopOffset, 5); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, aed->saveButton); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreate("next", XmSTRING_DEFAULT_CHARSET)); cnt++; 

  aed->nextButton = XtCreateManagedWidget("next", 
			xmPushButtonWidgetClass, aed->aed_form, args, cnt);
  XtAddCallback(aed->nextButton, XmNactivateCallback, nextAed, aed);

  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, aed->nextButton); cnt++;
  XtSetArg(args[cnt], XmNtopOffset, 5); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, aed->saveButton); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreate("prev", XmSTRING_DEFAULT_CHARSET)); cnt++; 

  aed->prevButton = XtCreateManagedWidget("previous", 
			xmPushButtonWidgetClass, aed->aed_form, args, cnt);

  XtAddCallback(aed->prevButton, XmNactivateCallback, prevAed, aed);

  return(aed);
}

/* getFileType: ***********************************************************
 * retrieve the type for the given file by a call to '/bin/file'.
 *************************************************************************/

Boolean getFileType(fcont, file)
     char *fcont, *file;
{
  FILE *f;
  char buf[255], cmd[128];
  char *str;

  sprintf(cmd, "/bin/file %s", file);

  if (! (f = popen(cmd, "r"))) {
    sprintf(msg, "cannot get file type -- %s -- !", file);
    exError("aed", msg, NULL);
    *fcont = EOS;
    return(False);
  }

  if (! fgets(buf, sizeof(buf), f)) {
    sprintf(msg, "cannot get file type -- %s -- !", file);
    exError("aed", msg, NULL);
    *fcont = EOS;
    pclose(f);
    return(False);
 }

  pclose(f);

  for (str = index(buf, ':'), str++; *str == ' ' || *str == '\t'; str++);
  str[strlen(str) - 1] = '\0';

  strcpy(fcont, str);

  if (strcmp(fcont,"directory") == 0)
    strcpy(fcont,"file names");
  return(True);
}

/* aedInfo: ******************************************************************
 * get the info related to the given file and display the prepared information
 * in the aed window 
 ****************************************************************************/

aedInfo(aed, indx)
     aedForm *aed;
     int indx;			/* index of file */
{
  Boolean isdevice = False;
  char *name; 
  struct tm *localtime();
  struct tm *tmp;
  struct stat sb;
  char res[256];
  char ftype[80];
  char fcont[80]; 
  
  if (indx >= aed->file_cnt || indx < 0) {
    Beep(XtDisplay(aed->aed_shell));
    return;
  }

  aed->current = indx;

  free(aed->file);

  if (aed->file_list[indx][0] == '/') 
    aed->file = aed->file_list[indx];
  else {
    aed->file = Malloc(strlen(aed->dir) + strlen(aed->file_list[indx]) + 2);
    strcpy(aed->file, aed->dir);
    strcat(aed->file, "/");
    strcat(aed->file, aed->file_list[indx]);
  }

  aed_set(aed->labels[NAME], aed->file);

  if(stat(aed->file, &sb) == -1) {
    sprintf(msg, "cannot stat -- %s -- !", aed->file);
    exError("aed", msg, NULL);
    return;
  }
    
  sprintf(res,"size             : %ld  bytes", sb.st_size);
  aed_set(aed->labels[SIZE], res);

 if((pw = getpwuid(sb.st_uid)) == NULL)
    name = "????";
  else
    name = pw->pw_name;

  aed->orig_owner = strCopy(name);
  editorFillBuffer(aed->owner->entryEditor, name, False);
  
  if((gr = getgrgid(sb.st_gid)) == NULL)
    name = "????";
  else
    name = gr->gr_name;

  aed->orig_group = strCopy(name);
  aed->groupBuf = strCopy(name);
  aed_set(aed->group, name);

  sprintf(res,"links            : %d",sb.st_nlink);
  aed_set(aed->labels[LINKS], res);
  
  if((sb.st_mode &S_IFMT) == S_IFIFO)
    sprintf(res, "fifo special");
  else if((sb.st_mode &S_IFMT) == S_IFCHR) {
    sprintf(res, "character special file");
    isdevice = True;
  } else if((sb.st_mode &S_IFMT) == S_IFBLK) {
    sprintf(res, "block special file");
    isdevice = True;
  } else if((sb.st_mode &  S_IFMT) == S_IFREG) {
    if (! getFileType(fcont, aed->file))
      return;
    sprintf(res, "regular file (%s)", fcont);
  } else if ((sb.st_mode & S_IFMT) == S_IFLNK)
    sprintf(res, "symbolic link");
  else if((sb.st_mode &S_IFMT) == S_IFSOCK)
    sprintf(res, "socket");
  else if ((sb.st_mode & S_IFMT) == S_IFDIR)
    sprintf(res, "directory");

  strcpy(ftype, res);
  sprintf(res, "file type        : %s", ftype);
  
  aed_set(aed->labels[TYPE], res);

  tmp = localtime(&sb.st_mtime);
  sprintf(res,"last modification: %02d.%02d.%02d   %02d:%02d",
	  tmp->tm_mday, tmp->tm_mon, tmp->tm_year, tmp->tm_hour, tmp->tm_min);
  aed_set(aed->labels[LASTACCESS], res);
  
  tmp = localtime(&sb.st_atime);
  sprintf(res,"last access      : %02d.%02d.%02d   %02d:%02d",
	  tmp->tm_mday, tmp->tm_mon, tmp->tm_year, tmp->tm_hour, tmp->tm_min);
  aed_set(aed->labels[LASTMOD], res);

  tmp = localtime(&sb.st_ctime);
  sprintf(res,"last change      : %02d.%02d.%02d   %02d:%02d",
	  tmp->tm_mday, tmp->tm_mon, tmp->tm_year, tmp->tm_hour, tmp->tm_min);
  aed_set(aed->labels[LASTCHANGE], res);

  aed->orig_mode = sb.st_mode & 07777;		/* only mode-bits */
  aed->mode = aed->orig_mode;

  sprintf(res, "%#4.4o <--> ", aed->orig_mode);
  aed_set(aed->labels[ORIG_ACCESS], res);

  sprintf(res,"%#4.4o", aed->mode);
  aed_set(aed->labels[CURR_ACCESS], res);

  set_bits(aed, sb.st_mode);
  
  sprintf(res, "device           : %d, %d", (sb.st_dev>>8) & 0377,
	    						 sb.st_dev & 0377);
  aed_set(aed->labels[RESIDESON], res);

  if(isdevice == True) 
    sprintf(res, "device number    : %d, %d", (sb.st_rdev>>8) & 0377,
	    						 sb.st_rdev & 0377);
   else
    strcpy(res, "device number    : ?, ?");
  
  aed_set(aed->labels[DEVICENO], res);

  sprintf(res, "inode            : %d", sb.st_ino);
  aed_set(aed->labels[INODE], res);
}

/* initAed: ***************************************************************
 * initialize subsystems and variables 
 *************************************************************************/

initAed(display)
     Display *display;
{
  register i, cnt;
  register char **ptr;
  group_elem *groups_list;

  exInitLayout(display);
  exInitMenus(display);

  access_bits[0] = XmStringCreate("-", XmSTRING_DEFAULT_CHARSET);
  access_bits[1] = XmStringCreate("r", XmSTRING_DEFAULT_CHARSET);
  access_bits[2] = XmStringCreate("w", XmSTRING_DEFAULT_CHARSET);
  access_bits[3] = XmStringCreate("x", XmSTRING_DEFAULT_CHARSET);
  access_bits[4] = XmStringCreate("s", XmSTRING_DEFAULT_CHARSET);

  /* get group-file entries */
  
  cnt = 0;

  while (gr = getgrent()) {
    if (! (cnt % 10)) {
      if (cnt) 
	groups_list = (group_elem *) realloc(groups_list, 
				   ((cnt + 10) * sizeof(group_elem)));
      else
	groups_list = (group_elem *) calloc(10, sizeof(group_elem));
    }
    groups_list[cnt].name = strCopy(gr->gr_name);
    groups_list[cnt].gid  = gr->gr_gid;

    i = 0;
    while (gr->gr_mem[i]) {
      if (! (i % 10)) {
	if (i) 
	  ptr = (char **) realloc(ptr,((i + 10) * sizeof(char *)));
	else
	  ptr = (char **) calloc(10, sizeof(char *));
      }
      ptr[i] = strCopy(gr->gr_mem[i]);
      i++;
    }
    groups_list[cnt].mem_cnt = i;
    groups_list[cnt].mem = ptr;
    cnt++;
  }
  group_list.groups = groups_list;
  group_list.group_cnt = cnt;
}


/* aed_set: ***************************************************************
 * assign the specified label a new value 
 *************************************************************************/
aed_set(w, str)
     Widget w;
     char *str;
{
  Arg args[1];
  int cnt;
  XmString s;

  cnt  = 0;
  XtSetArg(args[cnt], XmNlabelString, &s); cnt++;
  XtGetValues(w, args, cnt);

  if (s)
    XmStringFree(s);

  cnt = 0;
  XtSetArg(args[cnt], XmNlabelString, 
	    XmStringCreate(str, XmSTRING_DEFAULT_CHARSET)); cnt++; 
  XtSetValues(w, args, cnt);
}

/* aed_set_bits: **********************************************************
 * set the toggle-buttons 
 *************************************************************************/

aed_set_bits(r, b, m)
     exRights *r;
     Boolean b[];
     int m[];
{
  register i, cnt;
  Arg args[1];

  for (i = 0; i < 3; i++) {
    cnt = 0;
    XtSetArg(args[cnt], XmNlabelString, access_bits[b[i] ? i + 1 : 0]); cnt++;
    XtSetValues(r->toggles[i], args, cnt);
    r->bit[i] = m[i];
  }
}
  
/* aed_set_sbits: **********************************************************
 * set the s-bit toggle buttons (suid, sgid, sticky bits)
 *************************************************************************/

aed_set_sbits(r, b, m)
     exRights *r;
     Boolean b[];
     int m[];
{
  register i, cnt;
  Arg args[1];

  for (i = 0; i < 3; i++) {
    cnt = 0;
    XtSetArg(args[cnt], XmNlabelString, access_bits[b[i] ? 4 : 0]); cnt++;
    XtSetValues(r->toggles[i], args, cnt);
    r->bit[i] = m[i];
  }
}
  
/**************************************************************************
 *
 *				callbacks
 *
 *************************************************************************/

/* toggleAccessBit: *******************************************************
 * called when the user presses the mouse button over a mode-bit. The 
 * corresponding bit is toggled, set if reset and reset if set.
 *************************************************************************/

static void toggleAccessBit(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  exRights *bits;
  aedForm *aed;
  Arg args[1];
  int cnt, i, j, mode;
  int no = (int) client_data;
  char res[6];
  
  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &bits); cnt++;
  XtGetValues(w, args, cnt);
  
  aed = bits->aed;

  if (bits->indx < 3) {			/* owner or group or other */
    bits->bit[no] = (bits->bit[no]) ? 0 : 
      			(power(8, 2 - bits->indx) * power(2, 2 - no));
    cnt = 0;
    XtSetArg(args[cnt], 
	     XmNlabelString, access_bits[bits->bit[no] ? no + 1 : 0]); cnt++;
  } else {				/* s bits */
    bits->bit[no] = bits->bit[no] ? 0 : power(8, 3) * power(2, 2 - no);
    cnt = 0;
    XtSetArg(args[cnt], 
	     XmNlabelString, access_bits[bits->bit[no] ? 4 : 0]); cnt++;
  }

  XtSetValues(bits->toggles[no], args, cnt);
  
  mode = 0;
  for (i = 0; i < 4; i++) 
    for (j = 0; j < 3; j++) 
      mode += aed->bits[i].bit[j];
  
  sprintf(res,"%#4.4o", mode);
  aed_set(aed->labels[CURR_ACCESS], res);
  aed->mode = mode;
}

/* saveAed: ***************************************************************
 * save the edited settings
 *************************************************************************/

static void saveAed(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  char res[40];
  Boolean ctime_changed = False;
  aedForm *aed = (aedForm *) client_data;
  struct stat sb;
  struct tm *tmp;

  if (! needsUpdate(aed)) {
    Beep(XtDisplay(aed->aed_shell));
    return;
  }

  if (exNotify2User(aed->aed_form, 100, 100, "Do changes ?") == ANSWERYES) {
      if (chmod(aed->file, aed->mode) != 0) {
	sprintf(msg, "cannot change mode -- %s --!", aed->file);
	exError("aed", msg, NULL);
	return;
      } else {
	aed->orig_mode = aed->mode;
	sprintf(res, "%#4.4o <--> ", aed->orig_mode);
	aed_set(aed->labels[ORIG_ACCESS], res);
	ctime_changed = True;
      }

      pw = getpwnam(aed->ownerBuf);
      gr = getgrnam(aed->groupBuf);

      if (!pw || !gr) {
	sprintf(msg, "cannot retrieve group or owner -- %s -- !", aed->file);
	exError("aed", msg, NULL);
	return;
      }

      if (strcmp(aed->ownerBuf, aed->orig_owner) 
	  || strcmp(aed->groupBuf, aed->orig_group)) {
        if (chown(aed->file, pw->pw_uid, gr->gr_gid) != 0) {
	  sprintf(msg, "cannot change owner -- %s --!", aed->file);
	  exError("aed", msg, NULL);
	} else {
	  free(aed->orig_owner);
	  free(aed->orig_group);
	  aed->orig_owner = strCopy(aed->ownerBuf);
	  aed->orig_group = strCopy(aed->groupBuf);
	  ctime_changed = True;
	}
      }

      if (ctime_changed) {
	if(stat(aed->file, &sb) == -1) {
	  sprintf(msg, "cannot stat -- %s -- !", aed->file);
	  exError("aed", msg, NULL);
	  return;
	}
	tmp = localtime(&sb.st_ctime);
	sprintf(res,"last change      : %02d.%02d.%02d   %02d:%02d",
	   tmp->tm_mday, tmp->tm_mon, tmp->tm_year, tmp->tm_hour, tmp->tm_min);
	aed_set(aed->labels[LASTCHANGE], res);

      }
    }
}

/* revertAed: *************************************************************
 * revert the edited settings
 *************************************************************************/

static void revertAed(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  aedForm *aed = (aedForm *) client_data;
  char res[40];

  if (! needsUpdate(aed)) {
    Beep(XtDisplay(aed->aed_shell));
    return;
  }

  if (exNotify2User(aed->aed_form, 100, 100, "Undo changes ?") == ANSWERYES) {
    aed->mode = aed->orig_mode;
    sprintf(res,"%#4.4o", aed->mode);
    aed_set(aed->labels[CURR_ACCESS], res);
    set_bits(aed, aed->mode);

    free(aed->groupBuf);
    aed->groupBuf = strCopy(aed->orig_group);
    aed_set(aed->labels[GROUP_LABEL], aed->groupBuf);
    editorFillBuffer(aed->owner->entryEditor, aed->orig_owner, False);
  }
}

/* quitAed: ***************************************************************
 * quit aed
 *************************************************************************/

static void quitAed(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  closeAed(w, client_data, call_data);
  exit(1);
}

/* closeAed: **************************************************************
 * close aed
 *************************************************************************/

static void closeAed(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  aedForm *aed = (aedForm *) client_data;

  if (! needsUpdate(aed)) {
    XtPopdown(aed->aed_shell);
    aed->flag &= ~AED_OPEN;
    free(aed->file_list);
    free(aed->file);
    free(aed->dir);
    aed->file = NULL;
  } else 
    if (exNotify2User(aed->aed_shell, 100, 100, "Quit without doing changes ?") 
	== ANSWERYES) {
      XtPopdown(aed->aed_shell);
      aed->flag &= ~AED_OPEN;
      free(aed->file_list);
      free(aed->file);
      free(aed->dir);
      aed->file = NULL;
    }
}

/* nextAed: ***************************************************************
 * display next file in file-list
 *************************************************************************/

static void nextAed(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  aedForm *aed = (aedForm *) client_data;
  
  aedInfo(aed, aed->current + 1);
}


/* prevAed: ***************************************************************
 * display previous file in file-list
 *************************************************************************/

static void prevAed(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  aedForm *aed = (aedForm *) client_data;

  aedInfo(aed, aed->current - 1);
}

/* setGroup: **************************************************************
 * let the user specify a new group
 *************************************************************************/

static void setGroup(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  MHmenu *mptr;
  exMenu *menu;
  int groupCnt = 0;
  register i, j;
  char **groups, *str;
  group_elem *groupPtr;

  aedForm *aed = (aedForm *) client_data;

  groups = (char **) calloc(group_list.group_cnt, sizeof(char *));

  for (i = 0; i < group_list.group_cnt; i++) {
    groupPtr = &group_list.groups[i];
    for (j = 0; j < groupPtr->mem_cnt; j++) 
      if (! strcmp(groupPtr->mem[j], aed->ownerBuf)) {
	groups[groupCnt++] = groupPtr->name;
	break;
      }
  }

  if (groupCnt < 2)
    return;

  setWorkApplication();
  if (! (mptr = buildMenuFromStrings("valid groups", MDS_L_POPUP, 
			           groups, groups, groupCnt)))
    return;

  setSelCnt(mptr, 1);
  setSelType(mptr, MDS_CNT_EQ);
  setMenuPos(mptr, 5, 5, MDS_POS_CREL);

  if (! (menu = exMakeMenu((ApplicationNode *)NULL, (exMenubar *)NULL,
			   aed->aed_shell, mptr)))
    return;

  exCallMenu(menu);
  if (str = exGetMenuReturnValue(menu)) {
    str[strlen(str) - 1] = '\0';
    aed_set(aed->group, str);
    free(aed->groupBuf);
    aed->groupBuf = strCopy(str);
  }

  /*deleteMenu(currentApplication(), menu);*/
  deleteWorkMenu(mptr);
  cfree(groups);
}

/* openAed: ***************************************************************
 * map the aed-form
 *************************************************************************/

openAed(aed, dir, file_list, cnt)
     aedForm *aed;
     char *dir;
     char **file_list;
     int cnt;
{
  if (! (aed->flag & AED_OPEN)) {
    aed->flag |= AED_OPEN;
    XtPopup(aed->aed_shell, XtGrabNone);
  }
  aed->file_list = file_list;
  aed->dir = strCopy(dir);
  aed->file_cnt = cnt;
  aed->file = NULL;
}

static entryHelp(editor)
     Editor editor;
{
  help(editor);
}

static bell(editor)
     Editor editor;
{
  Bell();
}


Boolean needsUpdate(aed)
     aedForm *aed;
{
  if (! aed->orig_owner || ! aed->orig_group)
    return(False);

  /* compare initial with current values */
  if (aed->mode == aed->orig_mode 			&&
       strcmp(aed->ownerBuf, aed->orig_owner) == 0	&&
       strcmp(aed->groupBuf, aed->orig_group) == 0)
    return(False); 
  else
    return(True);
}

set_bits(aed, mode)
     aedForm *aed;
     int mode;
{
  Boolean b[3];
  int m[3];

  if (mode & 0400) {
    b[0] = True;
    m[0] = 0400;
  } else {
    b[0] = False;
    m[0] = 0;
  }

  if (mode & 0200) {
    b[1] = True;
    m[1] = 0200;
  } else {
    b[1] = False;
    m[1] = 0;
  }

  if (mode & 0100) {
    b[2] = True;
    m[2] = 0100;
  } else {
    b[2] = False;
    m[2] = 0;
  }

  aed_set_bits(&aed->bits[OWNER], b, m);
  
  if (mode & 040) {
    b[0] = True;
    m[0] = 040;
  } else {
    b[0] = False;
    m[0] = 0;
  }

  if (mode & 020) {
    b[1] = True;
    m[1] = 020;
  } else {
    b[1] = False;
    m[1] = 0;
  }

  if (mode & 010) {
    b[2] = True;
    m[2] = 010;
  } else {
    b[2] = False;
    m[2] = 0;
  }

  aed_set_bits(&aed->bits[GROUP], b, m);

  if (mode & 04) {
    b[0] = True;
    m[0] = 04;
  } else {
    b[0] = False;
    m[0] = 0;
  }

  if (mode & 02) {
    b[1] = True;
    m[1] = 02;
  } else {
    b[1] = False;
    m[1] = 0;
  }

  if (mode & 01) {
    b[2] = True;
    m[2] = 01;
  } else {
    b[2] = False;
    m[2] = 0;
  }

  aed_set_bits(&aed->bits[OTHER], b, m);

  if (mode & S_ISUID) {
    b[0] = True;
    m[0] = S_ISUID;
  } else {
    b[0] = False;
    m[0] = 0;
  }
  if (mode & S_ISGID) {
    b[1] = True;
    m[1] = S_ISGID;
  } else {
    b[1] = False;
    m[1] = 0;
  }
  if (mode & S_ISVTX) {
    b[2] = True;
    m[2] = S_ISVTX;
  } else {
    b[2] = False;
    m[2] = 0;
  }

  aed_set_sbits(&aed->bits[SBITS], b, m);
}
