/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include "fm.h"

static char SccsID[] = "@(#)file.c	1.3\t1/27/93";            /* sccs keyword */

static char msg[255];

Boolean insertNewFile();
char *makeItemRepresentation();

/* deleteFiles: **************************************************************
 * mark the file as deleted; the file is removed after user confirmation when
 * the window is removed, when file manager is exited or when the user chooses
 * 'commit' in the menu
 ****************************************************************************/

void deleteFiles(wnd)
     dirWindow *wnd;
{
  register dirItem *iptr;

  for (iptr = wnd->itemList; iptr; iptr = iptr->next) {
    SET_ITEM_STATE(wnd, iptr);
    if (iptr->flags & ITEM_SELECTED) {
      if (iptr->flags & ITEM_DISPLAYED) {
	sprintf(msg, "Cannot delete a displayed directory: %s",
							ITEM_NAME(iptr));
	exAlert(Head->root, 100, 100, msg);
	continue;
      }
      iptr->flags |= ITEM_DELETED;
      fmdisplayItem(iptr->dirWnd, iptr, True);
      if (! (wnd->flags & DIRW_CHNGED))	/* mark directory as changed */
	wnd->flags |= DIRW_CHNGED;
    }
  }
}

/* undeleteFiles: *************************************************************
 * reset the delete flag
 *****************************************************************************/

void undeleteFiles(wnd)
     dirWindow *wnd;
{
  register dirItem *iptr;

  for (iptr = wnd->itemList; iptr; iptr = iptr->next) {
    SET_ITEM_STATE(wnd, iptr);
    if (iptr->flags & ITEM_SELECTED) {
      iptr->flags &= ~ITEM_DELETED;
      fmdisplayItem(iptr->dirWnd, iptr, True);
    }
  }
}

/* renameFiles: ***************************************************************
 * mark file as renamed; do real renaming after user confirmation when
 * the window is removed, when file manager is exited or when the user chooses
 * 'commit' in the menu
 *****************************************************************************/

void renameFiles(wnd)
     dirWindow *wnd;
{
  char *newName;
  register dirItem *iptr;
  int x, y;
  
  getMouse_xy(XtDisplay(Head->root), &x, &y);
  for (iptr = wnd->itemList; iptr; iptr = iptr->next) { 
    SET_ITEM_STATE(wnd, iptr);
    if (iptr->flags & ITEM_SELECTED) {
      sprintf(msg,"(%s) Enter new name: ",ITEM_NAME(iptr));
      newName = exGetFilename(Head->root, x, y, msg, NULL, ITEMLEN);
      if (*newName != 10 && *newName != 13 && *newName != '\0') { 
	/* first look if corresponding dir-windows exist, change accordingly */
	if (iptr->flags & ITEM_ISDIR && iptr->flags & ITEM_DISPLAYED)
	  renameDirWindows(wnd, iptr, newName);
	iptr->flags |= ITEM_RENAMED;
	free(iptr->a_name);
	iptr->a_name = newName;
	fmdisplayItem(iptr->dirWnd, iptr, True);
	if (! (wnd->flags & DIRW_CHNGED))	/* mark directory as changed */
	  wnd->flags |= DIRW_CHNGED;
      }
    }
  }
}

/* unrenameFiles: ************************************************************
 * restore the original name of the file and reset the rename flag
 ****************************************************************************/
void unrenameFiles(wnd)
     dirWindow *wnd;
{
  register dirItem *iptr;
  
  for (iptr = wnd->itemList; iptr; iptr = iptr->next) { 
    SET_ITEM_STATE(wnd, iptr);
    if (iptr->flags & (ITEM_SELECTED | ITEM_RENAMED)) {
      if (iptr->flags & ITEM_ISDIR && iptr->flags & ITEM_DISPLAYED)
	renameDirWindows(wnd, iptr, iptr->D_name);
      iptr->flags &= ~ITEM_RENAMED;
      free(iptr->a_name);
      iptr->a_name = (char *) NULL;
      fmdisplayItem(iptr->dirWnd, iptr, True);
    }
  }
}

/* copyFiles: *****************************************************************
 * copy the files into an other directory; the user must select the 
 * destination directory with the mouse; if move is True, the files are 
 * moved (i.e. deleted in the destination directory)
 *****************************************************************************/

void copyFiles(wnd, move)
     dirWindow *wnd;
     Boolean move;
{
  register dirWindow *w;
  XEvent event;
  Arg args[2];
  int x, y, cnt;
  Dimension width, height;
  Window subw, rw;
  Display *d = XtDisplay(Head->root);
  
  XtGrabPointer(wnd->fmWindow, False, ButtonPressMask, 
		GrabModeAsync, GrabModeAsync, None, CopyCursor, CurrentTime);

  XMaskEvent(d, ButtonPressMask, &event);
  rw = XDefaultRootWindow(d);
  for (w = Head->first; w; w = w->next) {
    if (w == wnd)
      continue;

    cnt = 0;
    XtSetArg(args[cnt], XmNwidth, &width); cnt++;
    XtSetArg(args[cnt], XmNheight, &height); cnt++;
    XtGetValues(w->fmTop, args, cnt);			/* get bounds */

    XTranslateCoordinates(d, rw, XtWindow(w->fmWindow), 
	  event.xbutton.x_root, event.xbutton.y_root, &x, &y, &subw);

    if (x > 0 && y > 0 && x < width && y < height) {	/* within bounds */
      XtUngrabPointer(wnd->fmWindow, event.xbutton.time);
      insertFiles(wnd, w, move);
      if (! (w->flags & DIRW_CHNGED))	
	w->flags |= DIRW_CHNGED;
      if (move && (!(wnd->flags & DIRW_CHNGED)))	
	wnd->flags |= DIRW_CHNGED;
      break;
    }
  }
}

/* duplicateFiles: ************************************************************
 * duplicate the files in the current directory; the user is prompted to 
 * provide a new name for each file
 ****************************************************************************/

void duplicateFiles(wnd)
     dirWindow *wnd;
{
  register dirItem *iptr;
  int x, y;
  char *newName;

  getMouse_xy(XtDisplay(Head->root), &x, &y);

  for (iptr = wnd->itemList; iptr; iptr = iptr->next) {
    SET_ITEM_STATE(wnd, iptr);
    if (iptr->flags & (ITEM_SELECTED)) {
      sprintf(msg,"(%s) Enter new name: ",ITEM_NAME(iptr));
      newName = exGetFilename(Head->root, x, y, msg, NULL, ITEMLEN);
      if (*newName != 10 && *newName != 13 && *newName != '\0') {
	insertNewFile(wnd, iptr, wnd, newName, False);
	free(newName);
      }
    }
  }
  updateDirWindow(wnd);
  if (! (wnd->flags & DIRW_CHNGED))	/* mark directory as changed */
    wnd->flags |= DIRW_CHNGED;
}

/* commitChanges: *************************************************************
 * do the changes for the current directory; if parameter silent is True, the 
 * user is not informed if nothing remains to be done. If update is False, the 
 * directory window is not updated.
 *****************************************************************************/

void commitChanges(wnd, silent, update)
     dirWindow *wnd;
     Boolean silent;
     Boolean update;
{
  register dirItem *iptr;
  char path1[256], path2[256];
  int rc;

  if (wnd->flags & DIRW_CHNGED) {
    sprintf(msg,"Directory: %s\nReally do changes (cannot be undone) ?",
	    wnd->path);
    
    switch (exNotify2User(Head->root, 100, 100, msg)) {
    case ANSWERNO:
      return;
    case ANSWERYES:
      break;
    case ANSWERCANCEL:
      return;
    }
  } else if (! silent) {
    sprintf(msg, "Directory: %s\nNo changes remain to be done !", wnd->path);
    exAlert(Head->root, 100, 100, msg);
    return;
  }

  for (iptr = wnd->itemList; iptr; iptr = iptr->next) {
    if (iptr->flags & ITEM_DELETED) {
      sprintf(path1, "%s/%s", wnd->u_path, iptr->D_name);
      if (iptr->flags & ITEM_ISDIR) 
	rc = rmdir(path1);
      else
	rc = unlink(path1);

      if (rc < 0) {
	sprintf(msg, "delete file: -- %s -- cannot be deleted !", wnd->u_path);
	exAlert(Head->root, 100, 100, msg);
      }
    } else if (iptr->flags & ITEM_RENAMED) {
      sprintf(path1, "%s/%s", wnd->u_path, iptr->D_name);
      sprintf(path2, "%s/%s", wnd->u_path, iptr->a_name);
      rc = rename(path1, path2);
    }
    iptr->flags &= ~ITEM_SELECTED;
  }

  if (update)
    updateDirWindow(wnd);			/* show changes */
  wnd->flags &= ~DIRW_CHNGED;			/* reset state */
}

/* writeNames: ****************************************************************
 * write the names of the selected files to stdout; the parameter scope 
 * defines whether all directories or only the current directory are
 * searched for selected items 
 *****************************************************************************/

void writeNames(wnd, scope)
     dirWindow *wnd;
     int scope;
{
  register dirWindow *w;

  switch (scope) {
  case ALL_WINDOWS:
    for (w = Head->first; w; w = wnd->next)
      putNames(w);
    break;
  case CURRENT_WINDOW:
    putNames(wnd);
    break;
  }
}

/* putNames: ******************************************************************
 * write the names of the files associated with the selected items of a 
 * directory to stdout
 *****************************************************************************/

putNames(wnd)
     dirWindow *wnd;
{ 
  register dirItem *iptr;
  register char *ptr = wnd->u_path;
  
  for (iptr = wnd->itemList; iptr; iptr = iptr->next) {
    SET_ITEM_STATE(wnd, iptr);
    if (iptr->flags & (ITEM_SELECTED)) {
      write(1, ptr, strlen(ptr));			/* write dir name */
      if (ptr[strlen(ptr) - 1] != '/')
	write(1, "/", 1);
      write(1, iptr->D_name, strlen(iptr->D_name));	/* write file name */
      write(1, " ", 1);
    }
  }
}

/* insertNewFile: *************************************************************
 * copy (dflag = false) or move (dflag = True) the file corresponding to the
 * item from the directory associated with the window 'sw' into the directory
 * associated with the window 'dw';
 *****************************************************************************/

Boolean insertNewFile(sw, item, dw, name, dflag)
     dirWindow *sw;			/* source window */
     dirItem *item;			/* item to be copied */
     dirWindow *dw;			/* destination window */
     char *name;			/* new name */
     Boolean dflag;
{
  register dirItem *iptr;
  Boolean file_exists = False;
  Boolean copy_flag = True;

  if (sw == dw && strcmp(ITEM_NAME(item), name) == 0)
    return(False);

  for (iptr = dw->itemList; iptr->next; iptr = iptr->next) {
    SET_ITEM_STATE(dw, iptr);
    if (strcmp(name, ITEM_NAME(iptr)) == 0) {
      file_exists = True;
      break;
    }
  }

  if (file_exists) {
    sprintf(msg, "File -- %s -- already exists: overwrite ?", name);
    switch (exNotify2User(Head->root, 100, 100, msg)) {
    case ANSWERYES:
      copy_flag = True;
      break;
    case ANSWERCANCEL:
      return(False);
    default:
      copy_flag = False;
      break;
    } 
  }
    
  if (copy_flag)
    sys_copy(sw->u_path, item->D_name, dw->u_path, name, dflag);

  return(copy_flag);
}

/* insertFiles: ***************************************************************
 * insert each selected file into the destination window; copy if dflag is 
 * False, move if dflag is True
 *****************************************************************************/

insertFiles(sw, dw, dflag)
     dirWindow *sw;
     dirWindow *dw;
     Boolean dflag;
{
  register dirItem *iptr;

  for (iptr = sw->itemList; iptr; iptr = iptr->next) {
    SET_ITEM_STATE(sw, iptr);
    if (iptr->flags & (ITEM_SELECTED)) {
      insertNewFile(sw, iptr, dw, ITEM_NAME(iptr), dflag);
      fmdisplayItem(iptr->dirWnd, iptr, True);
    }
  }

  if (dflag)					/* files have been moved */
    updateDirWindow(sw);			/* update source window */
  updateDirWindow(dw);				/* update destination window */
}

/* markVisibleItems: **********************************************************
 * mark in the list of all items those items which names satisfy the given 
 * file pattern; additionally set the used file pattern in the window structure
 * to the used pattern string 
 *****************************************************************************/

markVisibleItems(wnd, pattern, forced)
     dirWindow *wnd;
     char *pattern;
{
  char *re_comp();
  char shellRexp[25];			/* holds the converted reg. exp. */
  register dirItem *iptr;

  if (! pattern)
    return(setError(ERR_NULL));

  convertRegExp(pattern, shellRexp);	/* convert the reg.exp. syntax */
  if (re_comp(shellRexp)) {		/* compile the regular expression */
    return(setError(ERR_ABORT));
  }

  for (iptr = wnd->itemList; iptr; iptr = iptr->next)
    if (re_exec(ITEM_NAME(iptr))) 	/* do the match */
      iptr->flags |= ITEM_VISIBLE;
    else
      iptr->flags &= ~ITEM_VISIBLE;

  setFilePattern(wnd, pattern);

  if (wnd->m) {
    if (wnd->m->scrollbar)
      XtDestroyWidget(wnd->m->scrollbar);
    freeMenu(wnd->m);
  }
  makeItems(wnd);
  fmdisplayItems(wnd, forced);
  return(0);
}

/* selectItems: *************************************************************
 * perform selections on the items of the current directory window: either
 * select or unselect all files or selecet those files that contain a certain
 * pattern.
 ****************************************************************************/

void selectItems(wnd, what)
     dirWindow *wnd;
     int what;
{
  register dirItem *iptr;
  char *pattern, *files;
  char buf[100];
  FILE *sysf, *popen();

  switch (what) {
  case SELECT_ALL:
    for (iptr = wnd->itemList; iptr; iptr = iptr->next) {
      SET_ITEM_STATE(wnd, iptr);
      if (iptr->flags & ITEM_VISIBLE && ! (iptr->flags & ITEM_SELECTED))
	markItemSelected(iptr, True);
    }
    break;
  case UNSELECT_ALL:
    for (iptr = wnd->itemList; iptr; iptr = iptr->next) {
      SET_ITEM_STATE(wnd, iptr);
      if (iptr->flags & ITEM_VISIBLE && iptr->flags & ITEM_SELECTED)
	markItemSelected(iptr, False);
    }
    break;
  case SELECT_BY_GREP:
    pattern = exGetEntry(Head->root, -1, -1, "search pattern: ", NULL, 20);
    if (pattern && *pattern) {
      files = exGetEntry(Head->root, -1, -1, 
			 "files (<cr> for all): ", NULL, 20);
      sprintf(buf, "cd %s; %s -l '%s' %s", 
	 wnd->u_path, FGREP, pattern, (files && *files) ? files : "*");
      selectItems(wnd, UNSELECT_ALL);
      sysf = popen(buf, "r");
      while (fgets(buf, sizeof(buf), sysf)) {
	if (buf[strlen(buf) - 1] == '\n')
	  buf[strlen(buf) - 1] = EOS;
	for (iptr = wnd->itemList; iptr; iptr = iptr->next) {
	  if (strcmp(iptr->D_name, buf) == 0 && iptr->flags & ITEM_VISIBLE)
	    markItemSelected(iptr, True);
	}
      }
      pclose(sysf);
    }
  }
}

/* setOrder: *****************************************************************
 * sets the sort criteria that is determines how the items are sorted; the 
 * sort criteria is one of: by last access data, by name, by modofication date
 * by size (see also sortItemList()).
 ****************************************************************************/

void setOrder(wnd, what)
     dirWindow *wnd;
     int what;
{
  if (wnd->sortFlag == what) 
    return;

  wnd->sortFlag = what;
  sortItemList(wnd);
  freeMenu(wnd->m);
  makeItems(wnd);
  fmdisplayItems(wnd, False);
}

/* setDsplOrder: *************************************************************
 * set the flags that determines how the items are displayed, either 
 * horizontally or vertically (see also makeItems()
 ****************************************************************************/

void setDsplOrder(wnd, what)
     dirWindow *wnd;
     int what;
{
  if (wnd->flags & what)
    return;

  if (what == DIRW_VERT) {
    wnd->flags &= ~DIRW_HOR;
    wnd->flags |= DIRW_VERT;
  } else {
    wnd->flags &= ~DIRW_VERT;
    wnd->flags |= DIRW_HOR;
  }
  makeItems(wnd);
  fmdisplayItems(wnd, False);
}

/* markItemSelected: *********************************************************
 * mark an item as beeing selected 
 ****************************************************************************/

markItemSelected(iptr, onoff)
     dirItem *iptr;
     int onoff;
{
  if (! (iptr->flags & ITEM_VISIBLE))		/* not a visible item */
    return;

  switch (onoff) {
  case True:					/* select item */	
    iptr->flags |= ITEM_SELECTED;
    iptr->dirWnd->m->selections[iptr->slot] = 1;
    fmdisplayItem(iptr->dirWnd, iptr, True);
    break;
  default:					/* deselect item */
    iptr->flags &= ~ITEM_SELECTED;
    iptr->dirWnd->m->selections[iptr->slot] = 0;
    fmdisplayItem(iptr->dirWnd, iptr, True);
    break;
  }
}

/* fmdisplayItems: ************************************************************
 * display visible items according to their associated flags 
 *****************************************************************************/

fmdisplayItems(wnd, forced)
     dirWindow *wnd;
     Boolean forced;
{
  register dirItem *iptr;

  for (iptr = wnd->itemList; iptr; iptr = iptr->next) 
    fmdisplayItem(wnd, iptr, forced);
}

/* fmdisplayItem: *************************************************************
 * display item according to associated flags 
 *****************************************************************************/

fmdisplayItem(wnd, iptr, forced)
     dirWindow *wnd;
     register dirItem *iptr;
     Boolean forced;
{
  char *name;
  
  name = makeItemRepresentation(iptr);

  setItemName(wnd->m, iptr->slot, name);
  if (forced)
    displayItem(wnd->m, iptr->slot, iptr->flags & ITEM_SELECTED);
}

