/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/

/*****************************************************************************
 * This file contains application functions invoked through the menus of the 
 * file-system browser; refer to file "makeMenus.c" for the description of 
 * the menus. 
 ****************************************************************************/

#include <X11/Xlib.h>
#include <Xm/Xm.h>

#include "fm.h"

static char SccsID[] = "@(#)menuFncts.c	1.4\t1/27/93";            /* sccs keyword */

extern dirWindow *toRemove;

/* remDirWnd ****************************************************************
 * remove the current window: ask whether to do the changes
 ****************************************************************************/

/* ARGSUSED */
remDirWnd(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    commitChanges(wnd, True, False);

  checkParentWindow(wnd->u_path, False);
  updateTree(wnd->u_path, False);

  /* don't remove the window here, because the current function has been 
     invoked by the menu system. If the window is removed here, Xt (or Motif) 
     has trouble to adjust its internal table (grab-list): it would bomb-out.
     So we set up to remove the window in the when we are back in the main 
     event loop after the menu has been popped down.
  */
  toRemove = wnd;
}

/* cleanupDirWindows *******************************************************
 * display all windows aligned with the current one
 ****************************************************************************/

/* ARGSUSED */  
cleanupDirWindows(parms)
     RequestParm *parms;
{
  register dirWindow *dirw;
  Position x, y;
  Dimension h;
  Arg args[3];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (! wnd)
    return;

  cnt = 0;
  XtSetArg(args[cnt], XmNx, &x); cnt++;
  XtSetArg(args[cnt], XmNy, &y); cnt++;
  XtSetArg(args[cnt], XmNheight, &h); cnt++;
  XtGetValues(wnd->fmTop, args, cnt);

  for (dirw = Head->first; dirw; dirw = dirw->next) {
    placeWindow(dirw, x, y);
    cnt = 0;
    XtSetArg(args[cnt], XmNheight, &h); cnt++;
    XtGetValues(dirw->fmTop, args, cnt);
    y = y + h + 40;		/* estimate wm decorations */
  }

  Head->curx = x;
  Head->cury = y;
}

/* listDirSel: ****************************************************************
 * for each selected item: if its a directory display it in a separate window
 *****************************************************************************/
/* ARGSUSED */
listDirSel(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    listDirectory(wnd, DIR_SELECTED);
}

/* listDirFather: *************************************************************
 * display the father directory
 ****************************************************************************/

/* ARGSUSED */
listDirFather(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    listDirectory(wnd, DIR_FATHER);
}

/* listDirAny: ***************************************************************
 * prompt the user for the name of a directory and display it 
 ****************************************************************************/

/* ARGSUSED */
listDirAny(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    listDirectory(wnd, DIR_ANY);
}

/* editAttr: ****************************************************************
 * interactively edit the file attributes of the selected files 
 ****************************************************************************/

/* ARGSUSED */
editAttr(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    editDirectory(wnd);
}

/* buildTree: ****************************************************************
 * display a graphical representation of the subtree below the current window
 ****************************************************************************/

/* ARGSUSED */
buildTree(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    buildFileTree(wnd);
}

/* remTree: ****************************************************************
 * remove the graphical structure
 ****************************************************************************/

/* ARGSUSED */
remTree(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);

  if (wnd)
    deleteTreeWindow(wnd);
}

/* delFile: ****************************************************************
 * for each selected item: mark the item as deleted
 ****************************************************************************/

/* ARGSUSED */
delFile(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);

  if (wnd)
    deleteFiles(wnd);
}

/* undelFile: *****************************************************************
 * undo the effect of a delete: remove the flag
 *****************************************************************************/

/* ARGSUSED */
undelFile(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    undeleteFiles(wnd);
}

/* renFile: ******************************************************************
 * foreach selected item: provide an alternative name
 *****************************************************************************/

/* ARGSUSED */
renFile(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    renameFiles(wnd);
}

/* unrenFile: ****************************************************************
 * undo the effect of the latest rename
 ****************************************************************************/

/* ARGSUSED */
unrenFile(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    unrenameFiles(wnd);
}

/* mvFile: *******************************************************************
 * let the user point to a window and move all files to the corresponding 
 * directory 
 *****************************************************************************/

/* ARGSUSED */
mvFile(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    copyFiles(wnd, True);
}

/* cpFile: *******************************************************************
 * let the user point to a window and copy all files to the corresponding 
 * directory 
 *****************************************************************************/

/* ARGSUSED */
cpFile(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    copyFiles(wnd, False);
}

/* dupFile: ******************************************************************
 * copy the files within the current directory: provide new names
 *****************************************************************************/

/* ARGSUSED */
dupFile(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    duplicateFiles(wnd);
}

/* putCurrentNames: ***********************************************************
 * provide the names of the selected items in the current window
 *****************************************************************************/

/* ARGSUSED */
putCurrentNames(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    writeNames(wnd, CURRENT_WINDOW);
}

/* putAllNames: ***************************************************************
 * provide the names of all selected items in all windows
 *****************************************************************************/

/* ARGSUSED */
putAllNames(parms)
     RequestParm *parms;
{
  writeNames((dirWindow *)NULL, ALL_WINDOWS);
}

/* commit: *******************************************************************
 * commit the changes in the current window; they cannot be undone thereafter
 *****************************************************************************/

/* ARGSUSED */
commit(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    commitChanges(wnd, True, True);
}

/* selAll: ********************************************************************
 * select all items in the current window
 *****************************************************************************/

/* ARGSUSED */
selAll(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    selectItems(wnd, SELECT_ALL);
}

/* unselAll: ******************************************************************
 * unselect all items in the current window
 *****************************************************************************/

/* ARGSUSED */
unselAll(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    selectItems(wnd, UNSELECT_ALL);
}

/* selByGrep: ****************************************************************
 * perform a grep on the files in the directory of the current window and 
 * select the items for the files that hit
 ****************************************************************************/

/* ARGSUSED */
selByGrep(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    selectItems(wnd, SELECT_BY_GREP);
}

/* sortByName: ****************************************************************
 * sort the items by name
 *****************************************************************************/

/* ARGSUSED */
sortByName(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    setOrder(wnd, ORDER_BYNAME);
}

/* sortByAccess: **************************************************************
 * sort the items by last access 
 *****************************************************************************/

/* ARGSUSED */
sortByAccess(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    setOrder(wnd, ORDER_BYACCESS);
}

/* sortByMod: ****************************************************************
 * sort items by modification date
 ****************************************************************************/

/* ARGSUSED */
sortByMod(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    setOrder(wnd, ORDER_BYMOD);
}

/* sortBySize: ****************************************************************
 * sort items by size
 *****************************************************************************/

/* ARGSUSED */
sortBySize(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    setOrder(wnd, ORDER_BYSIZE);
}

/* sortHorizontal: ************************************************************
 * display items in horizontal order
 *****************************************************************************/

/* ARGSUSED */
sortHorizontal(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    setDsplOrder(wnd, DIRW_HOR);
}

/* sortVertical **************************************************************
 * display items in vertical order
 ****************************************************************************/

/* ARGSUSED */
sortVertical(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    setDsplOrder(wnd, DIRW_VERT);
}

/* fmQuit *******************************************************************
 * quit gracefully: ask for each window whether to do the changes
 ****************************************************************************/

/* ARGSUSED */
fmQuit(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    quitFm(wnd, FMQUIT);
}

/* fmAbort ******************************************************************
 * abort: don't do changes
 ****************************************************************************/

/* ARGSUSED */
fmAbort(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd)
    quitFm(wnd, FMABORT);
}

/* getFilePattern: ************************************************************
 * set the pattern that is used to select the files of a directory that should
 * be presented in the directory window; then update the window
 *****************************************************************************/

/* ARGSUSED */
getFilePattern(parms)
     RequestParm *parms;
{
  char buf[21];
  register char *t = NULL;
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);

  buf[0] = EOS;
  sprintf(buf, "%s", wnd->filePattern);
  t = exGetEntry(parms->parent, -1, -1, "Enter file pattern: ", NULL, 20);

  if (t && *t) {
    markVisibleItems(wnd, t, False);
    setDirBoxSizeHints(wnd);			/* adapt window size */
  }
}      

/* showFileAttributes: ********************************************************
 * show the attributes of the selected files in the usual UNIX manner 
 *****************************************************************************/

/* ARGSUSED */
showFileAttributes(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);
  if (wnd) {
    displayAttributes(wnd, Head->attrEditor);
    XtPopup(Head->attrShell, XtGrabNone);
  }
}

/* makeDir: *******************************************************************
 * create a new dierctory in the current directory; the user is requested to 
 * provide a name for the new directory;
 *****************************************************************************/

/*ARGSUSED*/
makeDir(parms)
     RequestParm *parms;
{
  Arg args[1];
  int cnt;
  dirWindow *wnd;
  char *dir;
  char buf[PATHLEN];

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(((exMenu *)(parms->menu))->menubar->menubar),args, cnt);

  if (! wnd) {
    exError("fm", "makeDir: internal error: no current directory !", NULL);
    return;
  }

  dir = exGetEntry(parms->parent, -1, -1, "Enter directory name: ", NULL, 20);

  if (! dir || ! *dir)
    return;

  strcpy(buf, wnd->u_path);
  strcat(buf, "/");
  strcat(buf, dir);

  if (mkdir(buf, 0755) < 0)
    exError("fm", "makeDir: mkdir failed !", NULL);
}
