/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/******************************************************************************
 * This file contains the program to generate the History Tracers menus;
 * the generated application node "history menus" is stored in 
 * file "history.menu"
 ****************************************************************************/

#include <exBasics.h>
#include <exFonts.h>
#include <exErrors.h>
#include <exMenu.h>

static char helpText[200];

static void stopHistory();
static void contHistory();
static void saveHistory();
static void saveAsMacro();
static void restoreHistory();
static void quitHistory();

main(argc, argv)
     int argc;
     char *argv[]; 
{
  Widget top;
  ApplicationNode *HistoryMenus;
  ApplicationNode *createHistoryMenus();

#ifdef DEBUG
  malloc_debug(2);
#endif

  top = XtInitialize(NULL, "history editor menus", (XrmOptionDescRec *) NULL, 
		     0, &argc, argv);

  exInitLayout(XtDisplay(top));
  initMDS();

  if ((HistoryMenus = createHistoryMenus(top)) == (ApplicationNode *) 0)
    printf("application node edit menus not generated!\n");

  exit(0);
}

/* createHistoryMenus ***************************************************
 * this file contains the information to generate the menus needed
 * for the history tracer
 *************************************************************************/

ApplicationNode *createHistoryMenus(parent)
     Widget parent;
{
  ApplicationNode *HIST_MENUS = (ApplicationNode *) 0;  
  MHmenu *editMenu;
  MHitem *editItem;
  MHiReq  *Req;
  int actmask[6];

  /* new Application Node for history menus */
  HIST_MENUS = createApplNode();
  initApplNode(HIST_MENUS, "history menus");
  setAccess(HIST_MENUS, MDS_CHNGACC | MDS_SYS);
  
  if (insertApplNode(HIST_MENUS) == ERR_MAX)
    printf("cannot create Application Node; too many nodes\n");

  insertFunction(HIST_MENUS, stopHistory, "stopHistory");
  insertFunction(HIST_MENUS, contHistory, "contHistory");
  insertFunction(HIST_MENUS, saveHistory, "saveHistory");
  insertFunction(HIST_MENUS, saveAsMacro, "saveAsMacro");
  insertFunction(HIST_MENUS, restoreHistory, "restoreHistory");
  insertFunction(HIST_MENUS, quitHistory, "quitHistory");

  /* create edit menu QUIT for Application Node HIST_MENUS */

  editMenu = newMenu("quit");
  setSelCnt(editMenu,1);                             /* set selection count */
  setSelType(editMenu,MDS_CNT_EQ);                   /* set selection type */
  setMenuType(editMenu, MDS_M_PULLD);                /* set menu type */
  setMenuFlag(editMenu, MDS_READACC);                /* set access rights */
  sprintf(helpText, "Option to quit the history tracer.");
  editMenu->m_hlp->hlpmsg = strCopy(helpText);

  editItem = newItem(editMenu, "quit", EOS);
  setRetPattern(editItem, "$");
  Req = makeOneRequest(quitHistory, (char *) 0, (char *) 0, 0);
  insertRequest (editItem, Req, 0);
  sprintf(helpText, "Quit the history tracer.");
  editItem->i_hlp->hlpmsg = strCopy(helpText);

 /* create edit menu OPTIONS for Application Node HIST_MENUS */

  editMenu = newMenu("options");
  setSelCnt(editMenu,1);                             /* set selection count */
  setSelType(editMenu,MDS_CNT_EQ);                   /* set selection type */
  setMenuType(editMenu, MDS_M_PULLD);                /* set menu type */
  setMenuFlag(editMenu, MDS_READACC);                /* set access rights */
  sprintf(helpText, "Options to control the history tracer.");
  editMenu->m_hlp->hlpmsg = strCopy(helpText);

  setMask(actmask, 6, 0);
  actmask[0] = 1;
  actmask[2] = 1;
  actmask[3] = 1;
  actmask[4] = 1;
  actmask[5] = 1;
  setMenuInitialActivityMask(editMenu, actmask);      /* set activity mask */

  /* create items for edit menu OPTIONS */

  editItem = newItem(editMenu, "stop history", EOS);
  setRetPattern(editItem, "$");                      /* set return pattern */
  Req = makeOneRequest(stopHistory, "", (char *) 0, 0);
  insertRequest (editItem, Req, 0);                  /* set request */
  sprintf(helpText, "Stop recording your commands\nuntil the option <<continue history>> in this menu is selected.");
  editItem->i_hlp->hlpmsg = strCopy(helpText);

  editItem = newItem(editMenu, "continue history", EOS);
  setRetPattern(editItem, "$");
  Req = makeOneRequest(contHistory, (char *) 0, (char *) 0, 0);
  insertRequest (editItem, Req, 0);
  sprintf(helpText, "Resume recording\nif the option <<stop history>> in this menu has been selected before.");
  editItem->i_hlp->hlpmsg = strCopy(helpText);

  editItem = newItem(editMenu, "save history", EOS);
  setRetPattern(editItem, "$");
  Req = makeOneRequest(saveHistory, (char *) 0, (char *) 0, 0);
  insertRequest (editItem, Req, 0);
  sprintf(helpText, "Save the contents of the history tracer to a file.\nYou are requested for a file name.");
  editItem->i_hlp->hlpmsg = strCopy(helpText);

  editItem = newItem(editMenu, "save history as macro", EOS);
  setRetPattern(editItem, "$");
  Req = makeOneRequest(saveAsMacro, (char *) 0, (char *) 0, 0);
  insertRequest (editItem, Req, 0);
  sprintf(helpText, "Save the contents of the history tracer as macro.\nYou are requested to enter the name of a macro set\nand the name for the new macro.");
  editItem->i_hlp->hlpmsg = strCopy(helpText);

  editItem = newItem(editMenu, "load history", EOS);
  setRetPattern(editItem, "$");
  Req = makeOneRequest(restoreHistory, (char *) 0, (char *) 0, 0);
  insertRequest (editItem, Req, 0);
  sprintf(helpText, "Load a stored history in the history tracer's editor buffer.\nYou are requested to enter the name of the desired file.");
  editItem->i_hlp->hlpmsg = strCopy(helpText);

  setFile(HIST_MENUS, "history.menu");
  dumpToFile(parent); 

  return(HIST_MENUS);
}

static void stopHistory()
{
}

static void contHistory()
{
}

static void saveHistory()
{
}

static void saveAsMacro()
{
}

static void restoreHistory()
{
}

static void quitHistory()
{
}





