#ifndef _ex_editor_h_
#define _ex_editor_h_

#include <X11/X.h>

typedef int Editor;

/****************************************************************************
 *
 *				 constant definitions
 *
 ****************************************************************************/

#define CMD_CMD     	0       /* for 'normal commands */
#define ESC_CMD	    	1	/* for escape command table */
#define CX_CMD      	2	/* for ctrl-x commands */

#define ESC  		0x1b 

#define MAXBUF 		200	/* number of buffers (and editors) */
#define MAXLINE 	2000	/* longest text line that can be stored */
#define ED_MAXWIDTH 	200	/* longest line on screen */

#define ED_CHANGED 	0x1
#define ED_EXTERNBUFFER 0x2	/* editor buffer is provided by the caller */
#define ED_ACTIVE	0x4     /* mouse is in editor window */
#define ED_SCROLLBAR    0x8     /* provide a vertical scrollbar */
#define ED_MSGWINDOW    0x10	/* messages to window */
#define ED_VIEWMODE     0x20	/* editing not allowwd */
#define ED_HELP         0x40    /* provide a help window */
#define ED_CUROFF       0x80    /* make cursor invisible */
#define ED_AUTOKILL     0x100   /* delete automatically on buffer overflow */
#define ED_FRAMEOFF	0x200   /* hide cursor on WindowLeave events */
#define ED_MAPPED       0x400   /* editor window is mapped & visible */
#define ED_ENTRY        0x800   /* editor is actually an entry */
#define ED_LINEBREAK    0x1000  /* long lines are broken into rows */
#define ED_JUMPSCROLL	0x2000  /* scroll in jumps */

#define TABINTERVAL 8
#define MARGIN_JUMP 10		/* move left margin in such jumps */
#define ALLOCUNIT 10		/* amount to grow line */
				/* delete lines on buffer overflow */
#define KILLAMOUNT (bf->bufPtr->numlines / 10)

#define THERM_UPDATE_INCR  30	/* update thermometer in file-io */

#define LINE_SEP   "\\"

/******************************************************************************
 * the following drawing illustrates the layout of a character cell
 * with character coordinates (row, col):
 *
 * the pixel coordinates of the cell are:
 * x = col * FONTWIDTH
 * y = row * LINEHEIGHT
 * 'o' denotes the pixels of the character
 * 'l' denotes pixels that are used to draw the cursor or the frame
 *
 *  (x, y):      			    lllllllllll
 *  (x, y + FONTGAP):          		      ooooooo l
 *                             	             o       ol
 *                             	             o       ol
 *                             	             o       ol
 *                             	             o       ol
 *                             	             o       ol
 *                             	             o       ol
 *                             	             o       ol
 *                                            ooooooo l
 *  (x, y + LINEHEIGHT - FONTGAP            lllllllllll
 *  start of char in next line 
 *   
 ***************************************************************************/  
  
#define FONTGAP         1
#define LINEHEIGHT(f)   (exFontHeight(f) + 2 * FONTGAP)

#define REGION_DISPLAY  0x1	/* display region inverted */

/****************************************************************************
 *
 *				 data types  
 *
 ****************************************************************************/

/* used to pass information to the function for file name expansion */
typedef struct _expandInfo {
  char *word;			/* word to be expanded */
  int (* getcwd)();		/* function to extract the cur. directory */
  int matchCnt;			/* holds the number of matches */
  char *retBuf;			/* holds the marked up expansion list */
  Widget parent;		/* used to generate a menu */
} exExpandInfo;


struct Region {
   int m_line, m_pos;		/* line, pos of mark (buffer coordinates) */ 
   int d_line, d_pos;		/* line, pos of dot (buffer coordinates) */ 
   int flags;
};

typedef struct _editCmd {
  char key; 			/* key to call command */
  int (* func)();		/* function pointer */
} editCmd;

typedef struct _BufInfo {	/* info for each buffer */
  Widget w;			/* the editor Widget */
  Widget edt;			/* the text area */
  Widget msg;                	/* message Widget */
  Widget scrollbar;	       	/* the scrollbar */
  Editor hlp;			/* help Widget */
  int cols, lines;		/* screen's cols and lines */
  int height, width; 		/* screen dimensions in pixels */
  XFontStruct *font;		/* editor font */
  editCmd *cmdTable, 		/* holds the command table */
  	  *escTable,		/* the escape commands */
  	  *cxTable; 		/* the ctrl-x prefixed commands */
  char *filename;		/* file associated with buffer */
  char *name;                   /* editor name */
  int topline;			/* number of topmost line on screen */
  int curline;			/* number of current line in text */
  int curpos;			/* current cursor position in text */
  int currow;			/* current cursor row on screen */
  int curcol;			/* current cursor column on screen */
  int lmarg;			/* current left margin */
  struct Region region;
  int oldcol, oldrow;		/* last cursor position */
  struct _Buffer *bufPtr;	/* a pointer to the text buffer structure */
  int bufnum;			/* the buffer number for this buffer */
  char sbuff[20];		/* search buffer */
  int (* oldFnct)();		/* last executed command */
  int preservedCol;		/* in col preserving commands: stores col */
  int edPtr;			/* points to the char currently processed */
  enum {ISNEUTRAL, ISCONTR_X, ISESCAPE, ISNUMBER} lex_state;
  int lex_count;
  int flags;
  int (* inputFnct)();		/* interprets key events */
  char csrChr;			/* the char currently under the cursor */
  struct {			/* coordinates of cursor frame */
    int x, y, w, h;
  } csrFrame;
  int baseline;			/* baseline to draw characters */
  int fw, fh;			/* width and height of a charater */
  Cursor cursor;		/* cursor for editor main window */
  char *killBuff;		/* kill buffer */	
} BufInfo;

typedef struct _Buffer {	/* info for each buffer */
  char **text;			/* array of text lines */
  int numlines;			/* number of lines stored in buffer */
  int bufferLines;		/* size of text-buffer */
  Boolean changed;		/* buffer changed scince last write? */
} Buffer;

typedef struct _EditorDescr {
  int (* inputFnct)();			/* interprets key events */
  char *file;				/* initial file */
  char *name;				/* editor name */
  char **buffer;			/* extern buffer (if flag is on) */
  Widget msg;				/* the msg Widget if requested */
  int bufferLines;			/* bufferlines allocated */
  int lines, cols,			/* lines and cols of editor */
      width, height,			/* width and height of Widget */
      flags;
} EditorDescr;

typedef struct _cmdDescr {
  int (* func)();
  char *txt;
} cmdDescr;


static char exEditSccsID[] = "%W%\t%G%";            /* sccs keyword */

/****************************************************************************
 *
 *				 external declarations 
 *
 ****************************************************************************/

