/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include <exBasics.h>
#include "exEdit.h"
#include "display.h"

static char SccsID[] = "@(#)region.c	1.2\t11/4/92";            /* sccs keyword */

static char msg[100];

Boolean prepareRegion();
Boolean ETdeleteRegion();
char *Emtext(), *ETtext();

/* setDot: ****************************************************************
 * define a dot position, reset current mark; 
 ****************************************************************************/

setDot(bf, col, row)
     BufInfo *bf;
     int col;			/* col (x-pos on screen) */
     int row;			/* row (y-pos on screen) */
{
  screen2phys(bf, row, col, &bf->region.d_line, &bf->region.d_pos, False);
  bf->region.m_line = -1;
  bf->region.m_pos  = -1;
  sprintf(msg, "dot set (row,col)=(%d, %d)", row, col);
  ESmsg(bf, msg, False);
}

/* setMark: ****************************************************************
 * define a mark position; may not set unless a dot position has already 
 * been defined. It is guaranteed that the dot-positon is always befor the
 * mark (eventually, dot and mark are exchanged)
 ***************************************************************************/

Boolean setMark(bf, col, row)
     BufInfo *bf;
     int col;			/* col (x-pos on screen) */
     int row;			/* row (y-pos on screen) */
{
  if (bf->region.d_pos < 0 || bf->region.d_line < 0) {
    ESmsg(bf, "Cannot set mark; define dot first", False);
    return(False);
  }

  screen2phys(bf, row, col, &bf->region.m_line, &bf->region.m_pos, False);

  sprintf(msg, "mark set (row,col)=(%d, %d)", row, col);
  ESmsg(bf, msg, False);

  return(True);
}
  
/* dot2mark: *****************************************************************
 * exchange dot and mark
 ****************************************************************************/

dot2mark(bf)
     BufInfo *bf;
{
  int m_line, m_pos;

  m_line = bf->region.m_line;
  m_pos  = bf->region.m_pos;

  bf->region.m_line = bf->region.d_line;
  bf->region.m_pos  = bf->region.d_pos;

  bf->region.d_line = m_line;
  bf->region.d_pos  = m_pos;

  sprintf(msg, "dot (row, col)=(%d, %d), mark (row, col) = (%d, %d", 
	  		bf->region.d_line, bf->region.d_pos, 
	  		bf->region.m_line,bf->region.m_pos);
  ESmsg(bf, msg, False);
}

/* resetRegion: **************************************************************
 * reset the region 
 ****************************************************************************/

reset_region(bf)
     BufInfo *bf;
{
  bf->region.m_line = -1;
  bf->region.m_pos  = -1;
  bf->region.d_line = -1;
  bf->region.d_pos  = -1;
  bf->region.flags = 0;
}

/* copyRegion: ***************************************************************
 * return the text within the region defined by dot and mark
 ****************************************************************************/

char *copyRegion(bf)
     BufInfo *bf;
{
  register char *t, *l, *s;
  int i, line1, line2, pos1, pos2;
  register struct Region *r;
  
  r = &bf->region;

  if (! prepareRegion(r, &line1, &line2, &pos1, &pos2)) {
    ESmsg(bf, "Nothing copied !", False);
    return(NULL);			/* illegal region specified */
  }
  
  if (line1 == line2) {			/* dot & mark in one line */
    if (pos1 == pos2) {			/* NULL region */
      ESmsg(bf, "Nothing copied !", False);
      return(NULL);
    }
    t = bf->bufPtr->text[line1];
    l = Malloc(pos2 - pos1 + 2);

    bcopy((char *) &t[pos1], l, pos2 - pos1 + 1);
    sprintf(msg, "%d characters copied !", strlen(l));
    ESmsg(bf, msg, False);
    return(l);
  } else {				/* dot & mark in different lines */
    l = Malloc((line2 - line1 + 1) * MAXLINE);
    
    /* line with dot */
    t = bf->bufPtr->text[line1];
    strcpy(l, &t[pos1]);
    strcat(l, "\n");
    /* copy lines in between */
    for (i = line1 + 1; i < line2; i++) {
      strcat(l, bf->bufPtr->text[i]);
      strcat(l, "\n");
    }
    /* line with mark */
    t =  bf->bufPtr->text[line2];
    strncat(l, t, pos2 + 1);
    if (t[pos2] == '\0')
      strcat(l, "\n");

    s = strCopy(l);
    sprintf(msg, "%d characters copied !", strlen(l));
    ESmsg(bf, msg, False);
    free(l);
    return(s);
  }
}

/* ETdeleteRegion: ************************************************************
 * delete the region defined by dot and mark; 
 ****************************************************************************/

Boolean ETdeleteRegion(bf, lcnt, rcnt)
     register BufInfo * bf;
     int *lcnt;		/* number of lines deleted */
     int *rcnt;		/* number of corresponding rows (that are deleted) */
{
  register char *t, *l;
  int len, line1, line2, pos1, pos2;
  int row1, row2, col;
  register struct Region *r;
  char text[MAXLINE];

  r = &bf->region;

  if (! prepareRegion(r, &line1, &line2, &pos1, &pos2))
    return(-1);			/* illegal region specified */
  
  phys2screen(bf, line1, pos1, &row1, &col, False);
  phys2screen(bf, line2, pos2, &row2, &col, False);

  if (line1 == line2) {			/* dot & mark in one line */
    if (pos1 == pos2)			/* NULL region */
      return(False);

    t = bf->bufPtr->text[line1];	/* get line */

    if (pos2 >= strlen(t)) {		/* whole line */
      t[pos1] = EOS;			/* cut off rest of line */
      l = strCopy(t);			/* copy new line */
      free(t);
      bf->bufPtr->text[line1] = l;
      *lcnt = 0;
      *rcnt = row2 - row1;
      return(True);
    }

    l = Malloc(pos1 + strlen(t) - pos2 + 1);      

    bcopy(t, l, pos1);			/* copy first part */
    strcat(l, &t[pos2 + 1]);		/* append last part */
    free(t);
    bf->bufPtr->text[line2] = l;
    *lcnt = 0;
    *rcnt = row2 - row1;
    return(True);
  } else {				/* dot & mark in different lines */
    *rcnt = row2 - row1;

    bf->bufPtr->text[line1][pos1] = EOS;/* line with dot */
    t = strCopy(bf->bufPtr->text[line1]);
    free(bf->bufPtr->text[line1]);
    bf->bufPtr->text[line1] = t;
	
    if (t = bf->bufPtr->text[line2]) {	/* line with mark */
      len = strlen(t);
      if (pos2 == len) {	      	/* clear whole line */
	free(bf->bufPtr->text[line2]);
	bf->bufPtr->text[line2] = strCopy("");
      } else {				/* clear up to mark */
	l = strCopy(&t[pos2 + 1]);
	free(t);
	bf->bufPtr->text[line2] = l;
      } 
    }

    /* delete lines in between if any */

    ETdellines(bf, line1 + 1, line2 - line1 - 1);
    if (ETlen(bf, line1) + ETlen(bf, line1 + 1) > MAXLINE) {
      ESmsg(bf, "Line too long !", True);
      Bell();
    } else {
      strcpy(text, ETtext(bf, line1));
      strcat(text, ETtext(bf, line1 + 1));
      ETreplace(bf, line1, text);
      ETdellines(bf, line1 + 1, 1);
    }
    *lcnt = line2 - line1;
    *rcnt = row2 - row1;
    return(True);
  }
}

/* ESdeleteRegion: ***********************************************************
 * perform the visual effects when a region is deleted; make the start of
 * the region the new current position 
 ****************************************************************************/

ESdeleteRegion(bf, lcnt, rcnt)
     BufInfo *bf;
     int lcnt;	    	/* number of lines that have been deleted */
     int rcnt;	    	/* number of rows that have been deleted */
{
  int line1, line2, pos1, pos2, row, c, l;
  struct Region *r = &bf->region;

  csrChrForget(bf); 

  if (! prepareRegion(r, &line1, &line2, &pos1, &pos2))
    return;				/* illegal region specified */

  if (lcnt == 0) {
    ESdisplayLine(bf, line1, pos1);
    phys2screen(bf, line1, strlen((char *) EMtext(bf, line1, False)), 
		&row, &c, True);
    PSclear(bf, row, c);
    phys2screen(bf, line1, pos1, &bf->currow, &bf->curcol, True);
    ESshowcur(bf, True, bf->curcol, bf->currow);
    return;
  }
    
  (void) showline(bf, line1, pos1);	  /* first line of region */

  showRangeOfLines(bf, line1, line1);	/* show first line */
  phys2screen(bf, line1, strlen((char *) EMtext(bf, line1, False)), 
	      &row, &c, True);
  PSclear(bf, row, c);

  if (rcnt < 1)
    return;

  /* move lines up */
  ESmove_lines(bf, row + rcnt + 1, rcnt, DIR_UP);

  /* calculate last line that fits on screen */
  for (row = 0, l = bf->topline; l < ETnumlines(bf); l++) {
    c = screenLines(bf, l);
    row += c;
    if (row > bf->lines) {
      row -= c;
      l--;
      break;
    }
  }

  showRangeOfLines(bf, l - bf->topline - lcnt, l - bf->topline);
}


/* ESshowline: ***************************************************************
 * display the specified physical line 
 ****************************************************************************/

ESshowline(bf, line, attr)
     BufInfo *bf;
     int line;	
     int attr;
{
  int l, sl, slen, row, col, cnt, w;
  char *t;

  if (line > ETnumlines(bf))
    return;

  w = ESwidth(bf);

  if (! phys2screen(bf, line, 0, &row, &col, False))
    return;

  sl = screenLines(bf, line);

  t = strCopy(EMtext(bf, line, False));
  slen = strlen(t);

  /* display screen-lines */
  if (bf->flags & ED_LINEBREAK) {			/* break long lines */
    for (l = 0, cnt = 0; l < slen && row < bf->lines; l += w) {
      if (cnt == sl - 1) {	 			/* last row */
	PSwrite(bf, row, 0, strlen(&t[l]), &t[l], attr);
	free(t);
	return;
      } else {
	PSwrite(bf, row, 0, w, &t[l], attr);
	PSwrite(bf, row, w, 1, LINE_SEP, attr);
	row++;
	cnt++;
      }
    }
  } else {						/* no linebrak */
    if (bf->lmarg <= slen)
      PSwrite(bf, row, -bf->lmarg, strlen(&t[bf->lmarg]), &t[bf->lmarg], attr);
  }
  free(t);
}

/* ESshowlinerange: **********************************************************
 * display specified range of physical line 
 ****************************************************************************/

ESshowlinerange(bf, line, pos1, pos2, attr)
     BufInfo *bf;
     int line;			
     int pos1, pos2;     /* position in text buffer (without tabs expanded) */
     int attr;
{
  int col0, row0, col1, col2, row1, row2, sl = 0, w, l, i, num;
  char *t;

  if (line > ETnumlines(bf) || line < 0)
    return(-1);
  
  t = strCopy(EMtext(bf, line, False));

  if (strlen(t) > 0)
    sl = ETlen(bf, line)/*strlen(t)*/;

  w = ESwidth(bf);

  if (pos1 > sl)
    return(-1);

  if (pos2 < 0 || pos2 > sl)
    pos2 = sl;

  if (! phys2screen(bf, line, 0, &row0, &col0, False))
    return(-1);

  if (! phys2screen(bf, line, pos1, &row1, &col1, False))
    return(-1);

  if (! phys2screen(bf, line, pos2, &row2, &col2, False)) {
    if (row2 >= bf->lines) {  /* row containing pos2 does not fit on screen;
				 no scrolling */      
      row2 = bf->lines - 1;
      col2 = w;
    }
  }
    
  num = row1 - row0;    /* number of rows between beginning of line and pos1 */

  if (bf->flags & ED_LINEBREAK) {
    if ((row1 == row2) && (col2 < w)) {			/* one display line */
      PSwrite(bf, row1, col1, col2 - col1, &t[num * w + col1], attr);
      PSclear(bf, row1, col2);
    }
    else {
      l = num * w + col1;
      for (i = row1; i <= row2; i++) {
	if (i == row1) {				/* 1st line */
	  PSwrite(bf, i, col1, w - col1, &t[l], attr);
	  PSwrite(bf, i, w, 1, LINE_SEP, attr);
	  PSclear(bf, i, w + 1);			/* clear to eol */
	  l += w - col1;
	} else if ((i == row2) && (col2 < w)) {    	/* last line */
	  PSwrite(bf, i, 0, col2, &t[l], attr);
	  PSclear(bf, i, col2);		        	/* clear to eol */
	} else {					/* in between */
	  PSwrite(bf, i, 0, w, &t[l], attr);
	  PSwrite(bf, i, w, 1, LINE_SEP, attr);
	  PSclear(bf, i, w + 1);	       		/* clear to eol */
	  l += w;
	}
      }
    }
  } else {						/* no linebreak */
    PSwrite(bf,row1, col1 - bf->lmarg, col2 - col1, &t[col1 + bf->lmarg],attr);
    PSclear(bf, row1, col2 /*- bf->lmarg + 1*/);	     /* clear to eol */
  }
  free(t);
  return(0);
}

/* ESdisplayRegion: **********************************************************
 * display a region according to the given flag 
 ****************************************************************************/

ESdisplayRegion(bf, inverted)
     BufInfo *bf;
     Boolean inverted;
{
  struct Region *r = &bf->region;
  int color, i, line1, line2, pos1, pos2;

  if (! prepareRegion(r, &line1, &line2, &pos1, &pos2))
    return;					/* illegal region specified */
  
  if (inverted) {
    bf->region.flags |= REGION_DISPLAY;		/* set display flag */
    color = A_INVERSE;
  } else {
    bf->region.flags &= ~REGION_DISPLAY;	/* reset display flag */
    color = A_NORM;
  }

  if (line1 == line2) {				/* one line region */
    if (pos1 == pos2)				/* empty region */
      return;
    else 
      ESshowlinerange(bf, line1, pos1, pos2, color);

  } else {					/* multi-line region */
    ESshowlinerange(bf, line1, pos1, -1, color);/* first line */
    for (i = line1 + 1; i < line2; i++)		/* in between */
      ESshowlinerange(bf, i, 0, -1, color);
    ESshowlinerange(bf, line2, 0, pos2, color);	/* last line */
  }
}

/* prepareRegion: ************************************************************
 * check the given region and sort dot and mark in increasing order
 ****************************************************************************/

Boolean prepareRegion(r, l1, l2, p1, p2)
     struct Region *r;
     int *l1, *l2;			/* first and last line */
     int *p1, *p2;			/* first and last position */
{
  if (! r || r->d_line < 0 || r->m_line < 0)
    return(False);

  if (r->d_line < r->m_line) {		/* dot-line  befor mark-line */
    *l1 = r->d_line;
    *p1 = r->d_pos;
    *l2 = r->m_line;
    *p2 = r->m_pos;
  } else if (r->d_line == r->m_line) {	/* dot and mark in same line */
    if (r->d_pos < r->m_pos) {		/* dot befor mark */
      *l1 = r->d_line;
      *p1 = r->d_pos;
      *l2 = r->m_line;
      *p2 = r->m_pos;
    } else {				/* dot after mark */
      *l1 = r->m_line;
      *p1 = r->m_pos;
      *l2 = r->d_line;
      *p2 = r->d_pos;
    }
  } else {				/* mark-line after dot-line */
    *l1 = r->m_line;
    *p1 = r->m_pos;
    *l2 = r->d_line;
    *p2 = r->d_pos;
  }

  return(True);
}

/* ESyank: ******************************************************************
 * update the display after the killbuffer has been inserted 
 ****************************************************************************/

ESyank(bf, cnt, oldcol, oldtop)
     BufInfo *bf;
     int cnt;                           /* number of lines inserted */
     int oldcol;			/* old col (before insertText) */
     int oldtop;			/* old topline (before insertText) */
{
  if (cnt == 1) 			/* one line yanked */
    ESshowrow(bf, bf->currow, oldcol);
  else if (cnt > 1) {			/* more than one line yanked */
    if (oldtop == bf->topline)   	/* within visible area */         
      showRangeOfLines(bf, bf->curline - cnt, lastLineOnScreen(bf)
		       /*bf->curline*/);
    else {				/* visible area changed */
      if (! showline(bf, bf->curline, bf->curpos)) {
	ESscroll(bf, bf->curline - bf->topline, DIR_UP);
	bf->curline = bf->topline;
	setcol(bf);
      }
    }
  }
  else
    return;

  csrChrForget(bf);
  PSsetcur(bf, bf->curline, bf->curpos);
  ESshowcur(bf, True, bf->curcol, bf->currow);
  exEditSetScrollbar(bf);
}

/* ESmove_lines: ************************************************************* 
 * move lines on screen 
 ****************************************************************************/

ESmove_lines(bf, row, cnt, dir)
     BufInfo *bf;
     int row;			/* first row to move */
     int cnt;			/* number of rows to move */
     int dir;			/* direction */
{
  RECT rect;

  RASG(&rect, row, 0, bf->lines, bf->cols);
  PSslide(bf, &rect, dir, cnt);
} 

/* exCopy: *******************************************************************
 * copy the marked region into the killbuffer 
 ****************************************************************************/

Boolean exCopy(bf)
     BufInfo *bf;
{
  if (bf->region.d_line < 0) {			/* not dot set yet */
    ESmsg(bf, "Cannot copy: dot not set !", True);
    Bell();
    reset_region(bf);				/* reset region */
    return(False);
  }

  if (bf->region.m_line < 0) {			/* no mark, set to cur. pos */
    if (! setMark(bf, bf->curcol, bf->currow))
      return False;
    else
      ESdisplayRegion(bf, True);
  }
  
  setKillBuffer(bf, copyRegion(bf));
  
  return True;
}

/* exCut: *******************************************************************
 * copy the marked region into the killbuffer and delete it from the editor 
 * buffer
 ****************************************************************************/

Boolean exCut(bf)
     BufInfo *bf;
{
  int lcnt, rcnt;

  if (bf->flags & ED_VIEWMODE) {
    ESmsg(bf, "Read only buffer!", False);
    return(False);
  }

  if (bf->region.d_line < 0) {			/* not dot set yet */
    Bell();
    ESmsg(bf, "Cannot cut: no mark set !", True);
    reset_region(bf);				/* reset region */
    return(False);
  }

  if (bf->region.m_line < 0) {			/* no mark, set to cur. pos */
    if (! setMark(bf, bf->curcol, bf->currow))
      return False;
    else
      ESdisplayRegion(bf, True);
  }
  
  setKillBuffer(bf, copyRegion(bf));
  ESdisplayRegion(bf, False);			/* hide region */
  if (ETdeleteRegion(bf, &lcnt, &rcnt))		/* delete region lines */
    ESdeleteRegion(bf, lcnt, rcnt);		/* delete region rows */
  reset_region(bf);
  return True;
}

/* exYank: *******************************************************************
 * copy the contents of the killbuffer at the current position
 ****************************************************************************/

exYank(bf)
     BufInfo *bf;
{
  int cnt, oldcol, oldtop;

  oldcol = bf->curcol;
  oldtop = bf->topline;

  cnt = yankKillBuffer(bf);

  phys2screen(bf, bf->curline, bf->curpos, &bf->currow, &bf->curcol, True);
  ESyank(bf, cnt, oldcol, oldtop);
}

