/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/******************************************************************************
 * this file contains the functions to dump and restore application nodes on
 * files
 *****************************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/dir.h>
#include <X11/Intrinsic.h>
#include <exBasics.h>
#include <exLayout.h>
#include <exErrors.h>
#include "exMDS.h"
#include "exMenu.h"
#include <exThermometer.h>

static char SccsID[] = "@(#)file.c	1.2\t6/7/93";            /* sccs keyword */

#define NO_FUNCTION     -999	/* is stored, if no function is provided */
#define NOEXIST   	(int) 0	/* is stored if a null-pointer is found */
#define EXIST 		(int) 1 /* is stored if a record follows */
  
  extern char* getenv();

Boolean legalAccess();
Boolean Mtyp();
Boolean Cnttype();
Boolean reprPtype();
Boolean reprType();
Boolean colorType();

static char msg[100];
static char fullName[256];

/* dumpToFile **************************************************************
 * dump the contents of an application structure into the specified file 
 ***************************************************************************/

Boolean dumpToFile(w)
     Widget w;
{
  char *file;
  int fd, cnt;
  register i, j;
  MHmenu *mptr;
  MHitem *iptr;
  ApplicationNode *aptr = currentApplication();
  
  if (!(file = getMenuPath(aptr)))
    return(False);
  
  if ((fd = open(file, O_WRONLY | O_CREAT | O_TRUNC, MDS_UMASK)) == ERROR) {
    setError(ERR_OPEN);
    return(False);
  }
  
  sprintf(msg, "Dumping -- %s -- to file -- %s -- ...", 
	  aptr->applicationName, file); 
  
  initThermometer(XtDisplay(w), w, 0, 0, strlen(msg), msg, NULL);
  setThermometer(100, 100, 0);
  
  /* dump application structure */
  writeString(fd, aptr->applicationName);
  Write(fd, &aptr->a_access, sizeof(int));
  writeString(fd, aptr->file);
  
  /* dump the user defined functions */
  dumpUserFuncs(fd, aptr);
  
  /* dump the menus */
  /* we don't trust the menuCnt and count the menus ourself */
  for (i = 0, cnt =0; i < MDS_MAXMENU; i++)
    if ((mptr = aptr->m_table[i]) && ! (mptr->m_flags & MDS_TEMP))
      cnt++;
  
  /* dump the number of menus to be stored */
  Write(fd, &cnt, sizeof(int));
  
  for (i = 0; i < MDS_MAXMENU; i++) {	/* for each menu */
    
    if (cnt > 4)				/* show progress */
      setThermometer(100, 100, i * 100 / cnt); 
    
    if ((mptr = aptr->m_table[i]) == (MHmenu *) 0)
      continue;
    
    /* store only permanent menus */
    if (mptr->m_flags & MDS_TEMP) 
      continue;
    
    dump_mRepr(fd,mptr->m_repr);
    Write(fd, &(mptr->m_type), sizeof(MDS_MENU_TYPE));
    Write(fd, &(mptr->m_flags), sizeof(int));
    Write(fd, &(mptr->i_selCnt), sizeof(int));
    Write(fd, &(mptr->i_selType), sizeof(MDS_CNT_TYPE));
    Write(fd, &(mptr->maxItem), sizeof(int));
    Write(fd, &(mptr->m_pos), sizeof(Pos));
    dumpHlp(fd, aptr, mptr->m_hlp);
    
    /* dump the activityMask */
    Write(fd,&(mptr->activityMask),sizeof(act_mask));
    
    /* dump the key */
    Write(fd,&(mptr->m_key),1);
    
    /* dump the items */
    /* we don't trust the item count, so we count the items ourself */
    for (j = 0, cnt = 0; j < MDS_MAXITEM; j++) 
      if ((iptr = mptr->i_table[j]) != (MHitem *) 0)
	cnt++;
    
    /* write the number of items that are to be stored */
    Write(fd, &(mptr->itemCnt), sizeof(int));
    
    
    /* now we are ready to dump the items */
    for (j = 0; j < MDS_MAXITEM; j++) {
      if ((iptr = mptr->i_table[j]) == (MHitem *) 0)
	continue;
      
      dump_iRepr(fd,iptr->i_repr);
      
      Write(fd,&(iptr->i_Pos),sizeof(Pos));
      dumpHlp(fd, aptr, iptr->i_hlp);
      dumpRet(fd,iptr->i_return);
      
      dumpRequests(fd, aptr, iptr->i_requests);
      
      Write(fd,&(iptr->i_repr->key),1);
      
      Write(fd,&(iptr->activityMask),sizeof(act_mask));
      
      Write(fd,&(iptr->flag),sizeof(int));
    }
  }
  setThermometer(100, 100, 100);
  hideThermometer(10);
  close(fd);
  free(file);
  aptr->_changed = False;
  return(True);
}

/* restoreFromFile: **********************************************************
 * rebuild an application node from the information stored in a file; if the 
 * application is already loaded, the parameter 'forced' determines the level 
 * of user interaction:
 * -1: do not overwrite the already existing node, stop restoring
 *  0: ask the user whether to overwrite or to stop 
 *  1: overwrite without asking the user
 ****************************************************************************/

Boolean restoreFromFile(w, file, forced)
     Widget w;
     char *file;
     int forced;	/* -1, 0, 1 */
{
  ApplicationNode *aptr;
  ApplicationNode *getApplByName();
  Boolean ok = False;
  int fd;
  MHmenu *mptr;
  MHitem *iptr;
  register i,j;
  char *applName, *menupath;
  
  if (! file || ! *file) {
    setError(ERR_NULL);
    return(False);
  }
  
  if (*file == '/') {				/* absolute file name */
    strcpy(fullName, file);
  } else {					/* create file name */
    if (! (menupath = getenv("MENUPATH"))) {
      if (! (menupath = getenv("HOME"))) {
	setError(ERR_NOEXIST);
	return(False);
      }    
    }
    
    if (makeFullName(menupath, file, fullName, 256) < 0) {
      setError(ERR_MAX);
      return(False);
    }
  }
  
  file = fullName;
  
  if (access(file,R_OK) == ERROR) {	/* no read permission */
    setError(ERR_READ);
    return(False);
  }
  
  if ((fd = open(file,O_RDONLY)) == ERROR) { /* cannot open file */
    setError(ERR_OPEN);
    return(False);
  }
  
  ReadString(fd, &applName);
  
  if (aptr = getApplByName(applName)) {
    switch (forced) {
    case -1: 				/* do not overwrite application node */
      break;
    case 0:
      sprintf(msg, "Application node --%s-- already loaded; replace it ?",
	      applName);
      switch (exNotify2User(w, -1, -1, msg)) {/* get confirmation from user */
      case ANSWERYES:
	ok = True;
	break;
      case ANSWERCANCEL:
	return;
      case ANSWERNO:
	ok = False;
	break;
      }
      break;
    case 1:				/* overwrite application node */
      ok = True;
      break;
    default:
      setError(ERR_VALUE);
    }
    
    if (! ok) {		    		/* abort restore */
      free(applName);
      close(fd);
      printf("file not restored (%s)!\n", file);
      setCurrentApplication(aptr);
      return(True);
    } else 				/* replace application */
      removeApplNode(aptr);
  }
  
  sprintf(msg, "%s application -- %s -- ...", 
 	  ok ? "Overwriting" : "Loading", applName); 
  
  initThermometer(XtDisplay(w), w, 100, 100, strlen(msg), msg, NULL);
  setThermometer(-1, -1, 0);
  
  if (insertApplNode(aptr = createApplNode()) != ERR_NOERR) {
    setError(ERR_CREATE); 
    return(False);
  }
  
  /* read application structure */
  aptr->applicationName = applName;
  Read(fd,&(aptr->a_access), sizeof(int));
  ReadString(fd,&(aptr->file));
  
  /* restore the table of user defined functions */
  
  restoreUserFuncs(fd, aptr);
  
  /* restore the menus */
  
  /* read the number of menus that are stored */
  Read(fd,&(aptr->menuCnt),sizeof(int));
  
  if (aptr->menuCnt > MDS_MAXMENU || aptr->menuCnt < 0) {
    setError(ERR_VALUE);
    return(False);
  }
  for (i = 0; i < aptr->menuCnt; i++) {	/* for each menu */
    
    if (aptr->menuCnt > 4)	      	/* show progress */
      setThermometer(100, 100, i * 100 / aptr->menuCnt); 
    
    mptr = createMenu();
    
    if (restore_mRepr(fd,&(mptr->m_repr)) == ERROR)
      return(False);
    
    Read(fd,&(mptr->m_type),sizeof(MDS_MENU_TYPE));
    if (Mtyp(mptr->m_type) == False)
      return(False);
    
    Read(fd,&(mptr->m_flags),sizeof(int));
    if (legalAccess(mptr->m_flags) == False)
      return(False);
    
    Read(fd,&(mptr->i_selCnt),sizeof(int));
    if (mptr->i_selCnt < 0) {
      return(setError(ERR_VALUE));
    }
    
    Read(fd,&(mptr->i_selType),sizeof(MDS_CNT_TYPE));
    if (Cnttype(mptr->i_selType) == False)
      return(False);
    
    Read(fd,&(mptr->maxItem),sizeof(int));
    
    if (restorePos(fd,&(mptr->m_pos)) == ERROR)
      return(False);
    
    restoreHlp(fd, aptr, mptr->m_hlp);
    
    /* restore the activityMask */
    Read(fd,&(mptr->activityMask),sizeof(act_mask));
    
    /* restore the key */
    Read(fd,&(mptr->m_key),1);
    
    /* restore the items */
    /* read the number of items that have been stored */
    Read(fd,&(mptr->itemCnt),sizeof(int));
    if (mptr->itemCnt > MDS_MAXITEM || mptr->itemCnt < 0) {
      setError(ERR_VALUE);
      return(False);
    }
    /* check whether the number of visible iteme per menu is valid */
    if (mptr->maxItem < 0) {
      setError(ERR_VALUE);
      return(False);
    }
    
    for (j = 0; j < mptr->itemCnt; j++) {
      iptr = createItem(); 
      
      if (restore_iRepr(fd,&( iptr->i_repr)) == ERROR)
	return(False);
      
      if (restorePos(fd,&(iptr->i_Pos)) == ERROR)
	return(False);
      
      if (restoreHlp(fd, aptr, iptr->i_hlp) == ERROR)
	return(False);
      
      if (restoreRet(fd,iptr->i_return) == ERROR)
	return(False);
      
      if (restoreRequests(fd, aptr, iptr->i_requests) == ERROR)
	return(False);
      
      Read(fd,&(iptr->i_repr->key),1);
      
      Read(fd,&(iptr->activityMask), sizeof(act_mask));
      
      Read(fd,&(iptr->flag), sizeof(int));
      
      mptr->i_table[j] = iptr;	/* instal the item in the menu item-table */
    }
    aptr->m_table[i] = mptr;	/* install the menu in the appl. menu-table */
  }
  
  setThermometer(100, 100, 100);	       	
  hideThermometer(10);
  
  close(fd);

  if (ok) 
    setError(ERR_EXISTS);	/* existing application node replaced */
  else 
    setError(ERR_NOERR);	/* loaded new application */

  return(True);
}

/******************************************************************************
 *				dump specified components
 *****************************************************************************/

/* dumpUserFuncs: ************************************************************
 * dump the user-defined functions in a specific format: the number of 
 * functions is stored first, followed by an integer for each function that 
 * represents the index of that function in the MH_FUNCTONS array and a 
 * textual description of the function
 ****************************************************************************/

dumpUserFuncs(fd, aptr)
     int fd;
     ApplicationNode *aptr;
{
  int i, cnt = 0;
  
  for (i = 1; i < MDS_MAXFUNC; i++)
    if (aptr->ApplFunc[i].func != (int (*)()) 0)
      cnt++;
  
  Write(fd, &cnt, sizeof(int));
  
  if (cnt == 0)
    return(setError(ERR_NOERR));
  
  for (i = 1; i < MDS_MAXFUNC; i++) {
    if (aptr->ApplFunc[i].func != (int (*)()) 0) {
      cnt = whichFunction(aptr, aptr->ApplFunc[i].func);
      Write(fd, &cnt, sizeof(int));
      writeString(fd, aptr->ApplFunc[i].funcName);
    }
  }
  return(setError(ERR_NOERR));
}

/* dump_mRepr: ***************************************************************
 * dump the representation record
 ****************************************************************************/

dump_mRepr(fd,m_repr)
     int fd;
     MHmRepr *m_repr;
{
  static exist = EXIST;
  static noexist = NOEXIST;
  
  if (m_repr) {	/* store representation */
    Write(fd,&exist,sizeof(int));
    Write(fd, &(m_repr->reprType), sizeof(MDS_REPR_TYPE));
    writeString(fd,m_repr->name);
  } else
    Write(fd,&noexist,sizeof(int));
  
  return(OK);
}

/* dump_iRepr: **************************************************************
 * dump the representation record; if no representationis provided, 
 * a NULL (int 0) is stored 
 ****************************************************************************/

dump_iRepr(fd,repr)
     int fd;
     MHiRepr *repr;
{
  static exist = EXIST;
  static noexist = NOEXIST;
  
  if (repr) {	/* store representation */
    Write(fd,&exist,sizeof(int));
    writeString(fd,repr->name);
    Write(fd, &(repr->key),1);
  } else
    Write(fd,&noexist,sizeof(int));
  
  return(OK);
}

/* dumpHlp: ******************************************************************
 * the help record is dumped; the function is dumped as an index into the MDS
 * function array;
 *****************************************************************************/

dumpHlp(fd, aptr, hlp)
     int fd;
     ApplicationNode *aptr;
     MHhlp *hlp;
{
  int indx = whichFunction(aptr, hlp->hlpFnct);
  writeString(fd,hlp->hlpmsg);
  Write(fd,&(indx),sizeof(int));
  return(OK);
}

/* dumpRet: ******************************************************************
 * dump the return record; the function is dumped as an index into the MDS
 * function table
 *****************************************************************************/

dumpRet(fd,rptr)
     int fd;
     MHiRet *rptr;
{
  writeString(fd,rptr->retPattern);
  Write(fd,&(rptr->reqCnt),sizeof(int));
  return(OK);
}

/* dumpRequests: ************************************************************
 * dump the array of requests; a leading vount of really stored request-records
 * is stored 
 ***************************************************************************/

dumpRequests(fd, aptr, requests)
     int fd;
     ApplicationNode *aptr;
     MHiReq *requests[];
{
  register i;
  int cnt = 0;
  
  for (i = 0; i < MDS_MAXREQ; i++) 
    if (requests[i] != (MHiReq *) 0)
      cnt++;
  
  Write(fd,&cnt,sizeof(int));	/* write the count of records to be stored */
  
  for (i = 0; i < cnt; i++)
    dumpReq(fd, aptr, requests[i]);
  return(OK);
}

/* dumpReq: ******************************************************************
 * dump a request record; the function is stored in the usual manner 
 *****************************************************************************/

dumpReq(fd, aptr, rptr)
     int fd;
     ApplicationNode *aptr;
     MHiReq *rptr;
{
  int indx = whichFunction(aptr, (int (*)()) rptr->reqFnct);
  
  writeString(fd,rptr->reqData);
  writeString(fd,rptr->initialValue);
  writeString(fd,rptr->prelimInitialValue);
  Write(fd,&(rptr->flags), sizeof(int));
  Write(fd,&(rptr->size), sizeof(int));
  Write(fd,&indx,sizeof(int));
  Write(fd,&(rptr->cnt), sizeof(int));
  Write(fd, &(rptr->reqCntType), sizeof(MDS_CNT_TYPE));
  return(OK);
}

/*****************************************************************************
 *
 *			restore specified components
 *
 *****************************************************************************/

/* restoreUserFuncs: *********************************************************
 * restore  the set of user_defined request-functions
 *****************************************************************************/

restoreUserFuncs(fd, aptr)
     int fd;
     ApplicationNode *aptr;
{
  int cnt, i, indx;
  
  Read(fd, &cnt, sizeof(int));
  
  if (cnt <= 0)
    return(setError(ERR_VALUE));;
  
  for (i = 0; i < cnt; i++) {
    Read(fd,&(indx),sizeof(int));
    if (indx == NO_FUNCTION) {
      ReadString(fd,&(aptr->ApplFunc[0].funcName));
      return(OK);
    } else if (indx < 0) {
      aptr->ApplFunc[i].func = MDS_functions[-indx].func;
      ReadString(fd,&(aptr->ApplFunc[-indx].funcName));
    } else if (indx > MDS_MAXFUNC) {
      return(setError(ERR_VALUE));
    } else {				/* provided by application */
      aptr->ApplFunc[indx].func = MH_FUNCTIONS[indx].func;
      ReadString(fd,&(aptr->ApplFunc[indx].funcName));
    }
  }
  
  return(setError(ERR_NOERR));
}

/* restoreStringList: ********************************************************
 * restore an array of strings of max elements; first a count of the strings
 * is read, then a list of strings is restored 
 *****************************************************************************/

restoreStringList(fd,sptr,max)
     int fd;
     char **sptr;
{
  register i;
  int cnt;
  
  Read(fd,&cnt,sizeof(int));
  if (cnt < 0  || cnt > max) {
    return(setError(ERR_VALUE));
  }
  for (i = 0; i < cnt; i++) 
    ReadString(fd,&(sptr[i]));
  return(OK);
}

/* restore_mRepr: ************************************************************
 * restore the mRepr structure
 ****************************************************************************/

restore_mRepr(fd,repr)
     int fd;
     MHmRepr **repr;
{
  MDS_REPR_TYPE rtype;
  char *name;
  int flag;
  
  Read(fd,&flag,sizeof(int));
  switch (flag) {
  case EXIST: /* a representation has been stored */
    Read(fd,&rtype,sizeof(MDS_REPR_TYPE));
    ReadString(fd,&name);
    
    *repr = (MHmRepr *) Malloc(sizeof(MHmRepr));
    (*repr)->reprType = rtype;
    (*repr)->name = strCopy(name);
    break;
  case NOEXIST: /* no representation has been stored */
    *repr = (MHmRepr *) 0;
    break;
  default: /* illegal flag, abort restoring */
    return(setError(ERR_VALUE));
  }
  return(OK);
}

/* restore_iRepr: ************************************************************
 * restore the representation record; if no representation is provided, a 
 * dummy record has been stored consisting of the tripel(0,"",0)
 *****************************************************************************/

restore_iRepr(fd,repr)
     int fd;
     MHiRepr **repr;
{
  char *name;
  int flag;
  char key;
  
  Read(fd,&flag,sizeof(int));
  switch (flag) {
  case EXIST: /* a representation has been stored */
    ReadString(fd,&name);
    Read(fd,&key,1);
    *repr = new_iRepr(name, key);
    break;
  case NOEXIST: /* no representation has been stored */
    *repr = (MHiRepr *) 0;
    break;
  default: /* illegal flag, abort restoring */
    return(setError(ERR_VALUE));
  }
  return(OK);
}

/* restorePos: **************************************************************
 * read a position and check if x and y are valid 
 ****************************************************************************/
restorePos(fd,pos)
     int fd;
     Pos *pos;
{
  Read(fd,pos,sizeof(Pos));
  
  if (pos->posType != MDS_POS_NULL && pos->posType != MDS_POS_ABS &&
      pos->posType != MDS_POS_CREL)
    pos->posType = MDS_POS_NULL;      /* can live with that */
  
  return(OK);
}

/* restoreHlp: ***************************************************************
 * the help record is restored; the function is restored as an index into 
 * the MDS function array;
 ****************************************************************************/

restoreHlp(fd,aptr,hlp)
     int fd;
     ApplicationNode *aptr;
     MHhlp *hlp;
{
  int indx;
  
  ReadString(fd,&(hlp->hlpmsg));
  Read(fd, &indx, sizeof(int));
  if (indx == NO_FUNCTION) {
    hlp->hlpFnct = (int (*)())0;
    
    return(OK);
  } else if (indx < 0) {
    hlp->hlpFnct = MDS_functions[-indx].func;
    return(setError(ERR_NOERR));
  } else if (indx > MDS_MAXFUNC) {
    return(setError(ERR_VALUE));
  } else {
    hlp->hlpFnct = aptr->ApplFunc[indx].func;
    return(setError(ERR_NOERR));
  }
}

/* restoreRet: **************************************************************
 * restore the return record; 
 ***************************************************************************/
  
restoreRet(fd,rptr)
     int fd;
     MHiRet *rptr;
{
  ReadString(fd,&(rptr->retPattern));
  Read(fd,&(rptr->reqCnt),sizeof(int));
  
  return(OK);
}

/*restoreRequests: **********************************************************
 * the array of requests is restored; the counter is read to determine the 
 * number of actually records to be restored 
 ****************************************************************************/

restoreRequests(fd, aptr, requests)
     int fd;
     ApplicationNode *aptr;
     MHiReq *requests[];
{
  register i;
  int cnt;
  
  Read(fd,&cnt,sizeof(int));
  
  for (i = 0; i < cnt; i++) {
    requests[i] = (MHiReq *) Malloc(sizeof(MHiReq));
    if (! restoreReq(fd, aptr, requests[i])) 
      return(ERROR);
  }
  return(OK);
}

/* restoreReq: **************************************************************
 * restore a request record; the function is stored in the usual manner 
 ****************************************************************************/

restoreReq(fd, aptr, rptr)
     int fd;
     ApplicationNode *aptr;
     MHiReq *rptr;
{
  int indx;
  MDS_CNT_TYPE cntType;
  
  ReadString(fd, &(rptr->reqData));
  ReadString(fd, &(rptr->initialValue));
  ReadString(fd, &(rptr->prelimInitialValue));
  Read(fd, &(rptr->flags), sizeof(int));
  Read(fd, &(rptr->size), sizeof(int));
  Read(fd, &indx, sizeof(int));
  if (indx == NO_FUNCTION)
    rptr->reqFnct = (int (*)()) 0;
  else if (indx < 0) {
    rptr->reqFnct = MDS_functions[-indx].func;
  } else if (indx > MDS_MAXFUNC) {
    return(setError(ERR_VALUE));
  } else 
    rptr->reqFnct = aptr->ApplFunc[indx].func;
  Read(fd, &indx, sizeof(int));
  rptr->cnt = (indx < 0) ? 0 : indx;
  
  Read(fd, &cntType, sizeof(MDS_CNT_TYPE));
  if (Cnttype(cntType) == True)
    rptr->reqCntType = cntType;
  else 
    rptr->reqCntType = MDS_CNT_NULL;
  
  return(OK);
}

/* GetNameFromFile: **********************************************************
 * reads application name from input file
 *****************************************************************************/

char *GetNameFromFile(file)
     char *file;
{
  char *name;
  int fd;
  
  if (access(file, R_OK) == ERROR) {                /* no read permission */
    setError(ERR_READ);
    return ((char *) NULL);
  }
  
  if ((fd = open(file, O_RDONLY)) == ERROR) {       /* cannot open file */
    setError(ERR_READ);
    return ((char *) NULL);
  }
  
  if (readString(fd, &name) == ERROR)
    return((char *) NULL);
  
  return(name);
}

/* getApplIdentifier **********************************************************
 * stores application name and file from applications under directory 'path'
 *****************************************************************************/

getApplIdentifier(path, applList, length)
     char *path;
     applId *applList[];
     int length;
{
  char *rindex();
  struct direct *menufile;
  DIR *dirp;
  char *ending, *applName;
  register cnt = 0;
  
  if ((dirp = opendir(path)) == (DIR *) NULL) {
    printf("Cannot open file: %s ", path);
    return(-1);
  }
  
  for (menufile = readdir(dirp); menufile != (struct direct *) NULL && 
       cnt < length; menufile = readdir(dirp)) {
    if (menufile->d_namlen > strlen(MDS_FILESUFFIX)) {
      if ((ending = rindex(menufile->d_name, '.')) == (char *) NULL)
	continue;
      
      if (strcmp(ending, MDS_FILESUFFIX) == 0) {
	
	if (makeFullName(path, menufile->d_name, fullName, 256) < 0) {
	  setError(ERR_MAX);
	  return(-1);
	}
	
	/* read application name from user file */
	if ((applName = GetNameFromFile(fullName)) == (char *) NULL) {
	  setError(ERR_READ);
	  return(-1);
	}
	applList[cnt] = (applId *) Malloc(sizeof(applId));
	applList[cnt]->name = applName;
	applList[cnt]->file = strCopy(fullName);
	cnt++;
      }     
    }
  }  
  closedir(dirp); 
  
  return(cnt);
}

/* utility functions *********************************************************
 * determine the index of a function in the MDS function array; this index 
 * is dumped onto a file; so that the proper function an be restored when 
 * reading in the file (function pointers should be stored). The field zero of 
 * each array is left free, because in the MenuEditor the indices are stored 
 * as the values of the function pointer and a value of zero would indicate 
 * that no function g]has been provided.
 *****************************************************************************/

whichFunction(aptr, fnct)
     ApplicationNode *aptr;
     int (* fnct)();
{
  register i;
  
  /* first look in the application specific function table */
  for (i = 1; aptr->ApplFunc[i].func != (int(*)()) 0; i++)
    if (aptr->ApplFunc[i].func == fnct) 
      return(i);
  
  /* then look in the global MDS function table */
  for (i = 1; MDS_functions[i].func != (int (*)()) 0; i++) 
    if (fnct == MDS_functions[i].func)
      return(-i);
  
  return(NO_FUNCTION);
}

/* legalAccess: ***************************************************************
 * determine if the access that has been read is reasonable 
 *****************************************************************************/

Boolean legalAccess(acc)
     int acc;
{
  if ((acc == MDS_NOACC) || (acc & (MDS_READACC | MDS_CHNGACC | MDS_TEMP))) {
    setError(ERR_NOERR);
    return(True);
  } else {
    setError(ERR_VALUE);
    return(False);
  }
}

Boolean Mtyp(mtyp)
     MDS_MENU_TYPE mtyp;
{
  if (mtyp == MDS_M_NULL  ||
      mtyp == MDS_M_POPUP ||
      mtyp == MDS_M_PULLD ||
      mtyp == MDS_L_POPUP ||
      mtyp == MDS_L_PULLD) {
    setError(ERR_NOERR);
    return(True);
  } else { 
    setError(ERR_TYPE);
    return(False);
  }
}

Boolean Cnttype(Cnttpe)
     MDS_CNT_TYPE Cnttpe;
{
  if (Cnttpe == MDS_CNT_NULL ||
      Cnttpe == MDS_CNT_EQ   ||
      Cnttpe == MDS_CNT_LT   ||
      Cnttpe == MDS_CNT_LE   ||
      Cnttpe == MDS_CNT_GT   ||
      Cnttpe == MDS_CNT_GE ) {
    setError(ERR_NOERR);
    return(True);
  } else {
    setError(ERR_TYPE);
    return(False);
  }
}

Boolean reprType(rtype)
     MDS_REPR_TYPE rtype;
{
  if (rtype == MDS_RTYP_NULL ||
      rtype == MDS_RTYP_STRING) {
    setError(ERR_NOERR);
    return(True);
  } else {
    setError(ERR_TYPE);
    return(False);
  }
}

Boolean colorType(ctype)
     MDS_COLOR ctype;
{
  if (ctype == MDS_NOCOLOR ||
      ctype == MDS_WHITE ||
      ctype == MDS_BLACK ||
      ctype == MDS_GREY  ||
      ctype == MDS_LIGHTGREY ||
      ctype == MDS_DARKGREY) {
    setError(ERR_NOERR);
    return(True);
  } else {
    setError(ERR_TYPE);
    return(False);
  }
}

#undef NO_FUNCTION
#undef NOEXIST
#undef EXIST

