/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@gmd.de                                            *
 *              freitag@gmd.de                                          *
 *                                                                      *
 ************************************************************************/
#include <stdio.h>
#include <signal.h>
#include <sys/file.h>
#include <errno.h>

#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include <X11/IntrinsicP.h>
#include <X11/ShellP.h>

#include <Xm/DrawingA.h>
#include <Xm/Form.h>
#include <Xm/LabelG.h>

#include <exBasics.h>
#include <exEdit.h>
#include <exLayout.h>
#include <exErrors.h>
#include <exStxt.h>
#include <exMenu.h>
#include <exFkm.h>
#include <exEntry.h>
#include <cursors/assign>
#include <ife.h>
#include "cmds.h" 
#include "fkmedit.h"
#include "version.h"
#include "icons/fked"

static char SccsID[] = "%W%\t%G%";            /* sccs keyword */

static struct FKEditor *fkedit   = (struct FKEditor *) NULL;
static Widget AssignShell, AssignForm;
static Cursor AssignCursor;
static struct FunctionKey *copiedKey = (struct FunctionKey *) NULL;
static Widget FKW, FKS;			/* to display function key set */
static MenuDescr fkmeditMenus[3];
static fkmeditMenuCnt;
static char *m1 = "Please press the";
static char *m2 = "corresponding";
static char *m3 = "function key now !";

static GC normal_gc;
static Display *display;
static nk_cnt = 0;			/* counter for new keys per set */
static char *fkmPath = NULL;		/* directory where to store sets */
static char msg[256];

static void assignKey();
static void redrawAssignNotify();
static void deleteFunctionKey();
static struct FKEditor *createFKeditor();
static void clearFunctionKey();
static void fkmInit();
static void forcedQuit();
static void ife_inform();
static void setSizeHints();

static char *getFkmPath();
/* Quit */
static fkmeditQuit();

/* function key sets */
static fkmeditLoadSet();
static fkmeditNewSet();
static fkmeditSaveSet();
static fkmeditSaveSetAs();
static fkmeditDeleteSet();
static fkmeditShowSet();
static fkmeditHideSet();

/* function keys */
static fkmeditLoadKey();
static fkmeditNewKey();
static fkmeditSaveKey();
static fkmeditSaveKeyAs();
static fkmeditSaveKeyIn();
static fkmeditDeleteKey();

static fkmeditAssignKey();
static fkmeditCopyKey();
static fkmeditPasteKey();

static EditorDescr body_descr = {
  (int (*)()) 0,
  NULL,
  "fk-body",
  NULL, 
  NULL,
  6, 
  6,
  40, 
  0, 
  0,
  ED_MSGWINDOW | ED_HELP
};

main(argc, argv)
     int argc;
     char *argv[];
{
  XEvent event;
  int maxfd, cnt;
  static fd_set readfs[2];
  Widget top;
  int nfds, dpyno;
  XWMHints wm_hints;
  Boolean connected = False;
  int eshErrorHandler();

  if (argc > 1 && strcmp(argv[1], "-c") == 0)
    connected = True;

  top = XtInitialize(NULL, "FKE", (XrmOptionDescRec *) NULL, 0, &argc, argv);

  display = XtDisplay(top);
  XSetErrorHandler(eshErrorHandler);
  XSynchronize(display, True);
  signal(SIGHUP, forcedQuit);
  signal(SIGINT, SIG_IGN);
  signal(SIGQUIT, SIG_IGN);

  /* function key editor functions */

  cnt = 0;

  MH_FUNCTIONS[cnt].func = (int (*)()) 0;
  MH_FUNCTIONS[cnt].funcName = (char *) 0; cnt++;
  MH_FUNCTIONS[cnt].func = fkmeditQuit;
  MH_FUNCTIONS[cnt].funcName = "fkmeditQuit"; cnt++;

  MH_FUNCTIONS[cnt].func = fkmeditLoadSet;
  MH_FUNCTIONS[cnt].funcName = "fkmeditLoadSet"; cnt++;
  MH_FUNCTIONS[cnt].func = fkmeditNewSet;
  MH_FUNCTIONS[cnt].funcName = "fkmeditNewSet"; cnt++;
  MH_FUNCTIONS[cnt].func = fkmeditSaveSet;
  MH_FUNCTIONS[cnt].funcName = "fkmeditSaveSet"; cnt++;
  MH_FUNCTIONS[cnt].func = fkmeditSaveSetAs;
  MH_FUNCTIONS[cnt].funcName = "fkmeditSaveSetAs"; cnt++;
  MH_FUNCTIONS[cnt].func = fkmeditDeleteSet;
  MH_FUNCTIONS[cnt].funcName = "fkmeditDeleteSet"; cnt++;
  MH_FUNCTIONS[cnt].func = fkmeditShowSet;
  MH_FUNCTIONS[cnt].funcName = "fkmeditShowSet"; cnt++;
  MH_FUNCTIONS[cnt].func = fkmeditHideSet;
  MH_FUNCTIONS[cnt].funcName = "fkmeditHideSet"; cnt++;

  MH_FUNCTIONS[cnt].func = fkmeditLoadKey;
  MH_FUNCTIONS[cnt].funcName = "fkmeditLoadKey"; cnt++;
  MH_FUNCTIONS[cnt].func = fkmeditNewKey;
  MH_FUNCTIONS[cnt].funcName = "fkmeditNewKey"; cnt++;
  MH_FUNCTIONS[cnt].func = fkmeditSaveKey;
  MH_FUNCTIONS[cnt].funcName = "fkmeditSaveKey"; cnt++;
  MH_FUNCTIONS[cnt].func = fkmeditSaveKeyAs;
  MH_FUNCTIONS[cnt].funcName = "fkmeditSaveKeyAs"; cnt++;
  MH_FUNCTIONS[cnt].func = fkmeditSaveKeyIn;
  MH_FUNCTIONS[cnt].funcName = "fkmeditSaveKeyIn"; cnt++;
  MH_FUNCTIONS[cnt].func = fkmeditDeleteKey;
  MH_FUNCTIONS[cnt].funcName = "fkmeditDeleteKey"; cnt++;
  MH_FUNCTIONS[cnt].func = fkmeditAssignKey;
  MH_FUNCTIONS[cnt].funcName = "fkmeditAssignKey"; cnt++;
  MH_FUNCTIONS[cnt].func = fkmeditCopyKey;
  MH_FUNCTIONS[cnt].funcName = "fkmeditCopyKey"; cnt++;
  MH_FUNCTIONS[cnt].func = fkmeditPasteKey;
  MH_FUNCTIONS[cnt].funcName = "fkmeditPasteKey"; cnt++;

  MH_FUNCTIONS[cnt].func = (int (*)()) 0;
  MH_FUNCTIONS[cnt].funcName = (char *) 0;

  fkmInit(display);
  fkedit = createFKeditor(top, 50, 570, connected);

  initCmdTables(fkedit->fked_body, CmdTable, EscCmds, ctrlXCmds, fkmCmds);

  setbuf(stdin, NULL);
  setbuf(stdout, NULL);

  dpyno = ConnectionNumber(display);
  maxfd = getdtablesize();
  FD_ZERO(&readfs[0]);
  FD_SET(dpyno, &readfs[0]);
  FD_SET(fileno(stdin), &readfs[0]);

  openFKEditor(fkedit);

  setSizeHints(fkedit);

  fkmeditNewSet(NULL);			/* initial empty set */

  wm_hints.flags = IconPixmapHint | IconPositionHint;
  wm_hints.icon_pixmap = 
    XCreateBitmapFromData(display, XDefaultRootWindow(display),
			  		fked_bits, fked_width, fked_height);
  wm_hints.icon_x = 1060;
  wm_hints.icon_y = 700;

  XSetWMHints(display, XtWindow(fkedit->fked_shell), &wm_hints);

  while (True) {
    bcopy((char *) &readfs[0], (char *) &readfs[1], sizeof(readfs[0])); 

    nfds = select(maxfd, &readfs[1], 
		  (fd_set *) 0, (fd_set *) 0, (struct timeval *) 0); 

    if (nfds < 0 && errno != EINTR) {
      exAlert(top, -1, -1, "Function Key Editor: select error !");
      continue;
    }

    if (FD_ISSET(dpyno, &readfs[1])) {
      while (XtPending()) {
	XtNextEvent(&event);
	XtDispatchEvent(&event);
      }
    } else
      processMessage(fkedit);
  }
}

/* processMessage: ************************************************************
 * process the messages that are sent by the interface editor
 *****************************************************************************/

processMessage(fked)
     struct FKEditor *fked;
{
  int pid, opc, info;
  
  if (exReceiveMsg(fileno(stdin), &pid, &opc, &info) < 0) {
    fkmsg("Error receiving message from interface editor !");
    fkmeditQuit((RequestParm *) NULL);			/* never returns */
  }

  sprintf(msg, "pid=%d, opc=%d, info=%d", pid, opc, info);
  exError("FKE", "message received: ", msg, NULL);

  if (pid != getpid())
    return;
  
  switch (opc) {
  case IFE_OPEN:
    openFKEditor(fked);
    break;
  default:
    break;
  }
  XFlush(display);
}

/* openFKEditor: **************************************************************
 * display the function key editor on the screen
 *****************************************************************************/

openFKEditor(fked)
     struct FKEditor *fked;
{
  if (fked->flags & FK_MAPPED) 			/* raise if already open */
    return;

  fked->flags |= FK_MAPPED;
  XtPopup(fked->fked_shell, XtGrabNone);

  if (fked->flags & FK_CONNECT) 
    exSendMsg(fileno(stdout), getpid(), IFE_OPEN, 0); 
}

/* clearFKeditor: *************************************************************
 * clear the buffers of the function key editor
 *****************************************************************************/

clearFKeditor(fked)
     struct FKEditor *fked;
{
  clearBuffer(fked->nameEntry->entryEditor);	/* clear entry buffer */
  clearBuffer(fked->fked_body);		     	/* clear editor buffer */
}

/* createFKeditor: ************************************************************
 * create an instance of the function key editor at the given location 
 *****************************************************************************/

static struct FKEditor *createFKeditor(parent, x, y, connect)
     Widget parent;
     int x,y;
     Boolean connect;
{
  struct FKEditor *fked;
  int fh1, fh2;
  Arg args[15];
  int cnt;
  char fullName[256];
  static XtTranslations ptt = NULL;

  static char fkm_edit_translations[] = 
    " <Unmap>		: ife_inform() \n\
      <Map>             : ife_inform()";

  static XtActionsRec fkm_edit_actions[] = {
    { "ife_inform", ife_inform }
  };

  XtAddActions(fkm_edit_actions, 1);
  if (! ptt)
    ptt = XtParseTranslationTable(fkm_edit_translations);
  
  fked = (struct FKEditor *) Malloc(sizeof(struct FKEditor));

  fked->currentFile = NULL;
  fked->currentKey = (struct FunctionKey *) 0;
  fked->currentFKcnt = 0;
  
  /* create application shell */
  cnt = 0;
  XtSetArg(args[cnt], XmNx, x); cnt++;
  XtSetArg(args[cnt], XmNy, y); cnt++;
  XtSetArg(args[cnt], XmNtitle, VERSION); cnt++;
  XtSetArg(args[cnt], XmNiconName, "ERGO Function-key Manager"); cnt++;
  XtSetArg(args[cnt], XmNiconX, 1060); cnt++;
  XtSetArg(args[cnt], XmNiconY, 670); cnt++;
  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;  
  fked->fked_shell = XtCreatePopupShell(VERSION, 
			applicationShellWidgetClass, parent, args, cnt);

  XtOverrideTranslations(fked->fked_shell, ptt);

  /* create toplevel widget */
  cnt = 0;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  fked->fked_window = XtCreateManagedWidget("form", xmFormWidgetClass,
					   fked->fked_shell, args, cnt);

  /* create the menubar and load menus */
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;

  fked->menubar = 
    exCreateMenuBar(fked->fked_window, args, cnt, (void (*)()) 0);
		
  if (makeFullName(getMenuDir(), FKMENUS, fullName, sizeof(fullName)) < 0) 
    return((struct FKEditor *) 0);

  fkmeditMenuCnt = 3;
  if ( !initMenus(fked->menubar, fked->fked_window, fullName, 
	                         fkmeditMenus, &fkmeditMenuCnt)) {
    sprintf(msg, "Cannot load menus -- %s --: function key editor aborted!", 
	    							fullName);
    exAlert(parent, 100, 100, msg);
    exit(1);
  }

  /* create form for entry */
  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, fked->menubar->menubar); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
  
  fked->name_form =  XtCreateManagedWidget("name", xmFormWidgetClass,
			fked->fked_window, args, cnt);

  fked->nameEntry = exMakeEntry(fked->name_form, fked->name_buf, 1, 20, 
	  "Name: ??     ", CmdTable, EscCmds, ctrlXCmds, fkmCmds, True, False);


  /* create the message area */
  cnt = 0;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  fked->msgwin = XtCreateManagedWidget("msg", xmDrawingAreaWidgetClass,
			fked->fked_window, args, cnt);
 
  /* create key body editor window */
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, fked->name_form); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNbottomWidget, fked->msgwin); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  fked->fked_body = 
    makeEditor(fked->fked_window, "body", &body_descr, args, cnt);

  /* define height of msgwin */
  fh1 = exFontHeight(getEditorFont(fked->nameEntry->entryEditor));
  fh2 = exFontHeight(getEditorFont(fked->fked_body));

  cnt = 0;
  XtSetArg(args[cnt], XmNheight, 3 * MAX(fh1, fh2) / 2); cnt++;
  XtSetValues(fked->msgwin, args, cnt);
  
  /* provide the message area for both editors */
  editorSetMsg(fked->nameEntry->entryEditor, fked->msgwin);
  editorSetMsg(fked->fked_body, fked->msgwin);

 
  fked->flags = connect ? FK_CONNECT : 0;
  return(fked);
}

/* doEditFunctionKey: ********************************************************
 * put the given key-description into the editors data-structures and make it
 * ready for editing.
 ****************************************************************************/

static doEditFunctionKey(fk)
     struct FunctionKey *fk;
{
  openFKEditor(fkedit);

  if (fk->name)
    editorFillBuffer(fkedit->nameEntry->entryEditor, fk->name, False);
  else
    clearBuffer(fkedit->nameEntry->entryEditor);

  if (fk->body)
    editorFillBuffer(fkedit->fked_body, fk->body, False);
  else
    clearBuffer(fkedit->fked_body);

  setNeedSave(fkedit->nameEntry->entryEditor, False);
  setNeedSave(fkedit->fked_body, False);

  displayKey(fkedit);
}

/* assignKey: ***************************************************************
 * waits for the user to press a function key and assigns this physical key
 * to the active function key 
 ***************************************************************************/
/*ARGSUSED*/
static void assignKey(w, event, params, num_params)
     Widget w;
     XKeyPressedEvent *event;
     String params;
     int *num_params;
{
  int len;
  char s[5];
  KeySym sym;

  /* translate keycode to keyssym */
  sym = XKeycodeToKeysym(event->display, event->keycode, 0);

  if (IsFunctionKey(sym)) {
    fkedit->currentKey->sym = sym;
    exUngrabAll(AssignForm);
    XtPopdown(AssignShell);
    fkedit->flags |= FK_SETCHANGED;
    fkmsg("Function-key assigned !");
  } else {
    len = XLookupString(event, s, sizeof(s), NULL, NULL);
    if (len == 1 && *s == ESC) {
      exUngrabAll(AssignForm);
      XtPopdown(AssignShell);
      fkmsg("No function key assigned !");
    }
    Beep(display);				/* ring the bell */
  }
  displayKey(fkedit);
}

/* destroyFKeditor: ***********************************************************
 * remove the given function key editor instance
 *****************************************************************************/

destroyFKeditor(fked)
     struct FKEditor *fked;
{
  register i;

  XtDestroyWidget(fked->fked_shell);
  free(fked->currentFile);
  fked->currentFile = NULL;
  
  for (i = 0; i < fked->currentFKcnt; i++)
    clearFunctionKey(fked->currentFK[i]);
}

/* displayKey: ****************************************************************
 * display the name of the associated physical key 
 *****************************************************************************/

displayKey(fked)
     struct FKEditor *fked;
{
  char *key;
  char label[100];
  Arg args[2];
  int cnt;
  XmString xstr = NULL;

  key = getFKey(fked->currentKey);
  sprintf(label, "Name: %s", key ? key : "???");

  /* retrieve old string */
  cnt = 0;
  XtSetArg(args[cnt], XmNlabelString, &xstr); cnt++;
  XtGetValues(fked->nameEntry->label, args, cnt);

  if (xstr)					/* free string */
    XmStringFree(xstr);

  cnt = 0;					/* set new name */
  XtSetArg(args[cnt], XmNlabelString, XmStringCreate(label,  
				      XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetValues(fked->nameEntry->label, args, cnt);
}

/* editFunctionKey: ***********************************************************
 * invoke the function key editor to allow the user to edit the given 
 * function key
 *****************************************************************************/

static editFunctionKey(parms)
     RequestParm *parms;
{
  struct FunctionKey *fk;

  if (! fkedit)
    return;

  fk = (struct FunctionKey *) atoi(parms->data);
  free(parms->data);
  fkedit->currentKey = fk;
  
  doEditFunctionKey(fk);
}

/* deleteFunctionKey: *********************************************************
 * delete the specified function key from the current set
 *****************************************************************************/

static void deleteFunctionKey(parms)
     RequestParm *parms;
{
  register struct FunctionKey *fk;
  register i;
  char *name;

  if (! fkedit)
    return;

  fk = (struct FunctionKey *) atol(parms->data);
  free(parms->data);

  name = strCopy(fk->name);
  clearFunctionKey(fk);				/* clear the fk structure */

  for (i = 0; i < FK_MAXKEYS; i++)
    if (fkedit->currentFK[i] == fk) {		/* adjust key array */
      bcopy((char *) &fkedit->currentFK[i + 1], 
	     (char *) &fkedit->currentFK[i],
	     (FK_MAXKEYS - i - 1) * sizeof(fkedit->currentFK[0]));
      fkedit->currentFKcnt--;
      if (fkedit->currentKey == fk)	{	/* current key deleted */
	clearFKeditor(fkedit);			/* clear editor */

	/* if the current function-key has been deleted, try to
	   identify a new current key: if existing the next key is 
	   used, otherwise the previous one. 
	*/

	if (fkedit->currentFKcnt <= 0)		/* no more keys */
	  fkedit->currentKey =  (struct FunctionKey *) NULL;
	else if (i < fkedit->currentFKcnt)
	  fkedit->currentKey = fkedit->currentFK[i];
	else 
	  fkedit->currentKey = fkedit->currentFK[fkedit->currentFKcnt - 1];
      }
      
      if (fkedit->currentKey)
	doEditFunctionKey(fkedit->currentKey);

      sprintf(msg, "key -- %s -- deleted !", name);
      fkmsg(msg);
      free(fk);
      free(name);
      return;
    }

  fkmsg("Internal error: cannot locate function key !");
}

/* generateFKMenu: ************************************************************
 * generate a menu of function keys in the current set and let the user select
 *****************************************************************************/

static MHmenu *generateFKMenu(fks, cnt, funPtr)
     struct FunctionKey **fks;
     int cnt;
     int (* funPtr)();
{
  register i;
  register MHmenu *mptr;
  register MHitem *iptr;
  register MHiReq *req;
  char s[20];

  setWorkApplication();
  
  if ((mptr = createMenu()) == (MHmenu *) 0) {		/* create a new menu */
    setError(ERR_CREATE);
    return((MHmenu *) 0);
  }
  
  initMenu(mptr, MDS_M_POPUP, "function keys");
  insertMenu(mptr);
  setSelCnt(mptr, 1);
  setSelType(mptr, MDS_CNT_EQ);
  setMenuPos(mptr, 0, 0, MDS_POS_ABS);
  mptr->m_hlp->hlpmsg = 
    strCopy("Please select one of the function key descriptions !");
  
  for (i = 0; i < cnt; i++) {
    iptr = newItem(mptr, fks[i]->name ? fks[i]->name : "??", EOS);
    iptr->i_return->retPattern = strCopy("$");
    req = newRequest();
    setReqFnct(req, funPtr);
    sprintf(s, "%ld", (long) fks[i]);
    setRequestData(req, s);
    insertRequest(iptr, req, 0);
  }
  
  return(mptr);
}

/* fkmsg: ********************************************************************
 * forward a message
 ****************************************************************************/

fkmsg(str) 
     char *str;
{
  if (fkedit)
    editorMsg(fkedit->fked_body, str);
  else
    exNotifyUser(fkedit->fked_shell, -1, -1, str);
}

/* closeSet: *****************************************************************
 * close the file containing the description of the current set of function 
 * keys;
 * the user is informed if any changed function key descriptions have not been 
 * saved.
 *****************************************************************************/

/*ARGSUSED*/
static void closeSet(fked)
     struct FKEditor *fked;
{
  int i;
  int answer;

  if (needSave(fked->nameEntry->entryEditor) || needSave(fked->fked_body)) {
    answer = exNotify2User(fked->fked_shell, -1, -1, "Save key first ?");

    switch (answer) {
    case ANSWERYES:				/* save key first */
      fkmeditSaveKey(NULL);
      break;
    case ANSWERCANCEL:				/* cancel current function */
      return;
    default:					/* do not save key */
      break;
    }
  }

  if (fked->flags & FK_SETCHANGED) {
    answer = exNotify2User(fked->fked_shell, -1, -1, "Save set first ?");

    switch (answer) {
    case ANSWERYES:				/* save set first */
      fkmeditSaveSet(NULL);
      break;
    case ANSWERCANCEL:				/* cancel current function */
      return;
    default:					/* do not save set */
      break;
    }

    fked->flags &= ~FK_SETCHANGED;
  }
      
  for (i = 0; i < fked->currentFKcnt; i++) {
    clearFunctionKey(fked->currentFK[i]);
    fked->currentFK[i] = (struct FunctionKey *) NULL;
  }

  bzero((char *) fked->currentFK, FK_MAXKEYS * sizeof(struct FunctionKey *));
  
  free(fked->currentFile);
  fked->currentFile = NULL;
  fkedit->currentKey = NULL;
  fked->currentFKcnt = 0;

  clearFKeditor(fked);

  nk_cnt = 0;
  return;
}

/* saveSet: *******************************************************************
 * save the current function key set into the given file
 *****************************************************************************/

saveSet(fks, cnt, file)
     struct FunctionKey *fks[];
     int cnt;
     char *file;
{
  int fd, i;
  char fullName[256];
  
  if (*file != '/')			/* make a full-qualified name */
    makeFullName(fkmPath, file, fullName, sizeof(fullName));
  else
    strcpy(fullName, file);

  if ((fd = open(fullName, O_WRONLY | O_CREAT, 0662)) < 0) {
    sprintf(msg, "Cannot open file -- %s -- !", fullName);
    fkmsg(msg);
    return(-1);
  }
  
  write(fd, &cnt, sizeof(int));
  
  for (i = 0; i < cnt; i++) {
    write(fd, &(fks[i]->sym), sizeof(KeySym));
    writeString(fd, fks[i]->name);
    writeString(fd, fks[i]->body);
  }
  
  close(fd);
  
  sprintf(msg, "Current set saved into file -- %s -- !", fullName);
  fkmsg(msg);
  
  nk_cnt = 0;

  return(0);
}

/*ARGSUSED*/
static void ife_inform(w, event, params, num_params)
     Widget w;
     XExposeEvent *event;
     String *params;
     Cardinal *num_params;
{
  if (! (fkedit->flags & FK_CONNECT))
    return;

  switch (event->type) {
  case MapNotify:
    exSendMsg(fileno(stdout), getpid(), IFE_OPEN, 0); 
    fkedit->flags |= FK_MAPPED;
    break;
  case UnmapNotify:
    exSendMsg(fileno(stdout), getpid(), IFE_CLOSE, 0); 
    fkedit->flags &= ~FK_MAPPED;
    break;
  }
}

/******************************************************************************
 *
 *                          Quit function key editor 
 *
 *****************************************************************************/

/* fkmeditQuit: ***************************************************************
 * quit the editor, unmap and destroy its windows, free the data structures
 *****************************************************************************/
/*ARGSUSED*/
static fkmeditQuit(parms)
     RequestParm *parms;
{
  Boolean connected = False; 
  if (fkedit) {
    connected = fkedit->flags & FK_CONNECT;
    closeSet(fkedit);
    destroyFKeditor(fkedit);
  }
  
  free(fkmPath);

  if (connected) {
    exSendMsg(fileno(stdout), getpid(), IFE_INFO, IFE_STOP); 
    sleep(2);		/* make shure, that ifedit has received the message */
  }

  XCloseDisplay(display);
  exit(0);
}

/******************************************************************************
 *
 *                         function key sets
 *
 *****************************************************************************/

/* fkmeditLoadSet: ************************************************************
 * open a file containing the description of a set of function keys and load 
 * the first function key (if any) or create a new one; all function
 * key sets are in the directory FKPATH; if no such environment variable 
 * is set the user is prompted for a file name.
 *****************************************************************************/

/*ARGSUSED*/
static fkmeditLoadSet(parms)
     RequestParm *parms;
{
  char *file;
  char buf[10];
  int fd, cnt, i;
  struct FunctionKey **fks;
  RequestParm iparms;

  closeSet(fkedit);

  file = getFKSetFile(parms->parent);

  if (! file || ! *file)
    return(0);

  if (! fileExists(file)) {
    sprintf(msg, " File does not exist -- %s --: will be created !", file);
    fkmsg(msg);
    
    free(fkedit->currentFile);
    fkedit->currentFile = strCopy(file);
    fkedit->currentFKcnt = 0;
    free(file);
    return(0);
  }
  
  if ((fd = open(file, O_RDONLY)) < 0) {
    sprintf(msg, "Cannot open file -- %s -- !", file);
    fkmsg(msg);
    free(file);
    return(-1);
  }

  read(fd, &cnt, sizeof(int));
  if (cnt > FK_MAXKEYS || cnt < 0) {
    sprintf(msg, "Illegal number of keys in set (%d) !", cnt);
    fkmsg(msg);
    close(fd);
    return(-1);
  }

  free(fkedit->currentFile);
  fkedit->currentFile = NULL;
  for (i = 0; i < fkedit->currentFKcnt; i++)
    clearFunctionKey(fkedit->currentFK[i]);

  fks = fkedit->currentFK;

  for (i = 0; i < cnt; i++) {
    fks[i] = (struct FunctionKey *) Malloc(sizeof(struct FunctionKey));
    read(fd, &(fks[i]->sym), sizeof(KeySym));
    ReadString(fd, &(fks[i]->name));
    ReadString(fd, &(fks[i]->body));
  }

  close(fd);

  fkedit->currentFile = strCopy(file);
  fkedit->currentFKcnt = cnt;

  free(file);

  if (cnt > 0) {				/* load first key */
    sprintf(buf, "%ld", fkedit->currentFK[0]);
    iparms.data = strCopy(buf);
    editFunctionKey(&iparms);
    sprintf(msg,"Read -- %d -- function keys from file -- %s -- !", cnt, file);
    fkmsg(msg);
  } else if (fkmeditNewKey(&iparms) < 0)	/* create a new key */
    return(-1);

  fkedit->flags &= ~FK_NEWSET;
  fkedit->flags &= ~FK_NEWKEY;
  fkedit->flags &= ~FK_SETCHANGED;
  
  return(0);
}

/* fkmeditNewSet: *************************************************************
 * create a new function key set
 *****************************************************************************/
/*ARGSUSED*/
static fkmeditNewSet(parms)
     RequestParm *parms;
{
  closeSet(fkedit);
  
  fkedit->currentFile = strCopy("no_name");
  
  fkmeditNewKey((RequestParm *) NULL);

  fkedit->flags |= FK_NEWSET;
}

/* fkmeditSaveSet: ************************************************************
 * save the current function set into the current file 
 *****************************************************************************/

static fkmeditSaveSet(parms)
     RequestParm *parms;
{
  int rc;

  if (fkedit->flags & FK_NEWSET)
    return(fkmeditSaveSetAs(parms));

  rc = saveSet(fkedit->currentFK, fkedit->currentFKcnt, fkedit->currentFile);

  if (rc > -1) 
    fkedit->flags &= ~FK_SETCHANGED;

  return(rc);
}

/* fkmeditSaveSetAs: **********************************************************
 * save the current set into a file that the user is prompted for;
 * the file 
 *****************************************************************************/

/*ARGSUSED*/
static fkmeditSaveSetAs(parms)
     RequestParm *parms;
{
  char *file;
  int rc;

  file = exGetEntry(fkedit->fked_shell, -1, -1, "Name of set: ", NULL, 40);

  if (! file || ! *file)
    return(0);

  rc = saveSet(fkedit->currentFK, fkedit->currentFKcnt, file);

  if (rc > -1) {
    free(fkedit->currentFile);
    fkedit->currentFile = file;
    fkedit->flags &= ~FK_SETCHANGED;
  }

  return(0);
}

/* fkmeditDeleteSet: *********************************************************
 * remove the whole set 
 *****************************************************************************/
/*ARGSUSED*/
static fkmeditDeleteSet(parms)
     RequestParm *parms;
{
  char *file;

  file = getFKSetFile(parms->parent);	/* select a set */

  if (! file || ! *file)
    return;

  if (! fileExists(file)) {		/* file does not exist */
    sprintf(msg, "Set -- %s -- cannot be located: not removed !", file);
    fkmsg(msg);
    return;
  }
  
  if (unlink(file) < 0) {		/* remove file */
    sprintf(msg, "Set -- %s -- cannot be unlinked (%d) !", file, errno);
    fkmsg(msg);
    return;
  }
}

/* fkmeditShowSet: ************************************************************
 * show the function set as it would appear later on the screen 
 *****************************************************************************/

/*ARGSUSED*/
static fkmeditShowSet(parms)
     RequestParm *parms;
{
  Arg args[5];
  int cnt;

  if (FKS) {
    XtPopup(FKS, XtGrabNonexclusive);
    return;
  }

  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, 1); cnt++;
  XtSetArg(args[cnt], XmNwidth, 1); cnt++;
  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;

  FKS = XtCreatePopupShell("fked show shell", 
		    transientShellWidgetClass, fkedit->fked_shell, args, cnt);
  FKW = XtCreateManagedWidget(NULL, xmFormWidgetClass, FKS, args, cnt);

  displayFKSet(FKW, fkedit->currentFK, fkedit->currentFKcnt, False);

}
  
/* fkmeditHideSet: ************************************************************
 * hide the current function set
 *****************************************************************************/

/*ARGSUSED*/
static fkmeditHideSet(parms)
     RequestParm *parms;
{
  if (FKS)
    XtDestroyWidget(FKS);

  FKS = (Widget) NULL;
}

/******************************************************************************
 *
 *				function keys
 *
 *****************************************************************************/

/* fkmeditLoadKey: ************************************************************
 * load the a function key definition of the current set; for this purpose 
 * display a menu of all current function key descriptons and let the user 
 * select.
 *****************************************************************************/

/*ARGSUSED*/
static fkmeditLoadKey(parms)
     RequestParm *parms;
{
  exMenu *fkList;
  int x, y;
  static editFunctionKey();
  MHmenu *fkMenu;

  getMouse_xy(display, &x, &y);

  if (fkedit->currentFKcnt < 1) {
    fkmsg("No keys in current set !");
    return(-1);
  }

  if (! (fkMenu = generateFKMenu(fkedit->currentFK,
				 fkedit->currentFKcnt, editFunctionKey)))
    return(-1);

  setMenuPos(fkMenu, x, y, MDS_POS_ABS);

  if (! (fkList = exMakeMenu(NULL, NULL, fkedit->fked_shell, fkMenu))) {
    fkmsg("Internal error: canot create menu !");
    return(0);
  }

  exCallMenu(fkList);
  freeMenu(fkList);

  return(0);
}

/* fkmeditNewKey: *************************************************************
 * allow the user to enter a new key definition into an already opened set; if
 * no current set is defined, it is created.
 *****************************************************************************/
/*ARGSUSED*/
static fkmeditNewKey(parms)
     RequestParm *parms;
{
  register struct FunctionKey *fk;
  char buf[20];
  int answer;

  sprintf(buf,"no name [%d]", nk_cnt++);

  if (needSave(fkedit->nameEntry->entryEditor) || needSave(fkedit->fked_body)){
    answer = exNotify2User(fkedit->fked_shell, -1, -1, "Save key first ?");

    switch (answer) {
    case ANSWERYES:				/* save key first */
      fkmeditSaveKey(NULL);
      break;
    case ANSWERCANCEL:				/* cancel current function */
      return(-1);
    default:					/* do not save key */
      break;
    }
  }
  
  if (fkedit->currentFKcnt >= FK_MAXKEYS) {
    fkmsg("Max. number of keys already in set !");
    return(-1);
  }
  
  fk = (struct FunctionKey *) Malloc(sizeof(struct FunctionKey));
  fkedit->currentFK[fkedit->currentFKcnt++] = fk;
  fkedit->currentKey = fk;

  fk->sym = 0;
  fk->name = strCopy(buf);
  fk->body = "";
  
  clearFKeditor(fkedit);
  fkedit->flags |= FK_NEWKEY;

  doEditFunctionKey(fk);

  fkmsg("New key created !");
  
  return(0);
}

/* fkmeditSaveKey: ************************************************************
 * save the current function key description; i.e. store the editor buffers 
 * into the corresponding structure components
 *****************************************************************************/
 
/*ARGSUSED*/
static fkmeditSaveKey(parms)
     RequestParm *parms;
{
  register struct FunctionKey *fk;
  int pos;
  
  if (! fkedit || ! fkedit->currentKey) {
    fkmsg("No key to save !");
    return(-1);
  }

  fk = fkedit->currentKey;

  free(fk->name);
  fk->name = NULL;
  free(fk->body);
  fk->body = NULL;

  /* provide a name, if its a new key and we haven't edited the name buffer */
  if (fkedit->flags & FK_NEWKEY && ! needSave(fkedit->nameEntry->entryEditor)) 
    return (fkmeditSaveKeyAs(NULL));

  fk->name = getBufferContents(fkedit->nameEntry->entryEditor);

  if (fk->name && *fk->name)
    fk->name[strlen(fk->name) - 1] = EOS;
  else 					/* no name provided */
    return (fkmeditSaveKeyAs(NULL));

  fk->body = getBufferContents(fkedit->fked_body);
/*
  if (fk->body && *fk->body)
    fk->body[strlen(fk->body) - 1] = EOS;
*/
  if (! fk->flags & KEY_EXECUTE) {
    if (fk->body && *fk->body) {
      pos = strlen(fk->body) - 1;
      while (pos > 0 && (fk->body[pos] == '\n' || fk->body[pos] == '\r'))
	fk->body[pos--] = EOS;
    }
  }
  
  /* indicate that the key has been saved */
  fkedit->flags &= ~FK_NEWKEY;
  setNeedSave(fkedit->nameEntry->entryEditor, False);
  setNeedSave(fkedit->fked_body, False);

  if (! *fk->name || ! *fk->body)
    fkmsg("Warning: no name or body provided !");
  else {
    sprintf(msg, "Function key -- %s -- saved !", fk->name);
    fkmsg(msg);
  }
    
  fkedit->flags |=  FK_SETCHANGED;

  return(0);
}

/* fkmeditSaveKeyAs: **********************************************************
 * save the current function key description with a new name
 *****************************************************************************/
 
/*ARGSUSED*/
static fkmeditSaveKeyAs(parms)
     RequestParm *parms;
{
  register struct FunctionKey *fk;
  int pos;
  char *name;

  if (! fkedit || ! fkedit->currentKey) {
    fkmsg("No current function key to save !");
    return;
  }

  fk = fkedit->currentKey;

  name = exGetEntry(fkedit->fked_shell, -1, -1, "Key name: ", NULL, 15);

  if (! name || ! *name)
    return;

  editorFillBuffer(fkedit->nameEntry->entryEditor, name, False);

  free(fk->name);
  free(fk->body);

  fk->name = name;
  
  fk->body = getBufferContents(fkedit->fked_body);
  if (fk->body && *fk->body)
    fk->body[strlen(fk->body) - 1] = EOS;

  if (! fk->flags & KEY_EXECUTE) {
    if (fk->body && *fk->body) {
      pos = strlen(fk->body) - 1;
      while (pos > 0 && (fk->body[pos] == '\n' || fk->body[pos] == '\r'))
	fk->body[pos--] = EOS;
    }
  }

  if (! *fk->name || ! *fk->body)
    fkmsg("Warning: no name or body provided !");
  else {
    sprintf(msg, "Function key -- %s -- saved !", fk->name);
    fkmsg(msg);

    /* indicate that the key has been saved */
    fkedit->flags &= ~FK_NEWKEY;
    setNeedSave(fkedit->nameEntry->entryEditor, False);
    setNeedSave(fkedit->fked_body, False);
    fkedit->flags |=  FK_SETCHANGED;
  }
}

/* fkmeditSaveKeyIn: **********************************************************
 * save the current function key description into a different set
 *****************************************************************************/
 
/*ARGSUSED*/
static fkmeditSaveKeyIn(parms)
     RequestParm *parms;
{
  register struct FunctionKey *fk;
  
  if (! fkedit || ! fkedit->currentKey) {
    fkmsg("No current function key to save !");
    return(-1);
  }

  /* temporarily save current function key */
  fk = (struct FunctionKey *) Malloc(sizeof(struct FunctionKey));
  fk->flags = fkedit->currentKey->flags;
  fk->sym = fkedit->currentKey->sym;
  fk->name = strCopy(fkedit->currentKey->name);
  fk->body = strCopy(fkedit->currentKey->body);

  fkmeditLoadSet(parms);		/* load new set */
  fkmeditNewKey(parms);
  doEditFunctionKey(fk);

  free(fk->name);
  free(fk->body);
  free(fk);

  /* indicate that the key has been saved */
  setNeedSave(fkedit->nameEntry->entryEditor, False);
  setNeedSave(fkedit->fked_body, False);

  fkedit->flags &= ~FK_NEWKEY;
  fkedit->flags |=  FK_SETCHANGED;
  
  return(0);
}

/* fkmeditDeleteKey: **********************************************************
 * deletes a function key
 *****************************************************************************/

/*ARGSUSED*/
static fkmeditDeleteKey(parms)
     RequestParm *parms;
{
  exMenu *fkList;
  int x, y;
  MHmenu *fkMenu;

  if (fkedit->currentFKcnt <= 0) {
    fkmsg("No keys in set !");
    return(-1);
  }

  if (! (fkMenu = generateFKMenu(fkedit->currentFK, fkedit->currentFKcnt,
				 (int (* )())deleteFunctionKey)))
    return(-1);

  getMouse_xy(display, &x, &y);
  setMenuPos(fkMenu, x, y, MDS_POS_ABS);

  fkList = exMakeMenu(NULL, NULL, fkedit->fked_shell, fkMenu);
  exCallMenu(fkList);
  freeMenu(fkList);

  fkedit->flags |= FK_SETCHANGED;

  return(0);
}

/* fkmeditAssignKey: **********************************************************
 * assign a physical function key with the current function key description
 *****************************************************************************/

/*ARGSUSED*/
static fkmeditAssignKey(parms)
     RequestParm *parms;
{
  int x, y, w, h, cnt;
  Arg args[10];
  static XFontStruct *f = NULL;
  XEvent event;

  static char assign_translations[] = 
    " <KeyPress>: assignKey() 		\n \
      <Expose> 	: redrawAssignNotify()";

  static XtActionsRec assign_actions[] = {
    { "assignKey",   assignKey },
    { "redrawAssignNotify",  redrawAssignNotify }
  };
  
  static XtTranslations ptt = NULL;

  if (! fkedit || ! fkedit->currentKey)
    return;

  if (! ptt) {
    XtAddActions(assign_actions, 2);
    ptt = XtParseTranslationTable(assign_translations);
  }

  if (! f)
    f = exBoldFont(display);

  w = (strlen(m3) + 4) * exFontWidth(f);
  h = 5 * exFontHeight(f);

  getMouse_xy(display, &x, &y);
  
  if (! AssignShell) {
    cnt = 0;
    XtSetArg(args[cnt], XmNx, x + 20); cnt++;
    XtSetArg(args[cnt], XmNy, y + 20); cnt++;
    XtSetArg(args[cnt], XmNwidth, w); cnt++;
    XtSetArg(args[cnt], XmNheight, h); cnt++;
    XtSetArg(args[cnt], XmNborderWidth, 5); cnt++;
    XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;
    
    AssignShell = XtCreatePopupShell("fked assign shell", 
		    transientShellWidgetClass, fkedit->fked_shell, args, cnt);
    cnt = 0;
    XtSetArg(args[cnt], XmNuserData, fkedit); cnt++;
    XtSetArg(args[cnt], XmNwidth, w); cnt++;
    XtSetArg(args[cnt], XmNheight, h); cnt++;
    
    AssignForm = XtCreateManagedWidget("fked assign", 
		    xmDrawingAreaWidgetClass, AssignShell, args, cnt);
    
    XtOverrideTranslations(AssignForm, ptt);
  }

  XtPopup(AssignShell, XtGrabExclusive);

  while (((TransientShellWidget) AssignShell)->shell.popped_up) {
    XtNextEvent(&event);
    XtDispatchEvent(&event);
  }
}

/******************************************************************************
 *
 *				edit-menu 
 *
 *****************************************************************************/

/* fkmeditCopyKey: ************************************************************
 * put the current function key description into the copy buffer
 *****************************************************************************/

/*ARGSUSED*/
static fkmeditCopyKey(parms)
     RequestParm *parms;
{
  if (! fkedit || ! fkedit->currentKey) {
    fkmsg("Nothing to copy !");
    return;
  }

  if (! copiedKey)
    copiedKey = (struct FunctionKey *) Malloc(sizeof(struct FunctionKey));

  copiedKey->flags = fkedit->currentKey->flags;
  copiedKey->sym = fkedit->currentKey->sym;
  free(copiedKey->name);
  free(copiedKey->body);
  copiedKey->name = strCopy(fkedit->currentKey->name);
  copiedKey->body = strCopy(fkedit->currentKey->body);

  sprintf(msg, "Key copied -- %s -- !", fkedit->currentKey->name);
  fkmsg(msg);
}

/* fkmeditPasteKey: ***********************************************************
 * paste the function key in the copy buffer into the function key editor 
 *****************************************************************************/

/*ARGSUSED*/
static fkmeditPasteKey(parms)
     RequestParm *parms;
{
  RequestParm rParm;

  if (! copiedKey) {
    fkmsg("Nothing to paste !");
    return;
  }

  if (! fkedit || ! fkedit->currentKey) {
    fkmsg("No current key !");
    return;
  }

  fkedit->currentKey->flags = copiedKey->flags;
  fkedit->currentKey->sym = copiedKey->sym;
  free(fkedit->currentKey->name);
  fkedit->currentKey->name = strCopy(copiedKey->name);
  free(fkedit->currentKey->body);
  fkedit->currentKey->body = strCopy(copiedKey->body);
  sprintf(msg, "%d", (int) fkedit->currentKey);
  rParm.data = strCopy(msg);
  
  setNeedSave(fkedit->fked_body, True);
  setNeedSave(fkedit->nameEntry->entryEditor, True);

  editFunctionKey(&rParm);
}

/* redrawAssignNotify: ********************************************************
 * redraw the assign notifier window
 *****************************************************************************/

/*ARGSUSED*/
static void redrawAssignNotify(wd, event, params, num_params)
     Widget wd;
     XExposeEvent *event;
     String params;
     int *num_params;
{
  static XFontStruct *f = NULL;
  int x, y, w, h;

  if (! f)
    f = exBoldFont(display);

  w = (strlen(m3) + 4) * exFontWidth(f);

  x = (w - exTextWidth(f, m1)) / 2;
  y = exFontHeight(f) + f->max_bounds.ascent; 
  h = 5 * exFontHeight(f);

  XSetFont(display, normal_gc, f->fid);
  XDrawImageString(display, XtWindow(wd), normal_gc, x, y, m1, strlen(m1));
  y += exFontHeight(f); 
  x = (w - exTextWidth(f, m2)) / 2;
  XDrawImageString(display, XtWindow(wd), normal_gc, x, y, m2, strlen(m2));
  y += exFontHeight(f); 
  x = (w - exTextWidth(f, m3)) / 2;
  XDrawImageString(display, XtWindow(wd), normal_gc, x, y, m3, strlen(m3));

  exGrabAll(AssignForm, AssignCursor); 

  XWarpPointer(display, None, XtWindow(AssignForm), 0, 0, 0, 0, w /2, h / 2);
}

/* clearFunctionKey: ********************************************************
 * utility function: free allocated strorage 
 ***************************************************************************/

static void clearFunctionKey(fk)
     struct FunctionKey *fk;
{
  free((fk->name));
  free((fk->body));
}

/* fkmInit: ******************************************************************
 * intilalize internally used variables 
 ****************************************************************************/

static void fkmInit(d)
     Display *d;
{
  Window w;
  Pixmap pm;
  XColor fg, bg;

  fkmPath = strCopy(getFkmPath());

  w = XDefaultRootWindow(d);

  exInitLayout(d);
  exInitMenus(d);
  exInitFKM(d);

  pm = XCreateBitmapFromData (d, w, assign_bits, assign_width, assign_height);
  AssignCursor = XCreatePixmapCursor(d, pm, pm, &fg, &bg, 1, 1);
  /*
  AssignCursor = XCreateFontCursor(d, XC_box_spiral);
  */
  normal_gc = XCreateGC(d, w, 0, 0);
  XSetBackground(d, normal_gc, exWhitePixel(d));
  XSetForeground(d, normal_gc, exBlackPixel(d));
}

/* forcedQuit: ***************************************************************
 * this function is called upon receipt of a SIGHUP signal from the eshell in
 * order to terminate; 
 *****************************************************************************/

static void forcedQuit()
{
  RequestParm parms;
  
  parms.parent = fkedit->fked_shell;

  fkmeditQuit(&parms);
}

/* switchBuffer: ************************************************************
 * warp the mouse pointer into the opposite editor window 
 ***************************************************************************/

static switchBuffer(editor)
     Editor editor;
{
  BufInfo *bufinfoptr();

  if (editor == fkedit->nameEntry->entryEditor)
    XWarpPointer(display, None, XtWindow(bufinfoptr(fkedit->fked_body)->edt), 
		 0, 0, 0, 0, 0, 0);
  else if (editor == fkedit->fked_body)
    XWarpPointer(display, None, 
		 XtWindow(bufinfoptr(fkedit->nameEntry->entryEditor)->edt), 
		 0, 0, 0, 0, 0, 0);
}

/*ARGSUSED*/
int eshErrorHandler(dsp, event)
     Display *dsp;
     XErrorEvent *event;
{
  kill(getpid(), SIGQUIT);
}

/* getFkmPath: ****************************************************************
 * return the path where the user stores the function key sets. This should 
 * normally be the value of the environment variable "FKPATH"; if not set 
 * the valueof the variable "HOME" is used; if both variables are not defined,
 * the current directory is used.
 *****************************************************************************/

static char *getFkmPath()
{
  char *ptr;
  char *getenv();

  if (ptr = getenv("FKPATH"))
    return(ptr);

  exError("FKE", "environment variable FKPATH not set.", NULL);
    
  if (ptr = getenv("HOME"))
    return(ptr);
  
  exError("FKE", "environment variable HOME not set.", NULL);
  exError("FKE", "storing in current directory.", NULL);

  return(NULL);
}


/*****************************************************************************
 *
 *			editor functions 
 *
 ****************************************************************************/

/* saveFK: *******************************************************************
 * save the current function key in the current set 
 ****************************************************************************/

static saveFK(editor)
     Editor editor;
{
  fkmeditSaveKey(NULL);
}

/* writeFK: *****************************************************************
 * save the current function key with a new name in the current set 
 ****************************************************************************/

static writeFK(editor)
     Editor editor;
{
  fkmeditSaveKeyAs(NULL);
}

/* findFK: *******************************************************************
 * load a function key from the current set
 ****************************************************************************/

static findFK(editor)
     Editor editor;
{
  fkmeditLoadKey(NULL);
}

/* insertFK: ****************************************************************
 * insert the contents of a file in the current function key body
 ****************************************************************************/

static insertFK(editor)
     Editor editor;
{
  insertFile(fkedit->fked_body);
}

/* setSizeHints: ************************************************************
 * retrieve values from the various subcomponents of the macro editor and
 * use them to give size hints to the window manager 
 ****************************************************************************/

static void setSizeHints(fked)
     struct FKEditor *fked;
{
  XSizeHints size_hints;
  Dimension mb_height, name_height, msg_height;
  XFontStruct *f; 
  int cnt;
  Arg args[4];

  f = getEditorFont(fked->fked_body);
  
  name_height = 0;
  cnt = 0;
  XtSetArg(args[cnt], XmNheight, &name_height); cnt++;
  XtGetValues(fked->name_form, args, cnt);
  
  mb_height    = menuBarHeight(fked->menubar);
  msg_height   = msgHeight(fked->fked_body);

  size_hints.flags = PMinSize | PResizeInc;
  
  size_hints.min_width  = 20 * exFontWidth(f);
  size_hints.min_height = mb_height + name_height + msg_height + 
    				   LINEHEIGHT(f) + FONTGAP;

  size_hints.width_inc  = exFontWidth(f);
  size_hints.height_inc = LINEHEIGHT(f);

  XSetNormalHints(display, XtWindow(fked->fked_shell), &size_hints);
}
