/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/*****************************************************************************
 * This file contains all functions related to special subcomponents in the
 * menu editor (activity mask editor, previewer, help form, request form);
 ****************************************************************************/

#include "editDescr.h"
#include <signal.h>
#include <sys/file.h>

/*#include <X/Xkeyboard.h>
#include <stxt.h>
#include <cursors/dot>
#include <cursors/dotMask>*/

#include <X11/Shell.h>
#include <Xm/Form.h>
#include <Xm/LabelG.h>
#include <Xm/PushBG.h>
#include <Xm/ToggleBG.h>
#include <Xm/RowColumn.h>
#include <Xm/DrawingA.h>

static char SccsID[] = "@(#)subComps.c	1.2\t8/9/93";            /* sccs keyword */

static actEditInit();

int INTERNALS = 0;                 /* number of internal menus created
				      by previewer in '@MenuEditorInternal@' */

/***************** request form declarations **********************/

static struct _reqForm {
  Widget reqFormShell;              /* popup shell for request form */
  Widget form;                      /* form widget */
  struct _exMenubar *menuBar;       /* menu bar of request editor */
  exMenu *menup;                    /* run-time structure for permanent menu */
  Widget functp;                    /* panel for permanent menu */
  Widget titleReq,                  /* label holding current request index */
         titleItem;                 /* label holding current item name */
  Widget quitButton, saveButton;
  Widget reqLabel;                  /* corresponding item names */
  MHmenu *mptr;                     /* current menu */
  MHitem *iptr;                     /* current item */
  int reqIndex;                     /* current request index of iptr*/
  MHiReq *req;                      /* current request */
} reqForm;

static struct _reqFunctForm {
  Widget reqFunctShell;             /* popup shell for request form */
  Widget form;                      /* form widget */
  Widget entryForm;                 /* form widget for entries */
  Widget title;                     /* label holding current request index */
  Widget quitButton, saveButton;
  struct _formEntry *size; 
  struct _formEntry *data;  
  struct _formEntry *count; 
  struct _formEntry *initval;
  Boolean needSave;
} reqFunctForm;

void quitReqEdit();
static initReqForm1();
static void editReqForm2();
static void editReqViaFunct();
static void editReqViaWnd();
static void editReqViaMenu();
static void saveReqViaWnd(), quitReqViaWnd(); 
static void saveReqViaMenu(), quitReqViaMenu(); 
static void saveReqViaFunct(), quitReqViaFunct();  
static void getRequestResult();
static void resetFuncPanel();

static void exReqExposeProc();

static exMenu *menu1 = NULL, *menu2 = NULL, *menu3 = NULL;

/***************** activity mask editor declarations **********************/

typedef struct _actIndex {
  int x;
  int y;
} actIndex;

static box_width = 17;
static Cursor dotCursor = (Cursor) NULL;
static struct _actIndex *butIndex[ME_MAXITEM * (ME_MAXITEM + 1)];

static GC act_gc = NULL;

static struct _actEditor {
  Widget actEditShell;          /* popup shell for activity mask component */
  Widget actEditWin;            /* form widget */
  Widget actEditWork;           /* rowcolumn widget with toggle buttons */
  Widget quitButton, saveButton;
  Widget title1, title2;
  Widget actLabel[ME_MAXITEM + 1];            /* corresponding item names */
  Widget button[ME_MAXITEM + 1][ME_MAXITEM]; /* toggle buttons for item 
						  masks */
  int **activityMatrix;
  MHmenu *mptr;
  Boolean needSave;
} actEditor;

static void toggleProc();
static void saveProc(), quitProc();
static void destroyActEditWin();
static void printMask();

/***************** previewer declarations **********************************/

static EditorDescr retShowDescr = {
  (int (*)()) 0,                        /* input function */
  (char *) 0,                           /* file */
  (char *) 0,	                        /* editor name */
  (char **) 0,                          /* buffer */
  (Window) 0,                           /* message window */
  1,    				/* number of lines in buffer */
  1,					/* lines in window */
  0,		   		        /* cols  in window */
  0, 					/* width of window */
  0,					/* height of window */
  ED_LINEBREAK | ED_FRAMEOFF | ED_VIEWMODE,        /* flags */
};

static editCmd stxtCmdTable [] = {
  {ctrl('f'), right},		/* cursor right */
  {ctrl('b'), left},		/* cursor left  */
  {ctrl('n'), dwncol},		/* down in column */
  {ctrl('p'), upcol},		/* up in column */
  {ctrl('a'), beglin},		/* beginning of line */
  {ctrl('e'), endlin},		/* end of line */
  {ctrl('v'), uppage},		/* forward page */
  {ctrl('l'), newscr},		/* redraw screen */
  {ctrl('s'), fsearch},		/* search forward */
  {ctrl('r'), bwsearch},		/* search backward */
  {EOS,     (int (*)()) 0},	/* end of list */
};		

static editCmd ctrlxTable [] = {		       
  {'t', lineToTop},		/* current line to top of screen */
  {EOS, (int (*)()) 0},		
};

static editCmd escTable [] = {
  {'<', topBuffer},		/* goto top of file */
  {'>', botBuffer},		/* goto bottom of file */
  {' ', goLine},		/* goto specified line */
  {'=', memory},		/* display statistics */
  {'b', wordl},			/* move to left word */
  {'f', wordr},			/* move to right word */
  {'v', dwnpage},		/* display previous page */
  {'n', upScroll},		/* scroll up */
  {'p', dwnScroll},		/* scroll down */
  {'l', botscr},		/* cursor to last line of display */
  {'m', midscr},		/* cursor near middle of display */
  {'h', topscr},		/* cursor to top of display */
  {'s', ffsearch},		/* continue search forward */
  {'r', bbsearch},		/* continue search backword */
  {EOS, (int (*)()) 0},		
};

MHmenu *preModeMenu;
ApplicationNode *preappl;

static MHmenu *putInPreMode();
static changeUserFunct();
static void getPreviewResult();
static handlePreviewPopup();
static void SelectPreWnd();
static void DeselectPreWnd(); 
static void prevRepeat(), prevQuit();

/***************** help editor declarations **********************************/

static struct _hlpEditor {
  Widget hlpWin;
  Widget form, entryForm;
  Widget hlpTitle;
  Widget hlpLabel[ME_MAXITEM + 1];
  struct _formEntry *hlpEdit[ME_MAXITEM + 1];
  Widget quitButton, saveButton;
  MHmenu *mptr;
  Boolean needSave;
} hlpEditor; 

static void hlpSave(), hlpQuit();
 

/* editActivityMask **********************************************************
 * call activity mask editor
 *****************************************************************************/

editActivityMask(mptr, x, y)
     MHmenu *mptr;
     int x, y;
{
  register i, j, cnt = 0;
  int long_Item, long_Index;
  Arg args[20];
  XtCallbackRec call[2];  
  XtTranslations ptt;
 
  static XtActionsRec actMask[] = {
    {"printMask", printMask}
  };

  static char actMaskTable[] = 
    "<KeyPress>: printMask()";

  if (!(mptr = saveMenu(rme->menuEditDescr, mptr, False)) ||
      (mptr->itemCnt == 0)) {
    sprintf(msg, "No items defined yet\nfor current menu");
    exAlert(rme->menuEditDescr->meShell, -1, -1, msg);
    return(1);
  }

  actEditInit(XtDisplay(rme->menuEditDescr->meShell));

  actEditor.needSave = False;
  actEditor.mptr = mptr;
  
  for (i = 0; i < mptr->itemCnt * (mptr->itemCnt + 1); i++)
    butIndex[i] = (struct _actIndex *) Malloc(sizeof(struct _actIndex));

  actEditor.activityMatrix = (int **) Calloc(mptr->itemCnt + 1, sizeof(int *));
  
  for (i = 0; i <= mptr->itemCnt; i++) 
    actEditor.activityMatrix[i] = (int *) Malloc(sizeof(int) * mptr->itemCnt);

  for (i = 0; i < mptr->itemCnt; i++) {
    if (AM_ISSET(&mptr->activityMask, i))
      actEditor.activityMatrix[0][i] = 1;
    else 
      actEditor.activityMatrix[0][i] = 0;
    
    for (j = 0; j < mptr->itemCnt; j++)
      if (AM_ISSET(&mptr->i_table[j]->activityMask, i))
	actEditor.activityMatrix[j + 1][i] = 1;
      else
	actEditor.activityMatrix[j + 1][i] = 0;
  }

  cnt = 0;
  XtSetArg(args[cnt], XmNx, x); cnt++;
  XtSetArg(args[cnt], XmNy, y); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, 2); cnt++;
  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;

  actEditor.actEditShell = 
    XtCreatePopupShell("actShell", transientShellWidgetClass, 
		       rme->menuEditDescr->meShell, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, (caddr_t) mptr); cnt++;
  actEditor.actEditWin = XtCreateManagedWidget("actForm", xmFormWidgetClass, 
					 actEditor.actEditShell, args, cnt);

  XtAddActions(actMask, 1);
  ptt = XtParseTranslationTable(actMaskTable);
  XtOverrideTranslations(actEditor.actEditWin, ptt);

  /********************* title widgets ************************************/
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;

  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR(ACTEDITHEADING, XmSTRING_DEFAULT_CHARSET));
  cnt++;
  actEditor.title1 = XtCreateManagedWidget("acttitle", 
			  xmLabelGadgetClass, actEditor.actEditWin, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, actEditor.title1); cnt++;

  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR(mptr->m_repr->name, XmSTRING_DEFAULT_CHARSET));
  cnt++;
  actEditor.title2 = XtCreateManagedWidget("actmenuname", 
			  xmLabelGadgetClass, actEditor.actEditWin, args, cnt);

  /*********************** item name widgets ****************************/
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, actEditor.title1); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;

  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("initial mask", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  actEditor.actLabel[0] = XtCreateManagedWidget("actlabel", 
			  xmLabelGadgetClass, actEditor.actEditWin, args, cnt);

  /* determine the longest item name */
  long_Item = strlen("initial mask");
  long_Index = 0;

  for (i = 0; i < mptr->itemCnt; i++) {
    cnt = 0;
    XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNtopWidget, actEditor.actLabel[i]); cnt++;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;

    XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
    XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
    XtSetArg(args[cnt], XmNlabelString, 
	     XmStringCreateLtoR(mptr->i_table[i]->i_repr->name, 
				XmSTRING_DEFAULT_CHARSET)); cnt++;
    actEditor.actLabel[i + 1] = XtCreateManagedWidget("actlabel", 
			  xmLabelGadgetClass, actEditor.actEditWin, args, cnt);

    if (strlen(mptr->i_table[i]->i_repr->name) > long_Item) {
      long_Item = strlen(mptr->i_table[i]->i_repr->name);
      long_Index = i + 1;
    }
  }


  /*********************** widget for work area ************************/
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, actEditor.title1); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, actEditor.actLabel[long_Index]); 
  cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;

  XtSetArg(args[cnt], XmNborderWidth, 2); cnt++;
  XtSetArg(args[cnt], XmNentryClass, xmToggleButtonGadgetClass); cnt++;
  XtSetArg(args[cnt], XmNisHomogeneous, True); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
  XtSetArg(args[cnt], XmNorientation, XmHORIZONTAL); cnt++;
  XtSetArg(args[cnt], XmNpacking, XmPACK_COLUMN); cnt++;
  XtSetArg(args[cnt], XmNnumColumns, mptr->itemCnt + 1); cnt++;

  actEditor.actEditWork = XtCreateManagedWidget("actWork", 
			      xmRowColumnWidgetClass, actEditor.actEditWin, 
			      args, cnt);

  /*********************  initial mask *****************************/
  call[0].callback = toggleProc;
  call[0].closure = (caddr_t) &actEditor;
  call[1].callback = NULL;
  call[1].closure = NULL;

  for (i = 0; i < mptr->itemCnt; i++) {
    cnt = 0;
    XtSetArg(args[cnt], XmNheight, box_width); cnt++;
    XtSetArg(args[cnt], XmNwidth, box_width); cnt++;
    XtSetArg(args[cnt], XmNspacing, 0); cnt++;
    XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
    XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
    XtSetArg(args[cnt], XmNborderWidth, 0); cnt++;
    XtSetArg(args[cnt], XmNshadowThickness, 1); cnt++;
    XtSetArg(args[cnt], XmNhighlightThickness, 1); cnt++;
    XtSetArg(args[cnt], XmNlabelType, XmPIXMAP); cnt++;
    XtSetArg(args[cnt], XmNindicatorOn, False); cnt++;

    XtSetArg(args[cnt], XmNlabelPixmap, 
	     exWhitePM(XtDisplay(actEditor.actEditShell))); cnt++;
    XtSetArg(args[cnt], XmNselectPixmap, 
	     exBlackPM(XtDisplay(actEditor.actEditShell))); cnt++;

    XtSetArg(args[cnt], XmNvalueChangedCallback, call); cnt++;  

    if (actEditor.activityMatrix[0][i] == 1) {
      XtSetArg(args[cnt], XmNset, True); 
      cnt++; 
    }

    butIndex[i]->x = 0;
    butIndex[i]->y = i;
    XtSetArg(args[cnt], XmNuserData, (caddr_t) butIndex[i]); cnt++;

    actEditor.button[0][i] = XtCreateManagedWidget("actButton",
						   xmToggleButtonGadgetClass, 
						   actEditor.actEditWork, 
						   args, cnt);
  }

  /******************** item masks *********************************/
  call[0].callback = toggleProc;
  call[0].closure = (caddr_t) &actEditor;
  call[1].callback = NULL;
  call[1].closure = NULL;

  for (j = 1; j < mptr->itemCnt + 1; j++)
    for (i = 0; i < mptr->itemCnt; i++) {
      cnt = 0;
      XtSetArg(args[cnt], XmNheight, box_width); cnt++;
      XtSetArg(args[cnt], XmNwidth, box_width); cnt++;
      XtSetArg(args[cnt], XmNspacing, 0); cnt++;
      XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
      XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
      XtSetArg(args[cnt], XmNborderWidth, 0); cnt++;
      XtSetArg(args[cnt], XmNshadowThickness, 1); cnt++;
      XtSetArg(args[cnt], XmNhighlightThickness, 1); cnt++;
      XtSetArg(args[cnt], XmNlabelType, XmPIXMAP); cnt++;
      XtSetArg(args[cnt], XmNindicatorOn, False); cnt++;

      XtSetArg(args[cnt], XmNlabelPixmap, 
	       exWhitePM(XtDisplay(actEditor.actEditShell))); cnt++;
      XtSetArg(args[cnt], XmNselectPixmap, 
	       exBlackPM(XtDisplay(actEditor.actEditShell))); cnt++;
      XtSetArg(args[cnt], XmNlabelInsensitivePixmap, 
	       exLight_greyPM(XtDisplay(actEditor.actEditShell))); cnt++;
      XtSetArg(args[cnt], XmNselectInsensitivePixmap, 
	       exLight_greyPM(XtDisplay(actEditor.actEditShell))); cnt++;
      
      XtSetArg(args[cnt], XmNvalueChangedCallback, call); cnt++;  

      if (i + 1 == j) {                               /* not selectable */
	XtSetArg(args[cnt], XmNsensitive, False); 
	cnt++;
      }                                            /* activity mask set */
      else if (actEditor.activityMatrix[j][i] == 1) {
	XtSetArg(args[cnt], XmNset, True); 
	cnt++;      
      }

      butIndex[j * mptr->itemCnt + i]->x = j;
      butIndex[j * mptr->itemCnt + i]->y = i;
      XtSetArg(args[cnt], XmNuserData, 
	       (caddr_t) butIndex[j * mptr->itemCnt + i]); cnt++;

      actEditor.button[j][i] = XtCreateManagedWidget(NULL, 
						    xmToggleButtonGadgetClass, 
						     actEditor.actEditWork, 
						     args, cnt);
    }
					
/*  XDefineCursor(actEditor.actEditWork, dotCursor);*/

  /************************ dialogue buttons ******************************/   
  call[0].callback = quitProc;
  call[0].closure = (caddr_t) &actEditor;
  call[1].callback = NULL;
  call[1].closure = NULL;

  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, actEditor.actLabel[mptr->itemCnt]); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, 1); cnt++;
  XtSetArg(args[cnt], XmNlabelString, XmStringCreateLtoR("Quit", 
                                        XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetArg(args[cnt], XmNactivateCallback, call); cnt++;

  actEditor.quitButton = XtCreateManagedWidget("quitB", 
			      xmPushButtonGadgetClass, actEditor.actEditWin, 
			      args, cnt);

  call[0].callback = saveProc;
  call[0].closure = (caddr_t) &actEditor;
  call[1].callback = NULL;
  call[1].closure = NULL;

  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, actEditor.quitButton); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, actEditor.actLabel[mptr->itemCnt]); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, 1); cnt++;
  XtSetArg(args[cnt], XmNlabelString, XmStringCreateLtoR("Save", 
                                        XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetArg(args[cnt], XmNactivateCallback, call); cnt++;

  actEditor.saveButton = XtCreateManagedWidget("saveB", 
			      xmPushButtonGadgetClass, actEditor.actEditWin, 
			      args, cnt);

  subCompState(rme->menuEditDescr, 1, True);

  XtPopup(actEditor.actEditShell, XtGrabNone);
}

/* toggleProc ****************************************************************
 * activity mask editor application function to react on toggling the
 * activity mask buttons
 ****************************************************************************/

static void toggleProc(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  Arg args[10];
  register cnt;
  struct _actEditor *editor = (struct _actEditor *) closure;
  struct _actIndex *actInd;

  actInd = (struct _actIndex *) Malloc(sizeof(struct _actIndex));

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, (XtArgVal) &actInd); cnt++;
  XtGetValues(widget, args, cnt);

  if (editor->activityMatrix[actInd->x][actInd->y])
    editor->activityMatrix[actInd->x][actInd->y] = 0;
  else
    editor->activityMatrix[actInd->x][actInd->y] = 1;

  editor->needSave = True;
  free(actInd);
}
 
/* saveProc ******************************************************************
 * activity mask editor application function to quit
 ****************************************************************************/
/*ARGSUSED*/
static void saveProc(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  register i, j;
  struct _actEditor *editor = (struct _actEditor *) closure;

  if (editor->needSave) {
    for (i = 0; i < editor->mptr->itemCnt; i++)
      if (editor->activityMatrix[0][i])
	AM_SET(&editor->mptr->activityMask, i);
      else 
	AM_CLR(&editor->mptr->activityMask, i);

    for (j = 0; j < editor->mptr->itemCnt; j++) {
      for (i = 0; i < editor->mptr->itemCnt; i++)
	if (editor->activityMatrix[j + 1][i])
	  AM_SET(&editor->mptr->i_table[j]->activityMask, i);
	else 
	  AM_CLR(&editor->mptr->i_table[j]->activityMask, i);
    }
  }
  
  editor->needSave = False;
  rme->applPtr->_changed = True;
}

/* quitProc ****************************************************************
 * activity mask editor application function to save activity mask
 ****************************************************************************/
/*ARGSUSED*/
static void quitProc(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  struct _actEditor *editor = (struct _actEditor *) closure;

  if (editor->needSave)
    if (exNotifyUser(rme->menuEditDescr->meShell, -1, -1, 
		     "Activity masks not saved ! Quit anyway ?") == ANSWERNO)
      return;

  destroyActEditWin(editor);
  subCompState(rme->menuEditDescr, 1, False);
}

/* actEditInit ***************************************************************
 * initialize graphic context and cursor
 ****************************************************************************/

static actEditInit(d)
     Display *d;
{
  if (! act_gc) {
    act_gc = XCreateGC(d, XDefaultRootWindow(d), 0, 0);
    XSetForeground(d, act_gc, exBlackPixel(d));
    XSetBackground(d, act_gc, exWhitePixel(d));
    XSetFillStyle(d, act_gc, FillTiled);
    XSetTile(d, act_gc, exDark_greyPM(d));
    XSetFunction(d, act_gc, GXcopy);
  }

/*  if ( !dotCursor)
    dotCursor =  XCreateCursor(dot_width, dot_height, 
			       dot_bits, dotMask_bits, 
			       dot_x_hot, dot_y_hot, 
			       exForeground(), exBackground(), 
			       GXcopy);
*/
}

/* printMask ****************************************************************
 * print activity mask
 ****************************************************************************/

/*ARGSUSED*/
static void printMask(w, event, params, num_params)
     Widget w;
     XKeyEvent *event;
     String *params;
     Cardinal *num_params;
{
  register i, j;
  Arg args[1];
  MHmenu *mptr;

  XtSetArg(args[0], XmNuserData, (XtArgVal) &mptr);
  XtGetValues(w, args, 1);

  fprintf(stderr, "Activity masks of menu --%s--\n\n", mptr->m_repr->name);

  for (i = 0; i < mptr->itemCnt; i++)
    if (AM_ISSET(&mptr->activityMask, i))
      fprintf(stderr, "1");
    else
      fprintf(stderr, "0");
      fprintf(stderr, "\n");

  for (j = 0; j < mptr->itemCnt; j++) {
    for (i = 0; i < mptr->itemCnt; i++)
      if (AM_ISSET(&mptr->i_table[j]->activityMask, i))
	fprintf(stderr, "1");
      else
	fprintf(stderr, "0");
	fprintf(stderr, "\n");
  }
  fprintf(stderr, "\n\n");
}

/* checkActMaskEditor *********************************************************
 * quit activity mask editor if current menu is changed in menu editor
 ****************************************************************************/

static void checkActMaskEditor(mEditWin) 
     meditWindow *mEditWin;
{
  if ((mEditWin->com_flag & ACTEDON) != 0)
    quitProc((Widget) 0, &actEditor, (caddr_t) 0);
}

/* destroyActEditWin *********************************************************
 * 
 ****************************************************************************/

static void destroyActEditWin(editor)
     struct _actEditor *editor;
{
  register i;
  
  for (i = 0; i < editor->mptr->itemCnt * (editor->mptr->itemCnt + 1); i++)
    free(butIndex[i]);

  for (i = 0; i < editor->mptr->itemCnt; i++)
    free(editor->activityMatrix[i]);

  cfree(editor->activityMatrix/*, editor->mptr->itemCnt + 1, sizeof(int *)*/);
  /*  free(editor); */

  XtDestroyWidget(editor->actEditShell);
}

/* Previewer ****************************************************************
 * internal function to provide windows in which preview menus can be handled;
 * creates a menu representation from a menu description in a preview mode
 ****************************************************************************/

Previewer(mEditWin, aptr, mptr)
     meditWindow *mEditWin;
     ApplicationNode *aptr;   /* application node containing menu to preview */
     MHmenu *mptr;            /* menu to show in previewer */
{
  register cnt;
  Arg args[20];
  XtCallbackRec call[2];
  Position px, py;
  previewWindow *preWin;
  MenuDescr preMenuWin[MDS_MAXMENU];
  int preWinCnt = MDS_MAXMENU, fw;

  mEditWin->pre_win = (previewWindow *) Malloc(sizeof(previewWindow));
  preWin = mEditWin->pre_win;

  fw = exFontWidth(exBaseFont(XtDisplay(mEditWin->meShell)));

  cnt = 0;
  XtSetArg(args[cnt], XmNx, &px); cnt++;
  XtSetArg(args[cnt], XmNy, &py); cnt++;
  XtGetValues(mEditWin->meShell, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNx, px + 29 * fw); cnt++;
  XtSetArg(args[cnt], XmNy, py + menuBarHeight(mEditWin->menuBar)); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, 2); cnt++;
  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;

  preWin->pShell = XtCreatePopupShell("preview", transientShellWidgetClass, 
				 mEditWin->meShell, args, cnt);

  preWin->form = XtCreateManagedWidget("preForm", xmFormWidgetClass, 
				       preWin->pShell, (Arg *) 0, 0);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;

  preWin->menuBar = exCreateMenuBar(preWin->form, args,
				    cnt, getPreviewResult);

  preModeMenu = putInPreMode(aptr, mptr);

  setWorkApplication();
  mEditWin->preMenu = exMakeMenu(currentApplication(), preWin->menuBar, 
				 mEditWin->meShell, preModeMenu);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, preWin->menuBar->menubar); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("return value:", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  preWin->text = XtCreateManagedWidget("prelabel", 
			  xmLabelGadgetClass, preWin->form, args, cnt);

  /* create the editor window */
  retShowDescr.cols = E_ENTRYLEN3;
  retShowDescr.lines = 3;

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, preWin->text); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;

  preWin->retShow =
    makeEditor(preWin->form, "retShow", &retShowDescr, args, cnt);
  initCmdTables(preWin->retShow, stxtCmdTable, escTable, ctrlxTable, 
		(editCmd *) 0);

  switchCursor(preWin->retShow, FALSE);

  /************************ dialogue buttons ******************************/   
  call[0].callback = prevQuit;
  call[0].closure = (caddr_t) preWin;
  call[1].callback = NULL;
  call[1].closure = NULL;

  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, editorWidget(preWin->retShow)); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, 1); cnt++;
  XtSetArg(args[cnt], XmNlabelString, XmStringCreateLtoR("Quit", 
                                        XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetArg(args[cnt], XmNactivateCallback, call); cnt++;

  preWin->quitButton = XtCreateManagedWidget("quitB", 
			      xmPushButtonGadgetClass, preWin->form, 
			      args, cnt);


  call[0].callback = prevRepeat;
  call[0].closure = (caddr_t) preWin;
  call[1].callback = NULL;
  call[1].closure = NULL;

  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, preWin->quitButton); cnt++;
  XtSetArg(args[cnt], XmNleftOffset, 30); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, editorWidget(preWin->retShow)); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, 1); cnt++;
  XtSetArg(args[cnt], XmNlabelString, XmStringCreateLtoR(
		 "Show popup menu/list", XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetArg(args[cnt], XmNactivateCallback, call); cnt++;

  preWin->repButton = XtCreateManagedWidget("repB", 
			      xmPushButtonGadgetClass, preWin->form, 
			      args, cnt);

  XtPopup(preWin->pShell, XtGrabNonexclusive);
  mEditWin->com_flag |= PREVIEWON;
}

/* putInPreMode *************************************************************
 * for every menu of the application node containing the menu to be previewed
 * a copy is created, in which all user defined functions are set to a default 
 * function; the duplicated menus are added to the workapplication
 * which contains all popup-menus which have to be accessable while MenuEditor
 * is active (currentApplication=workApplication)
 *****************************************************************************/

static MHmenu *putInPreMode(aptr, mptr)
     ApplicationNode *aptr;   /* application node containing menu to preview */
     MHmenu *mptr;                /* menu to show in previewer */
{
  MHmenu *pmenu;
  int i, j, err;

  setWorkApplication();

  for (i = 0; i < aptr->menuCnt; i++) {
    if (strcmp(aptr->m_table[i]->m_repr->name, mptr->m_repr->name) != 0) {
      pmenu = createMenu();
      copyMenuDescr(aptr->m_table[i], pmenu);
      if ((err = insertMenu(pmenu)) == OK)
	INTERNALS++;
      else
	uim_error(err);

      if (checkFunc(aptr)	/*|| checkMenuReq(preModeMenu)*/)
	for (j = 0; j < pmenu->itemCnt; j++)
	  changeUserFunct(pmenu->i_table[j], aptr);
    }
  }

  pmenu = createMenu();
  copyMenuDescr(mptr, pmenu);
  if ((err = insertMenu(pmenu)) == OK)
    INTERNALS++;
  else
    uim_error(err);

  if (checkFunc(aptr) /*|| checkMenuReq(preModeMenu)*/)
    for (j = 0; j < pmenu->itemCnt; j++)
      changeUserFunct(pmenu->i_table[j], aptr);
  /*pmenu = getMenuByName(mptr->m_repr->name);*/
  return(pmenu);
}

/* checkMenuReq **************************************************************
 * function returns True if menu has requests via menu defined               *
 ****************************************************************************/

Boolean checkMenuReq(mptr)
MHmenu *mptr;
{
  register i, j;

  if (mptr == (MHmenu *) 0) {
    setError(ERR_NULL);
    return(False);
  }  

  for (i = 0; i < mptr->itemCnt; i++)
    for (j = 0; j < mptr->i_table[i]->i_return->reqCnt; i++)
      if (mptr->i_table[i]->i_requests[j]->reqFnct == reqViaMenu) 
        return(True);

  return(False);
}


/* changeUserFunct ************************************************************
 * resets user functions of menu in previewer to the default function 
 * "defaultUserFunct" and resets requests via menu to the default function
 * "defaultMenuFunct"
 *****************************************************************************/

static changeUserFunct(iptr, aptr)
     MHitem *iptr;                          /* item of menu in preview mode */
     ApplicationNode *aptr;     /* application node of menu in preview mode */
{
  int i, j = 0;

  for (i = 0; i < iptr->i_return->reqCnt; i++) 

    if (iptr->i_requests[i]->flags & MH_USERFUNCTION) {
      while ((iptr->i_requests[i]->reqFnct != aptr->ApplFunc[j].func) && 
	     (j < MDS_MAXFUNC))
	j++;

      setReqFnct(iptr->i_requests[i], defaultUserFunct);
      if (j < MDS_MAXFUNC)
	setRequestData(iptr->i_requests[i], aptr->ApplFunc[j].funcName);
    }
}

/* defaultUserFunct **********************************************************
 * application function of previewer which is executed instead of user defined 
 * functions in preview menu 
 *****************************************************************************/

defaultUserFunct(parms)
     RequestParm *parms;
{
  char *retstring;

  retstring = malloc(strlen(parms->data) + 70);
  sprintf(retstring, "if menu is not in preview mode\nuser function <%s> is executed.", parms->data);

  exAlert(rme->menuEditDescr->meShell, -1, -1, retstring);    
    /* button sollte "continue" heissen */
  free(retstring);

  setWorkApplication();
  return(0);
}

/* getPreviewResult ***********************************************************
 * handles return value from pulldown menus in preview mode
 ****************************************************************************/

static void getPreviewResult(w, event, params, num_params)
     Widget w;
     XPropertyEvent *event;
     String *params;
     Cardinal *num_params;
{
  char *value;
  register cnt = 0;
  Arg args[5];

  clearBuffer(rme->menuEditDescr->pre_win->retShow);

  if (value = strCopy(getMenuEventReturnValue(event))) {
    XtSetArg(args[cnt], XmNlabelString, 
	     XmStringCreateLtoR("return value:", XmSTRING_DEFAULT_CHARSET));
    editorAppend(rme->menuEditDescr->pre_win->retShow, value);
    cnt++;
  }
  else {
    XtSetArg(args[cnt], XmNlabelString, 
	     XmStringCreateLtoR("no return value", XmSTRING_DEFAULT_CHARSET));
    cnt++;
  }

  XtSetValues(rme->menuEditDescr->pre_win->text, args, cnt);

  free(value);
}

/* handlePreviewPopup *******************************************************
 * calls popup-menus in preview mode and handles return value
 ***************************************************************************/

static handlePreviewPopup(meWin)
     meditWindow *meWin;
{
  register cnt = 0;
  char *retvalue;
  Arg args[5];

  exCallMenu(meWin->preMenu);

  clearBuffer(rme->menuEditDescr->pre_win->retShow);

  if (retvalue = exGetMenuReturnValue(meWin->preMenu)) {
    XtSetArg(args[cnt], XmNlabelString, 
	     XmStringCreateLtoR("return value:", XmSTRING_DEFAULT_CHARSET)); 
    editorAppend(rme->menuEditDescr->pre_win->retShow, retvalue);
    cnt++;
  }
  else {
    XtSetArg(args[cnt], XmNlabelString, 
	     XmStringCreateLtoR("no return value", XmSTRING_DEFAULT_CHARSET));
    cnt++;
  }

  XtSetValues(meWin->pre_win->text, args, cnt);

  free(retvalue);
}

/* prevQuit *******************************************************************
 * previewer application function to quit
 *****************************************************************************/
/*ARGSUSED*/
static void prevQuit(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  previewWindow *prevWin = (previewWindow *) closure;

  stopPreview((RequestParm *) NULL);
}

/* prevRepeat ****************************************************************
 * previewer application function to repeat popup menu in preview mode
 ****************************************************************************/
/*ARGSUSED*/
static void prevRepeat(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  previewWindow *prevWin = (previewWindow *) closure;

  if (preModeMenu->m_type == MDS_M_POPUP 
      || preModeMenu->m_type == MDS_L_POPUP)
    handlePreviewPopup(rme->menuEditDescr);
}

/* createHelpForm ************************************************************
 * create a form containing the help information for the input menu and its
 * items
 ****************************************************************************/

createHelpForm(mEditWin, mptr, x, y)
     meditWindow *mEditWin;
     MHmenu *mptr;
     int x, y;                     /* position of form to edit help text */
{
  register cnt, i;
  int long_Item, long_Index;
  Arg args[20];
  XtCallbackRec call[2];
  char *hmsg;

  hlpEditor.needSave = False;
  hlpEditor.mptr = mptr;

  cnt = 0;
  XtSetArg(args[cnt], XmNx, x); cnt++;
  XtSetArg(args[cnt], XmNy, y); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, 2); cnt++;
  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;

  hlpEditor.hlpWin = XtCreatePopupShell("hlpShell", transientShellWidgetClass, 
		       mEditWin->meShell, args, cnt);

  exWidgetPos(hlpEditor.hlpWin, -1, -1, 0, 0);

  hlpEditor.form = XtCreateManagedWidget("hlpForm", xmFormWidgetClass, 
				      hlpEditor.hlpWin, NULL, 0);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  sprintf(msg, "edit help text for menu <<%s>>", mptr->m_repr->name);
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR(msg, XmSTRING_DEFAULT_CHARSET));
  cnt++;
  hlpEditor.hlpTitle = XtCreateManagedWidget("hlptitle", 
			  xmLabelGadgetClass, hlpEditor.form, args, cnt);

  /************************ labels *************************************/

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, hlpEditor.hlpTitle); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR(mptr->m_repr->name, XmSTRING_DEFAULT_CHARSET));
  cnt++;
  hlpEditor.hlpLabel[0] = XtCreateManagedWidget("hlpmenulabel", 
			  xmLabelGadgetClass, hlpEditor.form, args, cnt);

  /* determine the longest item name */
  long_Item = strlen(mptr->m_repr->name);
  long_Index = 0;

  for (i = 0; i < mptr->itemCnt; i++) {
    cnt = 0;
    XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNtopWidget, hlpEditor.hlpLabel[i]); cnt++;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNlabelString, 
	              XmStringCreateLtoR(mptr->i_table[i]->i_repr->name, 
					 XmSTRING_DEFAULT_CHARSET));
    cnt++;
    hlpEditor.hlpLabel[i + 1] = XtCreateManagedWidget("hlpitemlabel", 
			  xmLabelGadgetClass, hlpEditor.form, args, cnt);

    if (strlen(mptr->i_table[i]->i_repr->name) > long_Item) {
      long_Item = strlen(mptr->i_table[i]->i_repr->name);
      long_Index = i + 1;
    }
  }

  /************************ entries *************************************/

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, hlpEditor.hlpTitle); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, hlpEditor.hlpLabel[long_Index]); 
  cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;

  hlpEditor.entryForm = XtCreateManagedWidget("hlpWork", 
			      xmFormWidgetClass, hlpEditor.form, args, cnt);
  
  hlpEditor.hlpEdit[0] = createFormEntry(hlpEditor.entryForm, "",
					 E_ENTRYLEN3, E_ENTRYBUFLEN);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetValues(hlpEditor.hlpEdit[0]->entry->entryWindow, args, cnt);

  for (i = 0; i < mptr->itemCnt; i++) {
    hlpEditor.hlpEdit[i + 1] = createFormEntry(hlpEditor.entryForm, "",
					       E_ENTRYLEN3, E_ENTRYBUFLEN);

    cnt = 0;
    XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNtopWidget, 
	     hlpEditor.hlpEdit[i]->entry->entryWindow); cnt++;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
    XtSetValues(hlpEditor.hlpEdit[i + 1]->entry->entryWindow, args, cnt);
  }

  /************************ dialogue buttons ******************************/   
  call[0].callback = hlpQuit;
  call[0].closure = (caddr_t) &hlpEditor;
  call[1].callback = NULL;
  call[1].closure = NULL;

  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, hlpEditor.hlpLabel[mptr->itemCnt]); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, 1); cnt++;
  XtSetArg(args[cnt], XmNlabelString, XmStringCreateLtoR("Quit", 
                                        XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetArg(args[cnt], XmNactivateCallback, call); cnt++;

  hlpEditor.quitButton = XtCreateManagedWidget("quitB", 
			      xmPushButtonGadgetClass, hlpEditor.form, 
			      args, cnt);

  call[0].callback = hlpSave;
  call[0].closure = (caddr_t) &hlpEditor;
  call[1].callback = NULL;
  call[1].closure = NULL;

  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, hlpEditor.quitButton); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, hlpEditor.hlpLabel[mptr->itemCnt]); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;

  XtSetArg(args[cnt], XmNborderWidth, 1); cnt++;
  XtSetArg(args[cnt], XmNlabelString, XmStringCreateLtoR("Save", 
                                        XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetArg(args[cnt], XmNactivateCallback, call); cnt++;

  hlpEditor.saveButton = XtCreateManagedWidget("saveB", 
			      xmPushButtonGadgetClass, hlpEditor.form, 
			      args, cnt);

  /***************** fill in entries with help text ************************/

  if (hmsg = mptr->m_hlp->hlpmsg)
    fillEntry(hlpEditor.hlpEdit[0], hmsg, False);

  for (i = 0; i < mptr->itemCnt; i++)
    if (hmsg = mptr->i_table[i]->i_hlp->hlpmsg)
      fillEntry(hlpEditor.hlpEdit[i + 1], hmsg, False);

  XtPopup(hlpEditor.hlpWin, XtGrabNone);
}

/* hlpQuit *******************************************************************
 * help editor application function to quit
 *****************************************************************************/
/*ARGSUSED*/
static void hlpQuit(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  struct _hlpEditor *editor = (struct _hlpEditor *) closure;
  register i;

  if (!cmpFormEntry(editor->hlpEdit[0], editor->mptr->m_hlp->hlpmsg))
    editor->needSave = True;

  for (i = 0; i < editor->mptr->itemCnt; i++)
    if (!cmpFormEntry(editor->hlpEdit[i + 1], 
		     editor->mptr->i_table[i]->i_hlp->hlpmsg))
      editor->needSave = True;

  if (editor->needSave)
    if (exNotifyUser(rme->menuEditDescr->meShell, -1, -1, 
		     "help text not saved ! Quit anyway ?") == ANSWERNO)
      return;

  XtDestroyWidget(editor->hlpWin);
  subCompState(rme->menuEditDescr, 2, False);
}

/* hlpSave ******************************************************************
 * help editor application function to save help text
 ****************************************************************************/
/*ARGSUSED*/
static void hlpSave(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  struct _hlpEditor *editor = (struct _hlpEditor *) closure;
  char *hmsg;
  register i;

  hmsg = getBufferContents2(editor->hlpEdit[0]->entry->entryEditor);
  if (strCmp(editor->mptr->m_hlp->hlpmsg, hmsg)) {
    free(editor->mptr->m_hlp->hlpmsg);
    editor->mptr->m_hlp->hlpmsg = strCopy(hmsg);
    free(hmsg);
  }

  for (i = 0; i < editor->mptr->itemCnt; i++) {
    hmsg = getBufferContents2(editor->hlpEdit[i + 1]->entry->entryEditor);
    if (strCmp(editor->mptr->i_table[i]->i_hlp->hlpmsg, hmsg)) {
      free(editor->mptr->i_table[i]->i_hlp->hlpmsg);
      editor->mptr->i_table[i]->i_hlp->hlpmsg = strCopy(hmsg);
      free(hmsg);
    }  
  } 

  editor->needSave = False;
  rme->applPtr->_changed = True;
}

/* checkHelpEditor ***********************************************************
 * quit help editor if current menu is changed in menu editor
 ****************************************************************************/

static void checkHelpEditor(mEditWin) 
     meditWindow *mEditWin;
{
  if ((mEditWin->com_flag & HELPEDON) != 0)
    hlpQuit((Widget) 0, &hlpEditor, (caddr_t) 0);
}

/* editReqForm1 ***********************************************************
 * create form to edit requests
 *****************************************************************************/

editReqForm1(mptr, x, y)
     MHmenu *mptr;
     int x, y;
{
  register i, cnt = 0, ac = 0;
  int actmask[ME_MAXITEM], maxlen = 0, itemlen = 0;
  Arg args[20];
  MHmenu *itemList, *mhptr, *reqMenu1, *reqMenu2, *reqMenu3;
  char *nameList[ME_MAXITEM], str[100];
  static char *list2[] = {"quit"};
  static Boolean INIT = False;
 
  if (!(reqForm.mptr = saveMenu(rme->menuEditDescr, mptr, False)) ||
      (reqForm.mptr->itemCnt == 0)) {
    sprintf(msg, "No items defined yet\nfor current menu");
    exAlert(rme->menuEditDescr->meShell, -1, -1, msg);
    return(1);
  }

  subCompState(rme->menuEditDescr, 0, True);

  reqForm.iptr = NULL;
  reqForm.reqIndex = -1;

  for (i = 0; i < reqForm.mptr->itemCnt; i++)  /* search defined requests */
    if (reqForm.mptr->i_table[i]->i_return->reqCnt > 0) {
      reqForm.iptr = reqForm.mptr->i_table[i];
      reqForm.reqIndex = 0;
      reqForm.req = reqForm.iptr->i_requests[reqForm.reqIndex];
      break;
    }

  for (i = 0; i < reqForm.mptr->itemCnt; i++) {
    nameList[ac] = malloc(strlen(reqForm.mptr->i_table[i]->i_repr->name) 
			    + 5);
    sprintf(nameList[ac++], "%d %s", i + 1, 
	      reqForm.mptr->i_table[i]->i_repr->name);
  }

  if (menu1)
    XtDestroyWidget(menu1->button);
  if (menu2)
    XtDestroyWidget(menu2->button);
  if (menu3)
    XtDestroyWidget(menu3->button);

  reqMenu1 = generateListMenu(EDITREQITEMSEL, nameList, ac, genReqList, 
				NULL, TRUE);
  setMenuType(reqMenu1, MDS_M_PULLD); 
  
  reqMenu2 = generateListMenu(DELREQITEMSEL, nameList, ac, meDeleteRequest, 
				NULL, TRUE);
  setMenuType(reqMenu2, MDS_M_PULLD);

  reqMenu3 = generateListMenu(EDITREQQUIT, list2, 1, quitReqEdit, NULL, 
				TRUE);

  setMenuType(reqMenu3, MDS_M_PULLD);

  for (i = 0; i < reqForm.mptr->itemCnt; i++)
    free(nameList[i]);

  if (!INIT) {
    initReqForm1(x, y, mptr, reqMenu1, reqMenu2, reqMenu3);
    INIT = True;
  }
  else {
    cnt = 0;
    XtSetArg(args[cnt], XmNx, x); cnt++;
    XtSetArg(args[cnt], XmNy, y); cnt++;
    XtSetValues(reqForm.reqFormShell, args, cnt);

    cnt = 0;
    XtSetArg(args[cnt], XmNuserData, (caddr_t) mptr); cnt++;
    XtSetValues(reqForm.form, args, cnt);

    /* create menu descriptions for menus of request editor */
    if ((!(menu3 =  exMakeMenu(currentApplication(), reqForm.menuBar,
		      reqForm.reqFormShell, reqMenu3))) 
	|| (!(menu1 = exMakeMenu(currentApplication(), reqForm.menuBar,
			 reqForm.reqFormShell, reqMenu1))) 
	|| (!(menu2 = exMakeMenu(currentApplication(), reqForm.menuBar, 
			 reqForm.reqFormShell, reqMenu2)))) {
      fprintf(stderr, 
	      "error while creating menu description of request editor\n");
      return(-1);
    }

    /********************* title widgets ************************************/
    cnt = 0;
    if (reqForm.reqIndex < 0)
      sprintf(str, "No request defined");
    else
      sprintf(str, "Request <<%d>>", reqForm.reqIndex + 1);

    XtSetArg(args[cnt], XmNlabelString, XmStringCreateLtoR(str, 
					     XmSTRING_DEFAULT_CHARSET)); cnt++;
    XtSetValues(reqForm.titleReq, args, cnt);

    cnt = 0;
    if (reqForm.reqIndex < 0)
      sprintf(str, "for menu <<%s>>", reqForm.mptr->m_repr->name);
    else
      sprintf(str, "for item <<%s>>:", reqForm.iptr->i_repr->name);

    XtSetArg(args[cnt], XmNlabelString, XmStringCreateLtoR(str, 
					    XmSTRING_DEFAULT_CHARSET)); cnt++;
    XtSetValues(reqForm.titleItem, args, cnt);
  }
    
  /*********************** widget for functp ************************/

  setWorkApplication();
  mhptr = getMenuDescrByName(FUNCLISTMENU);

  itemList = generateListMenu("funcListPanel", FuncName, mhptr->itemCnt, 
				editReqForm2, NULL, TRUE);

  if (reqForm.reqIndex < 0) {
    setMask(actmask, itemList->itemCnt, 0);
    setMenuInitialActivityMask(itemList, actmask);
  }
  else {
    if ((i = mds_function(reqForm.req->reqFnct)) > 0)
      setDefaultItem(itemList->i_table[i - 2]);
    else if ((i = user_function(reqForm.req->reqFnct, rme->applPtr)) > 0)
      setDefaultItem(itemList->i_table[i + 3 /*MDS_FUNCT_CNT*/ - 1]);
  }

  if (reqForm.menup)
    freeMenu(reqForm.menup);

  reqForm.menup = exMakePanel(currentApplication(), reqForm.functp, 
			      itemList, 1, -1, False); 

  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, reqForm.menup->menuWidth); cnt++;
  XtSetArg(args[cnt], XmNheight, reqForm.menup->menuHeight); cnt++;
  XtSetValues(reqForm.functp, args, cnt);

  exWidgetPos(reqForm.reqFormShell, x, y, 0, 0);

  XtPopup(reqForm.reqFormShell, XtGrabNone);
}

/* initReqForm1 **************************************************************
 * initialize first form to edit requests
 ****************************************************************************/

static initReqForm1(x, y, mptr, rMenu1, rMenu2, rMenu3)
     int x, y;
     MHmenu *mptr, *rMenu1, *rMenu2, *rMenu3;
{
  register cnt, i;
  Arg args[20];
  char str[100];
  Widget label;

  cnt = 0;
  XtSetArg(args[cnt], XmNx, x); cnt++;
  XtSetArg(args[cnt], XmNy, y); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, 2); cnt++;
  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;

  reqForm.reqFormShell =  XtCreatePopupShell("reqShell", 
					     transientShellWidgetClass, 
    			      rme->menuEditDescr->meShell, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, (caddr_t) mptr); cnt++;
  reqForm.form = XtCreateManagedWidget("reqForm", xmFormWidgetClass, 
				       reqForm.reqFormShell, args, cnt);

  /********************* menubar ************************************/
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;

  reqForm.menuBar = exCreateMenuBar(reqForm.form, args,
				    cnt, getRequestResult);

  /* create menu descriptions for menus of request editor */
  if ((!(menu3 =  exMakeMenu(currentApplication(), reqForm.menuBar,
		      reqForm.reqFormShell, rMenu3))) 
      || (!(menu1 = exMakeMenu(currentApplication(), reqForm.menuBar,
			 reqForm.reqFormShell, rMenu1))) 
      || (!(menu2 =  exMakeMenu(currentApplication(), reqForm.menuBar, 
			 reqForm.reqFormShell, rMenu2)))) {
      fprintf(stderr, 
	      "error while creating menu description of request editor\n");
      return(-1);
    }


  /********************* title widgets ************************************/
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, reqForm.menuBar->menubar); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;

  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;

  if (reqForm.reqIndex < 0)
    sprintf(str, "No request defined");
  else
    sprintf(str, "Request <<%d>>", reqForm.reqIndex + 1);

  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR(str, XmSTRING_DEFAULT_CHARSET));
  cnt++;
  reqForm.titleReq = XtCreateManagedWidget("reqtitle1", 
			  xmLabelGadgetClass, reqForm.form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, reqForm.titleReq); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;

  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;

  if (reqForm.reqIndex < 0)
    sprintf(str, "for menu <<%s>>", reqForm.mptr->m_repr->name);
  else
    sprintf(str, "for item <<%s>>:", reqForm.iptr->i_repr->name);
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR(str, XmSTRING_DEFAULT_CHARSET));
  cnt++;
  reqForm.titleItem = XtCreateManagedWidget("reqtitle2", 
			  xmLabelGadgetClass, reqForm.form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, reqForm.titleItem); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("request function list:", 
			      XmSTRING_DEFAULT_CHARSET)); cnt++;  
  label = XtCreateManagedWidget("reqtitle3", 
			   xmLabelGadgetClass, reqForm.form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, label); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;

  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;

  reqForm.functp = XtCreateManagedWidget("functpanel", 
			   xmDrawingAreaWidgetClass, reqForm.form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNbottomOffset, 5); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("to specify further parameters!)", 
			      XmSTRING_DEFAULT_CHARSET)); cnt++;  
  label = XtCreateManagedWidget("infolabel", 
			   xmLabelGadgetClass, reqForm.form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNbottomWidget, label); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("(Click on request function", 
			      XmSTRING_DEFAULT_CHARSET)); cnt++;  
  label = XtCreateManagedWidget("infolabel", 
			   xmLabelGadgetClass, reqForm.form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNbottomWidget, label); cnt++;
  XtSetValues(reqForm.functp, args, cnt);

  return(0);
}

/* genReqList ****************************************************************
 * generates menu EDITREQ which lists the requests and fills the request form
 ****************************************************************************/

genReqList(parms)
     RequestParm *parms;
{
  register i, ac = 0, cnt = 0;
  exMenu *reqmenu;
  MHmenu *mptr;
  char *nameList[ME_MAXITEM + 1], *ptr;
  Arg args[10];
  
  if (!(reqForm.iptr = rme->menuPtr->i_table[atoi(parms->data)]))
    return;

  for (i = 0; i < reqForm.iptr->i_return->reqCnt; i++) {
    nameList[ac] = Malloc(10);
    sprintf(nameList[ac++], "request %d", i + 1);
  }
  nameList[ac++] = strCopy("new request");

  setWorkApplication();

  if (ac > 0) {
    mptr = buildMenuFromStrings(EDITREQ, MDS_L_POPUP, nameList, NULL, ac);
  }
  else {
    mptr = createMenu();
    initMenu(mptr, MDS_L_POPUP, EDITREQ);
  } 

  reqmenu = exMakeMenu(currentApplication(), NULL, 
		       rme->menuEditDescr->meShell, mptr);
  exCallMenu(reqmenu);
  
  if (ptr = exGetMenuReturnValue(reqmenu)) {
    reqForm.reqIndex = atoi(ptr);
    sprintf(msg, "Request <<%d>>", reqForm.reqIndex + 1);

    XtSetArg(args[cnt], XmNlabelString, 
	     XmStringCreateLtoR(msg, XmSTRING_DEFAULT_CHARSET)); cnt++;  
    XtSetValues(reqForm.titleReq, args, cnt);

    sprintf(msg, "for item <<%s>>:", reqForm.iptr->i_repr->name);

    XtSetArg(args[cnt], XmNlabelString, 
	     XmStringCreateLtoR(msg, XmSTRING_DEFAULT_CHARSET)); cnt++;
    XtSetValues(reqForm.titleItem, args, cnt);

    if (reqForm.iptr->i_requests[reqForm.reqIndex])
      reqForm.req = reqForm.iptr->i_requests[reqForm.reqIndex];
    else {
      reqForm.req = newRequest();
      setReqCnt(reqForm.req, 1, MDS_CNT_EQ);
    }

    resetFuncPanel();
  }

  freeMenu(reqmenu);
  deleteMenu(currentApplication(), mptr);
}

/* meDeleteRequest **********************************************************
 * generates menu DELREQ which lists the requests
 ****************************************************************************/

meDeleteRequest(parms)
     RequestParm *parms;
{
  register i, ac = 0;
  exMenu *reqmenu;
  MHmenu *mptr;
  MHitem *iptr;
  char *nameList[ME_MAXITEM], *retList[ME_MAXITEM], *ptr, *ret;
  int index, reqIndex, fIndex = -1;
  char *strtok();
  
  index = atoi(parms->data);
  if (iptr = rme->menuPtr->i_table[index])
    for (i = 0; i < iptr->i_return->reqCnt; i++) {
      nameList[ac] = Malloc(10);
      sprintf(nameList[ac], "request %d", i + 1);

      retList[ac] = Malloc(5);
      sprintf(retList[ac++], "%d,", i);
    }

  setWorkApplication();

  if (ac > 0) {
    mptr = buildMenuFromStrings(DELREQ, MDS_L_POPUP, nameList, retList, ac);
    setSelType(mptr, MDS_CNT_GE);
  }
  else {
    mptr = createMenu();
    initMenu(mptr, MDS_L_POPUP, DELREQ);
  } 

  reqmenu = exMakeMenu(currentApplication(), NULL, 
		       rme->menuEditDescr->meShell, mptr);
  exCallMenu(reqmenu);
  
  if (ptr = exGetMenuReturnValue(reqmenu)) {
    ret = strtok(ptr, ", ");
    fIndex = atoi(ret);
    while (ret) {
      reqIndex = atoi(ret);
      sprintf(msg, 
	      "Do you really want to delete\n request <%d> for item <%d>?", 
	      reqIndex + 1, index + 1);
      if (exNotifyUser(rme->menuEditDescr->meShell, -1, -1, msg) == ANSWERYES)
	deleteRequest(iptr, reqIndex);
      ret = strtok(NULL, ", ");
    }

    if ((fIndex > -1) && (iptr->i_return->reqCnt > fIndex)) {
                                             /* substitute deleted request */
      while (fIndex < iptr->i_return->reqCnt) {    /* comprise request list */
	if (! iptr->i_requests[fIndex]) {
	  i = fIndex + 1;
	  while (! iptr->i_requests[i])
	    i++;
	  iptr->i_requests[fIndex] = iptr->i_requests[i];
	  iptr->i_requests[i] = (MHiReq *) 0;
	}
	fIndex++;
      }
    }
  }
    
  freeMenu(reqmenu);
  deleteMenu(currentApplication(), mptr);
}

/* editReqForm2 ***********************************************************
 * builds up those elements of the form to edit a special request which are
 * common to all request types
 ****************************************************************************/

static void editReqForm2(parms)     
     RequestParm *parms;
{
  register cnt;
  int mx, my;
  Arg args[20];

  if (atoi(parms->data) == 2)                       /* reqViaForm */
    return;

  getMouse_xy(XtDisplay(rme->menuEditDescr->meShell), &mx, &my);

  cnt = 0;
  XtSetArg(args[cnt], XmNx, mx); cnt++;
  XtSetArg(args[cnt], XmNy, my); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, 2); cnt++;
  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;

  reqFunctForm.reqFunctShell = 
    XtCreatePopupShell("functShell", transientShellWidgetClass, 
		       rme->menuEditDescr->meShell, args, cnt);

  exWidgetPos(reqFunctForm.reqFunctShell, mx, my, 0, 0); 

  reqFunctForm.form = XtCreateManagedWidget("reqFunct", xmFormWidgetClass, 
				      reqFunctForm.reqFunctShell, NULL, 0);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	XmStringCreateLtoR("", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  reqFunctForm.title = XtCreateManagedWidget("functtitle", 
			  xmLabelGadgetClass, reqFunctForm.form, args, cnt);

  /************************ entries form ***********************************/

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, reqFunctForm.title); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;

  reqFunctForm.entryForm = XtCreateManagedWidget("functWork", 
			      xmFormWidgetClass, reqFunctForm.form, args, cnt);
  
  /************************ dialogue buttons ******************************/   

  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;

  XtSetArg(args[cnt], XmNborderWidth, 1); cnt++;
  XtSetArg(args[cnt], XmNlabelString, XmStringCreateLtoR("Quit", 
                                        XmSTRING_DEFAULT_CHARSET)); cnt++;

  reqFunctForm.quitButton = XtCreateManagedWidget("quitB", 
			      xmPushButtonGadgetClass, reqFunctForm.form, 
			      args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, reqFunctForm.quitButton); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;

  XtSetArg(args[cnt], XmNborderWidth, 1); cnt++;
  XtSetArg(args[cnt], XmNlabelString, XmStringCreateLtoR("Save", 
                                        XmSTRING_DEFAULT_CHARSET)); cnt++;

  reqFunctForm.saveButton = XtCreateManagedWidget("saveB", 
			      xmPushButtonGadgetClass, reqFunctForm.form, 
			      args, cnt);

  switch(atoi(parms->data)) {
  case 0:
    editReqViaWnd();
    break;
  case 1:
    editReqViaMenu();
    break;
  case 2:
    break;
  default:
    editReqViaFunct(atoi(parms->data));
  }

  exReqExposeProc(reqFunctForm.reqFunctShell);
  XtPopup(reqFunctForm.reqFunctShell, XtGrabExclusive);
}

/* editReqViaWnd ***********************************************************
 * form to specify a request of a menu option which presents a window with
 * a label to the user
 ****************************************************************************/

static void editReqViaWnd()     
{
  register cnt;
  Widget functLabel;
  Arg args[20];

  reqFunctForm.needSave = False;

  setReqFnct(reqForm.req, reqViaWnd);

  exWidgetPos(reqFunctForm.reqFunctShell, -1, -1, 0, 0); 

  XtAddCallback(reqFunctForm.quitButton, XmNactivateCallback, quitReqViaWnd,
		&reqFunctForm);

  XtAddCallback(reqFunctForm.saveButton, XmNactivateCallback, saveReqViaWnd,
		&reqFunctForm);

  cnt = 0;
  XtSetArg(args[cnt], XmNlabelString, 
	XmStringCreateLtoR("Request Via Window", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  XtSetValues(reqFunctForm.title, args, cnt);

  /************************ labels *************************************/

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, reqFunctForm.title); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("entry label", XmSTRING_DEFAULT_CHARSET)); cnt++;
  functLabel = XtCreateManagedWidget("functlabel", 
			    xmLabelGadgetClass, reqFunctForm.form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, functLabel); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	 XmStringCreateLtoR("entry length", XmSTRING_DEFAULT_CHARSET)); cnt++;
  functLabel = XtCreateManagedWidget("functlabel", 
			    xmLabelGadgetClass, reqFunctForm.form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, functLabel); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString,
	   XmStringCreateLtoR("number of input tokens", 
			      XmSTRING_DEFAULT_CHARSET)); cnt++;
  functLabel = XtCreateManagedWidget("functlabel", 
			    xmLabelGadgetClass, reqFunctForm.form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, functLabel); cnt++;
  XtSetValues(reqFunctForm.entryForm, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, functLabel); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
      XmStringCreateLtoR("(separated by blanks)", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  functLabel = XtCreateManagedWidget("infolabel", 
			    xmLabelGadgetClass, reqFunctForm.form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, functLabel); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString,
	 XmStringCreateLtoR("initial value", XmSTRING_DEFAULT_CHARSET)); cnt++;
  functLabel = XtCreateManagedWidget("functlabel", 
			    xmLabelGadgetClass, reqFunctForm.form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, functLabel); cnt++;
  XtSetValues(reqFunctForm.quitButton, args, cnt);
  XtSetValues(reqFunctForm.saveButton, args, cnt);

  /************************ entries *************************************/

  reqFunctForm.data = createFormEntry(reqFunctForm.entryForm, "",
				      E_ENTRYLEN1, 1);
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetValues(reqFunctForm.data->entry->entryWindow, args, cnt);

  reqFunctForm.size = createFormEntry(reqFunctForm.entryForm, "", 
				      E_ENTRYLEN2, 1);
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, reqFunctForm.data->entry->entryWindow); 
  cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetValues(reqFunctForm.size->entry->entryWindow, args, cnt);

  reqFunctForm.count = createFormEntry(reqFunctForm.entryForm, "", 
				       E_ENTRYLEN2, 1);
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, reqFunctForm.size->entry->entryWindow); 
  cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetValues(reqFunctForm.count->entry->entryWindow, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, reqFunctForm.count->entry->entryWindow); 
  cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("  ", XmSTRING_DEFAULT_CHARSET)); cnt++;
  functLabel = XtCreateManagedWidget("infolabel", xmLabelGadgetClass, 
				     reqFunctForm.entryForm, args, cnt);

  reqFunctForm.initval = createFormEntry(reqFunctForm.entryForm, "", 
					 E_ENTRYLEN1, 1);
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, functLabel); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetValues(reqFunctForm.initval->entry->entryWindow, args, cnt);

  /***************** fill in entries with values of request *****************/

  if (reqForm.req->reqData)
    fillEntry(reqFunctForm.data, reqForm.req->reqData, False);
  if (reqForm.req->initialValue)
    fillEntry(reqFunctForm.initval, reqForm.req->initialValue, False);

  sprintf(msg, "%d", reqForm.req->size);
  fillEntry(reqFunctForm.size, msg, False);

  sprintf(msg, "%d", reqForm.req->cnt);
  fillEntry(reqFunctForm.count, msg, False);
}

/* editReqViaMenu ***********************************************************
 * form to specify a request of a menu option which presents a submenu
 * to the user
 ****************************************************************************/

static void editReqViaMenu()     
{
  register cnt;
  Arg args[20];
  Widget functLabel;

  reqFunctForm.needSave = False;

  setReqFnct(reqForm.req, reqViaMenu);

  exWidgetPos(reqFunctForm.reqFunctShell, -1, -1, 0, 0); 

  XtAddCallback(reqFunctForm.quitButton, XmNactivateCallback, quitReqViaMenu,
		&reqFunctForm);

  XtAddCallback(reqFunctForm.saveButton, XmNactivateCallback, saveReqViaMenu,
		&reqFunctForm);

  cnt = 0;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("Request Via Menu", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  XtSetValues(reqFunctForm.title, args, cnt);

  /************************ labels *************************************/

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, reqFunctForm.title); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("menu name", XmSTRING_DEFAULT_CHARSET)); cnt++;
  functLabel = XtCreateManagedWidget("functlabel", 
			    xmLabelGadgetClass, reqFunctForm.form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, functLabel); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	 XmStringCreateLtoR("number of selectable items",
			    XmSTRING_DEFAULT_CHARSET)); cnt++;
  functLabel = XtCreateManagedWidget("functlabel", 
			    xmLabelGadgetClass, reqFunctForm.form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, functLabel); cnt++;
  XtSetValues(reqFunctForm.entryForm, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, functLabel); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString,
	 XmStringCreateLtoR("initial value", XmSTRING_DEFAULT_CHARSET)); cnt++;
  functLabel = XtCreateManagedWidget("functlabel", 
			    xmLabelGadgetClass, reqFunctForm.form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, functLabel); cnt++;
  XtSetValues(reqFunctForm.quitButton, args, cnt);
  XtSetValues(reqFunctForm.saveButton, args, cnt);

  /************************ entries *************************************/

  reqFunctForm.data = createFormEntry(reqFunctForm.entryForm, "", 
				      E_ENTRYLEN1, 1);
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetValues(reqFunctForm.data->entry->entryWindow, args, cnt);

  reqFunctForm.count = createFormEntry(reqFunctForm.entryForm, "",
				       E_ENTRYLEN2, 1);
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, reqFunctForm.data->entry->entryWindow); 
  cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetValues(reqFunctForm.count->entry->entryWindow, args, cnt);

  reqFunctForm.initval = createFormEntry(reqFunctForm.entryForm, "",
					 E_ENTRYLEN1, 1);
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, reqFunctForm.count->entry->entryWindow); 
  cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetValues(reqFunctForm.initval->entry->entryWindow, args, cnt);

  /***************** fill in entries with values of request *****************/

  if (reqForm.req->reqData)
    fillEntry(reqFunctForm.data, reqForm.req->reqData, False);
  if (reqForm.req->initialValue)
    fillEntry(reqFunctForm.initval, reqForm.req->initialValue, False);

  sprintf(msg, "%d", reqForm.req->cnt);
  fillEntry(reqFunctForm.count, msg, False);
}

/* editReqViaFunct ***********************************************************
 * form to specify a request of a menu option which on selection executes
 * a user defined function
 ****************************************************************************/

static void editReqViaFunct(findex)  
     int findex;                        /* index of user function */
{
  register cnt;
  Arg args[20];
  Widget functLabel;

  reqFunctForm.needSave = False;

  if (findex > 2)
    setReqFnct(reqForm.req, rme->applPtr->ApplFunc[findex - 3 + 1].func);

  exWidgetPos(reqFunctForm.reqFunctShell, -1, -1, 0, 0); 

  XtAddCallback(reqFunctForm.quitButton, XmNactivateCallback, quitReqViaFunct,
		&reqFunctForm);

  XtAddCallback(reqFunctForm.saveButton, XmNactivateCallback, saveReqViaFunct,
		&reqFunctForm);

  cnt = 0;
  XtSetArg(args[cnt], XmNlabelString, 
	 XmStringCreateLtoR("Request Via Function", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  XtSetValues(reqFunctForm.title, args, cnt);

  /************************ labels *************************************/

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, reqFunctForm.title); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("text data", XmSTRING_DEFAULT_CHARSET)); cnt++;
  functLabel = XtCreateManagedWidget("functlabel", 
			    xmLabelGadgetClass, reqFunctForm.form, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, functLabel); cnt++;
  XtSetValues(reqFunctForm.entryForm, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, functLabel); cnt++;
  XtSetValues(reqFunctForm.quitButton, args, cnt);
  XtSetValues(reqFunctForm.saveButton, args, cnt);

  /************************ entries *************************************/

  reqFunctForm.data = createFormEntry(reqFunctForm.entryForm, "", 
				      E_ENTRYLEN1, 1);
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetValues(reqFunctForm.data->entry->entryWindow, args, cnt);

  /***************** fill in entries with values of request *****************/

  if (reqForm.req->reqData)
    fillEntry(reqFunctForm.data, reqForm.req->reqData, False);
}

/* checkReqEdit ************************************************************
 * function to destroy request definition editor
 ****************************************************************************/

void checkReqEdit(mEditWin)
     meditWindow *mEditWin;
{
  if (mEditWin->com_flag & REQEDON)
    quitReqEdit((RequestParm *) 0);
}


/* quitReqEdit **************************************************************
 * request editor application function to save request definition
 ****************************************************************************/

void quitReqEdit(parms)
     RequestParm *parms;
{
  updateReqCnt(rme->menuEditDescr, rme->menuPtr, rme->menuPtr->itemCnt);

  if (reqFunctForm.reqFunctShell) {
    XtDestroyWidget(reqFunctForm.reqFunctShell);
    reqFunctForm.reqFunctShell = NULL;
  }

  XtPopdown(reqForm.reqFormShell);
  subCompState(rme->menuEditDescr, 0, False);
}


/* quitReqViaWnd **********************************************************
 * request editor application function to save request definition
 ****************************************************************************/
/*ARGSUSED*/
static void quitReqViaWnd(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  int imsg, LOOP = 1;
  struct _reqFunctForm *editor = (struct _reqFunctForm *) closure;

  while (LOOP) {

    if (!cmpFormEntry(editor->data, reqForm.req->reqData)) {	
      editor->needSave = True;
      break;
    }
  
    if (!cmpFormEntry(editor->initval, reqForm.req->initialValue)) {
      editor->needSave = True;
      break;
    }

    if (checkFormEntry(editor->size)) { 
      imsg = atoi(editor->size->string[0]);
      if (reqForm.req->size != imsg) {
	editor->needSave = True;
	break;
      }
    }

    if (checkFormEntry(editor->count)) { 
      imsg = atoi(editor->count->string[0]);
      if (reqForm.req->cnt != imsg) {
	editor->needSave = True;
	break;
      }
    }

    LOOP = 0;
  }

  if (editor->needSave)
    if (exNotifyUser(rme->menuEditDescr->meShell, -1, -1, 
	       "Request definition not saved ! Quit anyway ?") == ANSWERNO)
      return;

  exUngrabAll(editor->reqFunctShell);
  XtPopdown(editor->reqFunctShell);	  

  updateReqCnt(rme->menuEditDescr, rme->menuPtr, rme->menuPtr->itemCnt);
}

/* saveReqViaWnd ************************************************************
 * request editor application function to save request definition
 ****************************************************************************/
/*ARGSUSED*/
static void saveReqViaWnd(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  int imsg;
  struct _reqFunctForm *editor = (struct _reqFunctForm *) closure;
  
  if (! reqForm.iptr->i_requests[reqForm.reqIndex])
    insertRequest(reqForm.iptr, reqForm.req, reqForm.reqIndex);

  if (!cmpFormEntry(editor->data, reqForm.req->reqData)) {
    free(reqForm.req->reqData);
    reqForm.req->reqData = strCopy(editor->data->string[0]);
  }

  if (!cmpFormEntry(editor->initval, reqForm.req->initialValue)) {
    free(reqForm.req->initialValue);
    reqForm.req->initialValue = strCopy(editor->initval->string[0]);
  }

  if (checkFormEntry(editor->size)) {
    imsg = atoi(editor->size->string[0]);
    if (reqForm.req->size != imsg)
      reqForm.req->size = imsg;
  }

  if (checkFormEntry(editor->count)) { 
    imsg = atoi(editor->count->string[0]);
    if (reqForm.req->cnt != imsg)
      reqForm.req->cnt = imsg;
  }

  resetFuncPanel();

  editor->needSave = False;
  rme->applPtr->_changed = TRUE;
}

/* quitReqViaMenu **********************************************************
 * request editor application function to save request definition
 ****************************************************************************/
/*ARGSUSED*/
static void quitReqViaMenu(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  int imsg, LOOP = 1;
  struct _reqFunctForm *editor = (struct _reqFunctForm *) closure;

  while (LOOP) {

    if (!cmpFormEntry(editor->data, reqForm.req->reqData)) {	
      editor->needSave = True;
      break;
    }
  
    if (!cmpFormEntry(editor->initval, reqForm.req->initialValue)) {
      editor->needSave = True;
      break;
    }

    if (checkFormEntry(editor->count)) { 
      imsg = atoi(editor->count->string[0]);
      if (reqForm.req->cnt != imsg) {
	editor->needSave = True;
	break;
      }
    }

    LOOP = 0;
  }

  if (editor->needSave)
    if (exNotifyUser(rme->menuEditDescr->meShell, -1, -1, 
		 "Request definition not saved ! Quit anyway ?") == ANSWERNO)
      return;

  exUngrabAll(editor->reqFunctShell);
  XtPopdown(editor->reqFunctShell);	  

  updateReqCnt(rme->menuEditDescr, rme->menuPtr, rme->menuPtr->itemCnt); 
}

/* saveReqViaMenu ***********************************************************
 * request editor application function to save request definition
 ****************************************************************************/
/*ARGSUSED*/
static void saveReqViaMenu(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  int imsg;
  struct _reqFunctForm *editor = (struct _reqFunctForm *) closure;
  
  if (! reqForm.iptr->i_requests[reqForm.reqIndex])
    insertRequest(reqForm.iptr, reqForm.req, reqForm.reqIndex);

  if (!cmpFormEntry(editor->data, reqForm.req->reqData)) {
    free(reqForm.req->reqData);
    reqForm.req->reqData = strCopy(editor->data->string[0]);
  }

  if (!cmpFormEntry(editor->initval, reqForm.req->initialValue)) {
    free(reqForm.req->initialValue);
    reqForm.req->initialValue = strCopy(editor->initval->string[0]);
  }

  if (checkFormEntry(editor->count)) { 
    imsg = atoi(editor->count->string[0]);
    if (reqForm.req->cnt != imsg)
      reqForm.req->cnt = imsg;
  }

  resetFuncPanel();

  editor->needSave = False;
  rme->applPtr->_changed = TRUE;
}

/* quitReqViaFunct **********************************************************
 * request editor application function to save request definition
 ****************************************************************************/
/*ARGSUSED*/
static void quitReqViaFunct(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  int imsg;
  struct _reqFunctForm *editor = (struct _reqFunctForm *) closure;

  if (!cmpFormEntry(editor->data, reqForm.req->reqData))
    editor->needSave = True;

  if (editor->needSave)
    if (exNotifyUser(rme->menuEditDescr->meShell, -1, -1, 
		 "Request definition not saved ! Quit anyway ?") == ANSWERNO)
      return;

  exUngrabAll(editor->reqFunctShell);
  XtPopdown(editor->reqFunctShell);	  

  updateReqCnt(rme->menuEditDescr, rme->menuPtr, rme->menuPtr->itemCnt);
}

/* saveReqViaFunc ************************************************************
 * request editor application function to save request definition
 ****************************************************************************/
/*ARGSUSED*/
static void saveReqViaFunct(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  int imsg;
  struct _reqFunctForm *editor = (struct _reqFunctForm *) closure;
  
  if (! reqForm.iptr->i_requests[reqForm.reqIndex])
    insertRequest(reqForm.iptr, reqForm.req, reqForm.reqIndex);

  if (!cmpFormEntry(editor->data, reqForm.req->reqData)) {
    free(reqForm.req->reqData);
    reqForm.req->reqData = strCopy(editor->data->string[0]);
  }

  resetFuncPanel();

  editor->needSave = False;
  rme->applPtr->_changed = TRUE;
}

/* resetFuncPanel: ***********************************************************
 * adjust default item of panel with request functions
 ****************************************************************************/

static void resetFuncPanel()
{
  MHmenu *panel;
  int actmask[ME_MAXITEM];
  register i;

  panel = getMenuDescrByName("funcListPanel");

  setMask(actmask, panel->itemCnt, 1);
  setMenuInitialActivityMask(panel, actmask);
  resetSelection(panel, reqForm.menup);

  if ((i = mds_function(reqForm.req->reqFnct)) > 0)
    setDefaultItem(panel->i_table[i - 2]);
  else if ((i = user_function(reqForm.req->reqFnct, rme->applPtr)) > 0)
    setDefaultItem(panel->i_table[i + 3 - 1]);

  reqForm.menup->erase = False;
  exMakeDefaultMenu(reqForm.menup); 
                       
  XtUnmapWidget(reqForm.menup->pane);                 
  XtMapWidget(reqForm.menup->pane);                /* redisplay */
}


/* grabReqComp: ***************************************************************
 * grab pointer for request component
 ****************************************************************************/
/*ARGSUSED*/
static void grabReqComp(w, event, params, num_params)
     Widget w;
     XKeyEvent *event;
     String *params;
     Cardinal *num_params;
{
  exGrabAll(w, None);
}

/* exReqExposeProc: ********************************************************
 * add translations in order to actively grab pointer when a request 
 * component is exposed
 ****************************************************************************/

/*ARGSUSED*/
static void exReqExposeProc(w)
     Widget w;
{
  static Boolean INIT = False;
  static XtTranslations ptt;
 
  static XtActionsRec reqCmds[] = {
    {"grabReqComp", grabReqComp}
  };
  static char reqCmdTable[] = 
    "<Expose>		: grabReqComp()";

  if (!INIT) {
    XtAddActions(reqCmds, 1);
    ptt = XtParseTranslationTable(reqCmdTable);
    INIT = True;
  }
  
  XtOverrideTranslations(w, ptt);
}

/* getRequestResult ***********************************************************
 * 
 ****************************************************************************/

static void getRequestResult(w, event, params, num_params)
     Widget w;
     XPropertyEvent *event;
     String *params;
     Cardinal *num_params;
{
}


/* subCompState ************************************************************
 * records which subcomponents are running and adjusts activity mask of
 * menu 'edit features' accordingly
 ***************************************************************************/

subCompState(mEditWin, index, on)
     meditWindow *mEditWin;
     int index;             /* index of subcomponent in menu 'edit features' */
     Boolean on;            /* indicates if component corresponding to 'index'
			       is running */
{
  static int menuMask[3] = {1, 1, 1};
  exMenu *mWindow;
  int compFlag;

  if (!(mWindow = getMenuByName(EDITFEATURES, EditMenuWin, WinCnt))) {
    setError(ERR_NULL);
    return;
  }

  switch(index) {
  case 0:
    compFlag = REQEDON;
    break;
  case 1:
    compFlag = ACTEDON;
    break;
  case 2:
    compFlag = HELPEDON;
    break;
  default:
    return;
  }

  if (on) {
    mEditWin->com_flag |= compFlag;
    menuMask[index] = 0;
  }
  else {
    mEditWin->com_flag &= ~compFlag;
    menuMask[index] = 1;
  }

  setInitialMask(mWindow, menuMask);
}

/* checkSubComp **************************************************************
 * quit all running subcomponents if current menu is changed in menu editor
 *****************************************************************************/

checkSubComp(editw)
     meditWindow *editw;
{
  checkActMaskEditor(editw);
  checkReqEdit(editw);
  checkHelpEditor(editw);
}





